Package: seacarb
Title: Seawater Carbonate Chemistry
Version: 3.3.4
Date: 2026-04-16
LazyData: true
Authors@R: c(
  person("Jean-Pierre", "Gattuso", role = c("aut"), email = "jean-pierre.gattuso@imev-mer.fr"),
  person("James", "Orr", role = c("aut", "cre", "cph"), email = "james.orr@lsce.ipsl.fr"),
  person("Jean-Marie", "Epitalon", role = "aut", email = "jean-marie.epitalon@orange.fr"),
  person("Heloise", "Lavigne", role = "aut", email = "lavigne@obs-vlfr.fr"),
  person("Bernard", "Gentili", role = "ctb"),
  person("Mathilde", "Hagens", role = "ctb"),
  person("Andreas", "Hofmann", role = "ctb"),
  person("Jens-Daniel", "Mueller", role = "ctb"),
  person("Aurélien", "Proye", role = "ctb"),
  person("James", "Rae", role = "ctb"),
  person("Karline", "Soetaert", role = "ctb")
  )
Maintainer: James Orr <james.orr@lsce.ipsl.fr>
Depends: R (>= 3.5.0), oce, gsw, SolveSAPHE
Description: Calculates parameters of the seawater carbonate system and assists in design of ocean acidification perturbation experiments.
Encoding: UTF-8
URL: https://CRAN.R-project.org/package=seacarb
License: GPL (>= 2)
Repository: CRAN
RoxygenNote: 7.1.1
NeedsCompilation: no
Packaged: 2026-04-16 19:34:18 UTC; orr
Author: Jean-Pierre Gattuso [aut],
  James Orr [aut, cre, cph],
  Jean-Marie Epitalon [aut],
  Heloise Lavigne [aut],
  Bernard Gentili [ctb],
  Mathilde Hagens [ctb],
  Andreas Hofmann [ctb],
  Jens-Daniel Mueller [ctb],
  Aurélien Proye [ctb],
  James Rae [ctb],
  Karline Soetaert [ctb]
Date/Publication: 2026-04-17 10:10:02 UTC
Built: R 4.6.0; ; 2026-04-26 01:09:08 UTC; unix

Package: seacarb
Title: Seawater Carbonate Chemistry
Version: 3.3.4
Date: 2026-04-16
LazyData: true
Authors@R: c(
  person("Jean-Pierre", "Gattuso", role = c("aut"), email = "jean-pierre.gattuso@imev-mer.fr"),
  person("James", "Orr", role = c("aut", "cre", "cph"), email = "james.orr@lsce.ipsl.fr"),
  person("Jean-Marie", "Epitalon", role = "aut", email = "jean-marie.epitalon@orange.fr"),
  person("Heloise", "Lavigne", role = "aut", email = "lavigne@obs-vlfr.fr"),
  person("Bernard", "Gentili", role = "ctb"),
  person("Mathilde", "Hagens", role = "ctb"),
  person("Andreas", "Hofmann", role = "ctb"),
  person("Jens-Daniel", "Mueller", role = "ctb"),
  person("Aurélien", "Proye", role = "ctb"),
  person("James", "Rae", role = "ctb"),
  person("Karline", "Soetaert", role = "ctb")
  )
Maintainer: James Orr <james.orr@lsce.ipsl.fr>
Depends: R (>= 3.5.0), oce, gsw, SolveSAPHE
Description: Calculates parameters of the seawater carbonate system and assists in design of ocean acidification perturbation experiments.
Encoding: UTF-8
URL: https://CRAN.R-project.org/package=seacarb
License: GPL (>= 2)
Repository: CRAN
RoxygenNote: 7.1.1
NeedsCompilation: no
Packaged: 2026-04-16 19:34:18 UTC; orr
Author: Jean-Pierre Gattuso [aut],
  James Orr [aut, cre, cph],
  Jean-Marie Epitalon [aut],
  Heloise Lavigne [aut],
  Bernard Gentili [ctb],
  Mathilde Hagens [ctb],
  Andreas Hofmann [ctb],
  Jens-Daniel Mueller [ctb],
  Aurélien Proye [ctb],
  James Rae [ctb],
  Karline Soetaert [ctb]
Date/Publication: 2026-04-17 10:10:02 UTC
Built: R 4.6.0; ; 2026-04-27 10:39:13 UTC; unix

# Export all names exept "caculate_carb", which is internal

export(amp, bjerrum, bor, buffergen, buffer, buffesm, carbb, carbfull, carb, d2p, derivnum)
export(eos2teos_chem, eos2teos_geo, errors, f2pCO2, fCO2insi, K0, K1p, K1, K2p, K2, K2si, K3p, Kb)
export(kconv, Kf, Khs, Kn, Ksi, Kspa, Kspc, Ks, Kw, oa, Om, p2d, p2fCO2, p2xCO2, pCa)
export(pCO2insi, Pcoeffs, Pcorrect, pgas, pHconv, pHinsi, pH, pHslope, pHspec, pmix, ppH, psi, pTA)
export(rho, sa2sp_chem, sa2sp_geo, sf_calc, sf_calib)
export(sir_b, sir_full, sir, sp2sa_chem, sp2sa_geo, speciation, teos2eos_chem, teos2eos_geo)
export(theta, tris, vapress, x2pCO2)
export(fH, pHnbs2sws, pHsws2nbs)

importFrom(utils, data)
importFrom("graphics", "arrows", "contour", "matlines", "matplot",
             "mtext", "par")
importFrom("stats", "lm", "nls", "uniroot", "rnorm")
importFrom(oce, swTheta)
importFrom(gsw, gsw_CT_from_t, gsw_SA_from_SP, gsw_SP_from_SA, gsw_t_from_CT)
importFrom(SolveSAPHE, solve_pH_from_AT)

seacarb
=======

Carbonate chemistry with R

In 2003, Aurélien Proye and I put together seacarb, an R package that calculates various parameters of the carbonate system in seawater. The package was subsequently upgraded. In 2008, a new version (2.0) was built with the assistance of Héloïse Lavigne. Five additional functions were included to assist the design of perturbation experiments to investigate ocean acidification and other functions were revised in order to strictly follow the "Guide to best practices for ocean CO2 measurements" (Dickson et al., 2007). Seacarb uses equations mostly from the following sources:

- Dickson A. G., Sabine C. L. & Christian J. R., 2007. Guide to best practices for ocean CO2 measurements. PICES Special Publication 3:1-191.
- DOE, 1994. Handbook of methods for the analysis of the various parameters of the carbon dioxide system in sea water; version 2. Dickson, A. G. and Goyet, C., editors. ORNL/CDIAC-74, 1994.
- Frankignoulle M., 1994. A complete set of buffer factors for acid/base CO2 system in seawater. Journal of Marine Systems 5: 111-118.
- Orr J. C., Epitalon J.-M., Dickson A. G. & Gattuso J.-P., 2018. Routine uncertainty propagation for the marine carbon dioxide system. Marine Chemistry 207:84-107.
- Zeebe R. E. & Wolf-Gladrow D. A., 2001. CO2 in seawater: equilibrium, kinetics, isotopes. Amsterdam: Elsevier, 346 pp.

A portion of the code has been adapted, with permission from the authors, from the Matlab files mentioned above. I am grateful to Richard Zeebe and Dieter Wolf-Gladrow for that. Héloïse Lavigne contributed a lot to seacarb in 2008 and 2009; this was critical to the launch of version 2.0 and subsequent updates. Portions of code and/or corrections have also been contributed by Jean-Marie Epitalon (2004, 2018), Bernard Gentili (2006), Karline Soetaert (2007) and Jim Orr (2007, 2010, 2018). Jean-Marie Epitalon considerably improved the code, leading to much faster calculations in version 3.0.

Orr et al. (2015) assessed seacarb together with other packages which calculate the seawater carbonate chemistry. In 2018, option to calculate uncertainty propagation was added (Orr et al., 2018).

- Orr J. C., Epitalon J.-M., Dickson A. G. & Gattuso J.-P., 2018. Routine uncertainty propagation for the marine carbon dioxide system. Marine Chemistry 207:84-107. http://dx.doi.org/10.1016/j.marchem.2018.10.006
- Orr J. C., Epitalon J.-M. & Gattuso J.-P., 2015. Comparison of ten packages that compute ocean carbonate chemistry. Biogeosciences 12:1483-1510. http://www.biogeosciences.net/12/1483/2015/

This program is provided free under the GNU General Public License (GNU GPL). It will be improved using the comments that I will receive. If you are new to R, please check the manuals and FAQs available on the R-project web site to get information on how to install R and the seacarb package on your system. Please only report and comment on seacarb, not on general problems related to R.

Briefly, after installing R and if you have an Internet connection, here is the simplest way to install seacarb:

- Launch R
- To install seacarb (to be done only once), type the following command: install.packages("seacarb")
- To load the seacarb package into memory in order to use it (to be done each time R is launched), type the following command: library(seacarb)

The seacarb package can be downloaded from the Comprehensive R Archive Network (CRAN; see link below). The documentation is included in the package and is accessible using standard R commands. Please give due credit to the publications mentioned above and cite seacarb as follows:

Gattuso J.-P., Epitalon J.-M., Lavigne H. & Orr J., 2021. seacarb: seawater carbonate chemistry. R package version 3.3.0. http://CRAN.R-project.org/package=seacarb

# Copyright (C) 2009 Jean-Pierre Gattuso
# Copyright (C) 2018 Jens Daniel Mueller, update
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"pHspec" <-
function(S=35, T=25, R=1, d="mCP", k="m18", warn="y")
  {
  #-------Harmonize input vector length-----
  
  n <- max(length(S), length(T), length(d), length(k))
  
  if(length(S)!=n){S <- rep(S[1],n)}
  if(length(T)!=n){T <- rep(T[1],n)}
  if(length(R)!=n){R <- rep(T[1],n)}
  if(length(d)!=n){d <- rep(d[1],n)}
  if(length(k)!=n){k <- rep(k[1],n)}


  #-------Initialise output vector------- 
  
  pHspec <- rep(NA, n)
  method <- rep(NA, n)
  
  #-------Change temperature scale from [deg C] to [K] ----
  
  TK = T + 273.15;           # T [C]; TK [K]
  
  #-------Calculate sample pH based on input values-------
  
 	#--------------------------------------------------------------
	#-------mCP characterization by Mueller and Rehder 2018-----------
	#
	#       https://doi.org/10.3389/fmars.2018.00177  
	#     
	#       Calculates pH of seawater, brackish, and freshwater samples
	#       based on absorbance ratios R obtained from spectrophotometric 
	#       measurements with the dye m-Cresol purple 
	#
	#       pH-scale: total scale
	#        
	#       correct for T range : 5 - 35 °C
	#       correct for S range : 0 - 40
	#--------------------------------------------------------------
	
  pHspec_m18 <-
    
    #firsTK seTK of coefficienTKs defines pK2e2 = f(Sal, TKem)
    
    1.08071477e+03                      -
    1.35394946e-01  *S^0.5            -   
    1.98063716e+02  *S^1.5            +
    6.31924397e+01  *S^2              -
    5.18141866e+00  *S^2.5            -
    2.66457425e+04  *TK^-1             +
    5.08796578e+03  *S^1.5 * TK^-1   -
    1.62454827e+03  *S^2 * TK^-1     +
    1.33276788e+02  *S^2.5 * TK^-1   -
    1.89671212e+02  *log(TK)           +
    3.49038762e+01  *S^1.5 * log(TK) -
    1.11336508e+01  *S^2 * log(TK)   +
    9.12761930e-01  *S^2.5 * log(TK) +
    3.27430677e-01  *TK              -
    7.51448528e-04  *S^0.5 * TK      +
    3.94838229e-04  *S * TK          -
    6.00237876e-02  *S^1.5 * TK      +
    1.90997693e-02  *S^2 * TK        -
    1.56396488e-03  *S^2.5 * TK      +
    
    #second seTK of coefficienTKs includes TKhe definiTKion of mCP absorpTKiviTKy raTKios e1 and e3/e3
    #as deTKermined by Liu eTK al. (2011) and defines TKhe log-TKerm calculaTKion 
    
    log10(
      (R -
         (-0.007762 + 4.5174e-5*TK)) /
        (1 - (R *  (- 0.020813 + 2.60262e-4*TK + 1.0436e-4*(S-35))))
    )
  
  
  
	
  #-------------------------------------------------------------------
	#-------Choose between mCP characterizations (=method)----------
	
  is_m18 <- (k=='m18')
	pHspec[is_m18] <- pHspec_m18[is_m18]
	
	#-------Assign method -------------------
	
  method[is_m18] <- "Mueller and Rehder (2018)"
	
	#-------Set warnings-----------
	is_w <- warn == "y"
	if (any(is_w & is_m18 & (T>35 | T<5 | S>40))) 
	  {warning("S, and/or T is outside the range of validity for the mCP characterization by Mueller and Rehder (2018).")}
	
	#-------Assign attributes and define return value-----------
	
	attr(pHspec, "method") = method
	attr(pHspec, "pH scale") = "total scale"
	attr(pHspec,"unit") <- "mol/kg-soln"
	return(pHspec)

}

# Copyright (C) 2008 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"pCa" <- function(flag,  var1,  var2, Ca, S=35, T=20, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", eos="eos80", long=1.e20, lat=1.e20){
  # if the concentrations of total silicate and total phosphate are NA
  # they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  
# Only two options for eos
if (eos != "teos10" && eos != "eos80")
    stop ("invalid parameter eos: ", eos)

# if use of EOS-10 standard
if (eos == "teos10")
{
    # Must convert temperature and salinity from TEOS-10 to EOS-80
    # convert temperature: from Conservative (CT) to in-situ temperature
    # and salinity from Absolute to Practical (SP)
    STeos <- teos2eos_geo (S, T, P, long, lat)
    InsT <- STeos$T
    SP <- STeos$SP
}
else
{
    InsT <- T
    SP <- S
}

# pass to carb() original S and T to have them insert in the result table
ci <- carb(flag, var1= var1, var2= var2, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks="d", pHscale=pHscale, b=b, eos=eos, long=long, lat=lat) # initial carbonate chemistry
	#0.01028*(S/35) # calcium concentration (Dickson et al., 2007)
	#co <- ci
	#if (length(Ca) > 1) {for(i in 1:(length(Ca)-1)) {co <- rbind(co,c)}}
	Oa <- Ca * ci$CO3 / Kspa(S=SP, T=InsT, P=P)
	Oc <- Ca * ci$CO3 / Kspc(S=SP, T=InsT, P=P)
	cf <- ci
	if (length(Ca) > 1) {for(i in 1:(length(Ca)-1)) {cf <- rbind(cf,ci)}}
	co <- as.data.frame(c("pCa-initial", rep("pCa-final", nrow(cf))))
	cf$OmegaAragonite <- Oa
	cf$OmegaCalcite <- Oc
	out <- rbind(ci, cf)
	out <- cbind(co, out)
	names(out)[1] <- "comment"
	return(out[,1:20])
}

# Copyright (C) 2016 Mathilde Hagens (M.Hagens@uu.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

K2si <- function (S=35, T=25, P=0, pHscale="T",
                  kSWS2scale="x", ktotal2SWS_P0="x") 
{
  
  nK <- max(length(S), length(T), length(P), length(pHscale), 
            length(kSWS2scale), length(ktotal2SWS_P0))
  
  ##-------- Creation of vector for all entries
  
  if (length(S) != nK) {S <- rep(S[1], nK)}
  if (length(T) != nK) {T <- rep(T[1], nK)}
  if (length(P) != nK) {P <- rep(P[1], nK)}
  if (length(pHscale) != nK) {pHscale <- rep(pHscale[1], nK)}
  
  #-------Constants----------------
  
  tk = 273.15       # [K] (for conversion [deg C] <-> [K])
  TC = T + tk       # TC [C]; T[K]
  
  #--------------------------------------------------------------
  # Dissociation constant of SiO(OH)3 on total scale
  # Source: Wischmeyer et al. (2003), including corrections by 
  # D. Wolf-Gladrow and M. Hagens
  #--------------------------------------------------------------
  
  logK <- 8.9613 + (-4465.18) / TC + (-0.021952) * TC
  K2si <- 10^(logK)
  
  ## ---- Conversion from Total scale to seawater scale before pressure corrections
  
  # if correction factor (from Total scale to seawater at P=0) not given
  if (missing(ktotal2SWS_P0) || ktotal2SWS_P0[1] == "x")
  {
    # Compute it
    ktotal2SWS_P0 <- kconv(S=S, T=T, P=0)$ktotal2SWS
  }
  else
    # Check its length
    if(length(ktotal2SWS_P0)!=nK) ktotal2SWS_P0 <- rep(ktotal2SWS_P0[1], nK)
  K2si <- K2si * ktotal2SWS_P0
  
  # If needed, pressure correction. Conversion factors for pressure
  # correction are calculated within Pcorrect
  
  if (any(P != 0)) {
    K2si <- Pcorrect(Kvalue = K2si, Ktype = "K2si", T = T, S = S, 
                     P = P, pHscale = "SWS") }
  
  ###----------------pH scale corrections
  
  # Which pH scales are required ?
  is_total <- pHscale=="T"
  is_free   <- pHscale=="F"
  
  # if any pH scale correction required (from total scale)
  if (any(is_total) || any(is_free))
  {
    # if pH scale correction factor not given
    if (missing(kSWS2scale) || kSWS2scale[1] == "x")
    {
      # Compute it
      kSWS2scale <- rep(1.0,nK)
      if (any(is_total)) 
        {kSWS2scale[is_total] <- kconv(S=S[is_total], T=T[is_total], P=P[is_total])$kSWS2total}
      if (any(is_free))  
        {kSWS2scale[is_free]  <- kconv(S=S[is_free], T=T[is_free], P=P[is_free])$kSWS2free}
    }
    else
      # Check its length
      if(length(kSWS2scale)!=nK){kSWS2scale <- rep(kSWS2scale[1], nK)}
    # Apply pH scale correction
    K2si <- K2si*kSWS2scale
  }
  
  # Return full name of pH scale
  pHsc <- rep(NA,nK)
  pHsc[is_total] <- "total scale"
  pHsc[is_free]  <- "free scale"
  pHsc[!is_total & !is_free] <- "seawater scale"
  
  ##------------Warnings
  ## No warning is given; range of validity for this equation not clear from Wischmeyer et al (2003)
  
  attr(K2si, "unit") = "mol/kg-soln"
  attr(K2si, "pH scale") = pHsc
  return(K2si)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Heloise Lavigne and Aurelien Proye # with a most valuable contribution of Bernard Gentili <gentili@obs-vlfr.fr> 
# and valuable suggestions from Jean-Marie Epitalon <epitalon@lsce.saclay.cea.fr> 
# # This file is part of seacarb. 
# # Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version. 
# # Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details. 
# # You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA # #  

# New version of buffesm accounts for effects of nutrient (Si and P) concentrations
# ---------------------------------------------------------------------------------

buffesm <-  
  function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", warn="y",  eos="eos80", long=1.e20, lat=1.e20){
    n <- max(length(flag), length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), length(ks), length(b))
    if(length(flag)!=n){flag <- rep(flag[1],n)}
    if(length(var1)!=n){var1 <- rep(var1[1],n)}
    if(length(var2)!=n){var2 <- rep(var2[1],n)}
    if(length(S)!=n){S <- rep(S[1],n)}
    if(length(T)!=n){T <- rep(T[1],n)}
    if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
    if(length(P)!=n){P <- rep(P[1],n)}
    if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
    if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
    if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
    if(length(kf)!=n){kf <- rep(kf[1],n)}
    if(length(ks)!=n){ks <- rep(ks[1],n)}
    if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
    if(length(b)!=n){b <- rep(b[1],n)}

    # if the concentrations of total silicate and total phosphate are NA
    # they are set at 0
    Sit[is.na(Sit)] <- 0
    Pt[is.na(Pt)] <- 0

    # Only two options for eos
    if (eos != "teos10" && eos != "eos80")
        stop ("invalid parameter eos: ", eos)
    
    # if use of EOS-10 standard
    if (eos == "teos10")
    {
        # Must convert temperature and salinity from TEOS-10 to EOS-80
        # convert temperature: from Conservative (CT) to in-situ temperature
        # and salinity from Absolute to Practical (SP)
        eos <- teos2eos_geo (S, T, P, long, lat)
        InsT <- eos$T
        SP <- eos$SP
    }
    else
    {
        InsT <- T
        SP <- S
    }

     Carb <- carb(flag=flag, var1=var1, var2=var2, S=SP, T=InsT, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b)
 	P    <- Carb$P
 	pH   <- Carb$pH
	h    <- 10^(-pH)
	CO2  <- Carb$CO2
	HCO3 <- Carb$HCO3
	CO3  <- Carb$CO3
	DIC  <- Carb$DIC
	ALK  <- Carb$ALK
	Oa   <- Carb$OmegaAragonite
	Oc   <- Carb$OmegaCalcite

    #-------Constants----------------  
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK = InsT + tk;           # TK [K]; InsT[C]
    
    Cl = SP / 1.80655;            # Cl = chlorinity; SP = practical salinity (psu)
    ST = 0.14 * Cl/96.062        # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
    FLUO = 6.7e-5 * Cl/18.9984   # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
    bor = bor(S=SP , b=b)         # (mol/kg) total boron

    #---------------------------------------------------------------------
    #--------------------- compute K's ----------------------------------
    #---------------------------------------------------------------------
    
    # Ks (free pH scale) at zero pressure and given pressure
    Ks_P0 <- Ks(S=SP, T=InsT, P=0, ks=ks, warn=warn)
    Ks    <- Ks(S=SP, T=InsT, P=P, ks=ks, warn=warn)
    
    # Kf on free pH scale
    Kff_P0 <- Kf(S=SP, T=InsT, P=0, pHscale="F", kf=kf, Ks_P0, Ks)
    Kff <- Kf(S=SP, T=InsT, P=P, pHscale="F", kf=kf, Ks_P0, Ks)
    # Kf on given pH scale
    Kf <- Kf(S=SP, T=InsT, P=P, pHscale=pHscale, kf=kf, Ks_P0, Ks)
    
    # Conversion factor from total to SWS pH scale at zero pressure
    ktotal2SWS_P0 <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks_P0,Kff=Kff_P0)$ktotal2SWS

    # Conversion factor from SWS to chosen pH scale
    conv <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks,Kff=Kff)
    kSWS2chosen <- rep(1.,n)
    kSWS2chosen [pHscale == "T"] <- conv$kSWS2total [pHscale == "T"]
    kSWS2chosen [pHscale == "F"] <- conv$kSWS2free [pHscale == "F"]  

   # Commented out lines below when specific K is not used in subsequent buffer factor calculations
   #K1 <- K1(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)   
   #K2 <- K2(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)
    Kw <- Kw(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
   #K0 <- K0(S=SP, T=InsT, Patm=Patm, P=P, warn=warn)
    Kb <- Kb(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0, warn=warn)
   K1p <- K1p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
   K2p <- K2p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
   K3p <- K3p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
   Ksi <- Ksi(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
   #Kspa <- Kspa(S=SP, T=InsT, P=P, warn=warn)
   #Kspc <- Kspc(S=SP, T=InsT, P=P, warn=warn)

   #rho <- rho(S=SP,T=InsT,P=P)

   # Compute potential K0 with S, potential temperature, and atmospheric pressure (usually 1 atm)
   #K0pot <- K0(S=SP, T=theta(S=SP, T=InsT, P=P, Pref=0), Patm=Patm, P=0)

   #--------------------------------------------------------------------- 
   #--------------------    buffer effects    --------------------------- 
   #---------------------------------------------------------------------  
   #   	Buffer Factors from Egleston et al. (2010), Glob. Biogeochem. Cycles, 24, GB1002, doi:10.1029/2008GB003407  
   #       Std definitions needed to comput buffer factors 
   
   Alkc = (2*CO3 + HCO3) 	
   Borate  = (bor / (1 + h/Kb))  
   # could also use equivalent formulation of Borate = bor * Kb / (Kb + h) 	
   oh = Kw / h

   # Phosphorus inorganic species
   # [h2po4-] = K1p * [h3po4] / [H+]
   # [hpo4--] = K1p * K2p * [h3po4] / [H+]²
   # [po4---] = K1p * K2p * K3p * [h3po4] / [H+]³
   # Pt       = [h3po4] * (1 + K1p/[H+] + K1p*K2p/[H+]² + K1p*K2p*K3p/[H+]³)
   h3po4 = Pt / (1 + K1p/h + K1p*K2p/h^2 + K1p*K2p*K3p/h^3)
   h2po4 = K1p * h3po4 / h
   hpo4  = K2p * h2po4 / h
   po4   = K3p * hpo4  / h
   
   # Silicon inorganic species
   # [SiO(OH)3-] = Ksi * [Si(OH)4] / [H+]
   # Sit = [Si(OH)4] * (1 + Ksi / [H+])
   sioh4 = Sit /(1 + Ksi/h)
   # sioh3 = Ksi * sioh4/h
   sioh3 = Ksi * Sit/(h+Ksi)
      
   # Special definitions needed for buffer-factor calculations 
   #  - originally from Table 1 of Egleston et al;  
   #  - later modified to comply w/ formulas in Excel sheet of Chris Sabine (23 Aug 2010) 	
   # ------------------------------------------------------------------------------------
   # BAD formula in Table 1 (Egleston et al., 2010) - last sign is inversed
   # Segle   = (HCO3 + 4*CO3 + (h*Borate/(Kb + h)) + h - oh)  
   # GOOD formula with last sign above inverted (1st line of code just below),
   # ...  also added are effects from phosphoric and silicic acid systems (code lines 2-6 just below)
   #      The addition of these 2 acid systems from J.-M. Epitalon, 2016 (expanded equation from Egleston)
   SegleC = ( HCO3 + 4*CO3 + (h*Borate/(Kb + h)) + h + oh)

   numPt <-  - h3po4 * (-h2po4 - 2*hpo4 - 3*po4)
             + hpo4  * (2*h3po4 + h2po4 - po4)
             + 2*po4 * (3*h3po4 + 2*h2po4 + hpo4)
   # Protect against division by zero
   SegleP <- rep(0.0,n)
   SegleP[Pt > 0] <- numPt[Pt > 0] / Pt[Pt > 0]

   SegleSi =  h * sioh3/(Ksi + h)

   Segle <- SegleC + SegleP + SegleSi

   # GOOD formula from Sabine (Excel sheet, 23 Aug 2010)
   Pegle  = (2*CO2 + HCO3)                                  

   # GOOD formula from Table 1 (Egleston et al.)
   # Qegle  = (HCO3 - (h*Borate/(Kb + h)) - h - oh)           
   # GOOD formula from Sabine (Excel sheet, 23 Aug 2010)
   Qegle  = 2*Alkc - Segle                                 
   #Formula derived by J. Orr (same result as line just above)

   # #Compute 6 buffer factors: 
   #  *** NOTE - units of buffer factors (mol/kg) 
   #           - to convert to mmol/kg (units shown by Egleston), multiply each factor by 1000 after calling this routine 
   	
   gammaDIC = (DIC - (Alkc*Alkc)/Segle) 	
   gammaALK = ( (Alkc*Alkc - DIC*Segle) / Alkc ) 	
   betaDIC  = ( (DIC*Segle - Alkc*Alkc) / Alkc ) 	
   betaALK  = (Alkc*Alkc/DIC - Segle) 	

   # BAD Formula from Table 1 (Egleston et al.) - replace HCO3 by Qegle 	
   # omegaDIC = ( DIC - (Alkc*Pegle/HCO3) )                  
   # GOOD Formula from Sabine (Excel sheet, 23 Aug 2010) 	
   omegaDIC = ( DIC - (Alkc*Pegle/Qegle) )                  

   # BAD Formula from Table 1 (Egleston et al.) - replace HCO3 by Qegle 	
   # omegaALK = (Alkc - DIC*HCO3/Pegle)                       
   # GOOD Formula from Sabine (Excel sheet, 23 Aug 2010)  
   omegaALK = ( Alkc - DIC*Qegle/Pegle )                    

   # Revelle Factor (agrees exactly with that computed in seacarb's "buffer" function (BetaD, from Frankignoulle, 1994)  
   # when total dissolved inorganic phosphorus & silica concentrations are zero (Orr and Epitalon, 2015; Orr et al., 2015)
   # Now in this new version of 'buffesm', R accounts for these 2 acid systems.
   R = (DIC/gammaDIC)  	

   col <- c("gammaDIC", "betaDIC", "omegaDIC", "gammaALK", "betaALK", "omegaALK", "R") 	
   res <- data.frame(gammaDIC, betaDIC, omegaDIC, gammaALK, betaALK, omegaALK, R) 	
   names(res) <- col 
    
   return(res)
  }

# Copyright (C) 2003 Jean-Pierre Gattuso and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"bor" <-
function(S=35, b="u74"){  #three formulations possible Lee et al. (2010) : "l10"; Uppstrom (1974) : "u74"; "Kulinski et al. (2018)" : "k18"

n <- max(c(length(S), length(b)))
if(length(S)!=n){ S <- rep(S[1],n)}
if(length(b)!=n){ b <- rep(b[1],n)}
bor <- rep(NA, n)
method <- rep(NA, n)

# Lee et al. (2010)
  bor_l10 <- 0.1336*S #boron in mg/kg
  method <- "Lee et al. (2010)"

# Uppstrom (1974)
  bor_u74 <- 0.1284*S #boron in mg/kg
  method <- "Uppstrom (1974)"

# Kulinski et al. (2018)
  bor_k18 <- (11.405*S + 11.869)*10.811*10^-3 #boron in mg/kg
  method <- "Kulinski et al. (2018)"

#--------------- choice between the formulations -------------------
is_l10 <- (b=='l10')
is_u74 <- (b=='u74')
is_k18 <- (b=='k18')
bor[is_l10] <- bor_l10[is_l10]
bor[is_u74] <- bor_u74[is_u74]
bor[is_k18] <- bor_k18[is_k18]

method <- rep(NA, n)
method[is_l10] <- "Lee et al. (2010)"
method[is_u74] <- "Uppstrom (1974)"
method[is_k18] <- "Kulinski et al. (2018)"

# conversion from mg/kg to mol/kg
bor <- bor*10^(-3)/10.811
attr(bor,"unit") <- "mol/kg"
attr(bor, "method") <- method

return(bor)
}

# Copyright (C) 2008 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# sys is 0 if the system is closed and 1 if it is open
# co3 and hco3 are the   amount added in mol/kg


# One adds a certain volume (vol) of HCl (vol is then negative) or NaOH (vol is then positive) to 1 kg of seawater. 
# vol is in liter
# N is the normality of the HCl or NaOH
# HCl or NaOH are, respectively, a strong acid and a strong base which are therefore fully dissociated. The concentration of H+ or OH- is N mole/l
# HCl 1N: 0.1 ml therefore adds 0.1 e-3 mol of H+, decreasing TA
# NaOH 1N: 0.1 ml therefore adds 0.1 e-3 mol of OH-, increasing TA
# DIC is constant in a closed system
"ppH" <-
function(flag, sys, var1, var2, pCO2a, vol, N, S=35, T=20, P=0, Pt=0, Sit=0, pHscale="T", k1k2='x', kf='x', ks="d", eos="eos80", long=1.e20, lat=1.e20){
  # if the concentrations of total silicate and total phosphate are NA
  # they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  if (sys==0) {
		ci <- carb(flag=flag, var1=var1, var2=var2, S=S ,T=T, P=P, Pt=Pt, Sit=Sit,  pHscale=pHscale, k1k2=k1k2, kf=kf, ks=ks, eos=eos, long=long, lat=lat)
		alkf <- (ci$ALK + vol*N)/(1+abs(vol)) #final alk - dilution is taken into account
		dicf <- (ci$DIC)/(1+abs(vol))  #final dic - dilution is taken into account
		cf <- carb(flag=15, var1=alkf, var2=dicf, S=S, T=T, P=P,  Pt=Pt, Sit=Sit, pHscale=pHscale, k1k2=k1k2, kf=kf, ks=ks, eos=eos, long=long, lat=lat)
		co <- as.data.frame(c("ppH-closed-initial", rep("ppH-closed-final", nrow(cf)))) 
	}
	if (sys==1) {
		ci <- carb(flag=flag, var1=var1, var2=var2, S=S ,T=T, P=P, Pt=Pt, Sit=Sit,  pHscale=pHscale, k1k2=k1k2, kf=kf, ks=ks, eos=eos, long=long, lat=lat)
		alkf <- (ci$ALK + vol*N)/(1+abs(vol)) # final total alkalinity  - dilution is taken into account
		dicf <- (ci$DIC)/(1+abs(vol))	# final dic  before requilibration - dilution is taken into account
		#pHf <- ci$pH + ci$PhiH * (-vol) *N	# final pH using a buffer factor (see Frankignoulle, 1994)
		cc <- carb(flag=15, var1=alkf, var2=dicf, S=S, T=T, P=P, Pt=Pt, Sit=Sit,  pHscale=pHscale, k1k2=k1k2, kf=kf, ks=ks, eos=eos, long=long, lat=lat) #  before requilibration
		cf <- carb(flag=24,var1=pCO2a, var2=alkf, S=S, T=T, P=P, Pt=Pt, Sit=Sit,  pHscale=pHscale, k1k2=k1k2, kf=kf, ks=ks, eos=eos, long=long, lat=lat)
		co <- as.data.frame(c("ppH-open-initial", rep("ppH-open-final", nrow(cf))))
	}
	out <- rbind(ci, cf)
	out <- cbind(co, out)
	names(out)[1] <- "comment"
	return(out)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye
# Revised by James Orr, 2012-01-17
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"Kf" <-
function(S=35,T=25,P=0,kf='x',pHscale="T",Ks_p0="x",Ks_p="x",warn="y"){

    nK <- max(length(S), length(T), length(P), length(kf), length(pHscale), length(Ks_p0), length(Ks_p))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(kf)!=nK){kf <- rep(kf[1], nK)}
    if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}
    
##----------Make kf a global variable to facilitate pH scale changes in multiple routines

    if (missing(kf)) {
        if (exists("kfg",envir=parent.frame()) )
           {kf <- get("kfg", envir = parent.frame()) }
	else    
           {assign("kfg", "x", envir = parent.frame()) }
	}
    else
        {assign("kfg", kf, envir = parent.frame()) } 

    ##----------Check the validity of the method regarding the T/S range
    is_x <- kf=='x'
    is_outrange <- T>33 | T<10 | S<10 | S>40
    kf[is_x] <- "pf"  ## Perez and Fraga by default
    kf [is_x & is_outrange] <- "dg"
    
    #-------Constantes----------------

    #---- issues de equic----
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK = T + tk;           # TC [C]; T[K]
    Cl = S / 1.80655;      # Cl = chlorinity; S = salinity (per mille)
    iom0 = 19.924*S/(1000-1.005*S);
    ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
    FT = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)

    #---------------------------------------------------------------------
    #---------------------- Kf Perez and Fraga ---------------------------
    #  Kf = [H+][F-]/[HF]  
    #
    #   Perez and Fraga, 1987 in Guide to the Best Practices for Ocean CO2 Measurements
    #   Dickson, Sabine and Christian, 2007, Chapter 5, p. 14)
    #  
    #   pH-scale: 'total'   

    lnKfpf <- 874/TK - 9.68 + 0.111*S^(1/2)
    Kfpf <- exp(lnKfpf)

    # --------------- Conversion to free scale for pressure corrections
    
    # if Ks at zero pressure is not given
    if (missing(Ks_p0) || Ks_p0[1] == "x")
        # Ks at zero pressure NOT given --> compute it
    	Ks = Ks(S=S, T=T, P=rep(0,nK))                 # on free pH scale
    else
    {
    	Ks <- Ks_p0
    	if (length(Ks)!=nK) Ks <- rep(Ks[1], nK)
    }
    
    Cl = S / 1.80655;             # Cl = chlorinity; S = salinity (per mille)
    ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
    total2free  = 1/(1+ST/Ks)      # Kfree = Ktotal*total2free
    total2free <- as.numeric(total2free)	

    factor <- total2free

    Kfpf <- Kfpf * factor

    #---------------------------------------------------------------------
    # --------------------- Kf Dickson and Goyet -------------------------
    #  Kf = [H+][F-]/[HF]  
    #
    #   (Dickson and Riley, 1979 in Dickson and Goyet, 
    #   1994, Chapter 5, p. 14)
    #   pH-scale: 'free' (require to convert in total scale after pressure corrections 

    lnKfdg = 1590.2/TK - 12.641 + 1.525*sqrt(iom0) + log(1-0.001005*S);

    Kfdg <- exp(lnKfdg)

    # ---------- Choice between methods (Perez and Dickson) ----------

    is_pf <- (kf!='dg')    # everything that is not "dg" is "pf" by default
    Kf <- Kfpf
    Kf[!is_pf] <- Kfdg[!is_pf]


    # ------------------- Pressure effect --------------------------------
    if (any (P != 0)) Kf <- Pcorrect(Kvalue=Kf, Ktype="Kf", T=T, S=S, P=P, pHscale="F", warn=warn) 
    # All the Kf constants are on free scale, whatever the method or the pressure

    # Which pH scales are required ?
    is_total <- pHscale=="T"
    is_SWS   <- pHscale=="SWS"

    # pH scale correction factor
    factor <- rep(1.0,nK)

    # if any pH scale corrections required (from free scale)
    if (any(is_total) || any(is_SWS))
    {
        # If Ks at given pressure is not given 
        if (missing(Ks_p) || Ks_p[1] == "x")
            # compute Ks
            Ks = Ks(S=S, T=T, P=P)                 # on free pH scale
        else
        {
        	Ks <- Ks_p
        	if (length(Ks)!=nK) Ks <- rep(Ks[1], nK)
        }
        # Compute pH scale correction factor
        factor[is_total] <- 1 + ST[is_total]/Ks[is_total]
        factor[is_SWS]   <- 1 + ST[is_SWS]/Ks[is_SWS] + FT[is_SWS]/Kf[is_SWS]
    }
    # Perform pH scale correction
    Kf <- Kf*factor

    # Return full name of pH scale
    pHsc <- rep(NA,nK)
    pHsc[is_total] <- "total scale"
    pHsc[is_SWS]   <- "seawater scale"
    pHsc[!is_total & !is_SWS] <- "free scale"


    ##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & is_pf & (T>33 | T<9 | S<10 | S>40)) ) {warning("S and/or T is outside the range of validity of the formulation chosen for Kf.")}
    if (any(is_w & ( T>45 | S>45 ))) {warning("S and/or T is outside the range of validity of the formulations available for Kf in seacarb.")}

    ##---------------Attributes
    method <- rep(NA,nK)
    method[ is_pf] <- "Perez and Fraga (1987)"
    method[!is_pf] <- "Dickson and Riley (1979 in Dickson and Goyet, 1994)"


    attr(Kf,"unit")     = "mol/kg-soln"
    attr(Kf,"pH scale") = pHsc
    attr(Kf, "method") = method
    return(Kf)
}

# Copyright (C) 2008 Jean-Marie Epitalon and Jean-Pierre Gattuso and Heloise Lavigne and Aurelien Proye
# with a most valuable contribution of Bernard Gentili <gentili@obs-vlfr.fr>
# and valuable suggestions from Jean-Marie Epitalon <epitalon@lsce.saclay.cea.fr>
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#

buffer <- 
  function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", warn="y", eos="eos80", long=1.e20, lat=1.e20){
    n <- max(length(flag), length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), length(ks), length(b))
    if(length(flag)!=n){flag <- rep(flag[1],n)}
    if(length(var1)!=n){var1 <- rep(var1[1],n)}
    if(length(var2)!=n){var2 <- rep(var2[1],n)}
    if(length(S)!=n){S <- rep(S[1],n)}
    if(length(T)!=n){T <- rep(T[1],n)}
    if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
    if(length(P)!=n){P <- rep(P[1],n)}
    if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
    if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
    if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
    if(length(kf)!=n){kf <- rep(kf[1],n)}
    if(length(ks)!=n){ks <- rep(ks[1],n)}
    if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
    if(length(b)!=n){b <- rep(b[1],n)}

    # if the concentrations of total silicate and total phosphate are NA
    # they are set at 0
    Sit[is.na(Sit)] <- 0
    Pt[is.na(Pt)] <- 0
    
    # Only two options for eos
    if (eos != "teos10" && eos != "eos80")
        stop ("invalid parameter eos: ", eos)
    
    # if use of EOS-10 standard
    if (eos == "teos10")
    {
        # Must convert temperature and salinity from TEOS-10 to EOS-80
        # convert temperature: from Conservative (CT) to in-situ temperature
        # and salinity from Absolute to Practical (SP)
        STeos <- teos2eos_geo (S, T, P, long, lat)
        InsT <- STeos$T
        SP <- STeos$SP
    }
    else
    {
        InsT <- T
        SP <- S
    }

    Carb <- carb(flag=flag, var1=var1, var2=var2, S=SP, T=InsT, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b)
    
	pH   <- Carb$pH
	h    <- 10^(-pH)
	CO2  <- Carb$CO2
	HCO3 <- Carb$HCO3
	CO3  <- Carb$CO3
	DIC  <- Carb$DIC
	ALK  <- Carb$ALK
	Oa   <- Carb$OmegaAragonite
	Oc   <- Carb$OmegaCalcite
    
    #-------Constants----------------
    
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK = InsT + tk;           # TK [K]; InsT[C]
    
    Cl = SP / 1.80655;            # Cl = chlorinity; SP = practical salinity (psu)
    ST = 0.14 * Cl/96.062        # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
    FLUO = 6.7e-5 * Cl/18.9984   # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
    BOR = bor(S=SP , b=b);        # (mol/kg) total boron

    #---------------------------------------------------------------------
    #--------------------- compute K's ----------------------------------
    #---------------------------------------------------------------------
    
    # Ks (free pH scale) at zero pressure and given pressure
    Ks_P0 <- Ks(S=SP, T=InsT, P=0, ks=ks, warn=warn)
    Ks    <- Ks(S=SP, T=InsT, P=P, ks=ks, warn=warn)
    
    # Kf on free pH scale
    Kff_P0 <- Kf(S=SP, T=InsT, P=0, pHscale="F", kf=kf, Ks_P0, Ks)
    Kff <- Kf(S=SP, T=InsT, P=P, pHscale="F", kf=kf, Ks_P0, Ks)
    # Kf on given pH scale
    Kf <- Kf(S=SP, T=InsT, P=P, pHscale=pHscale, kf=kf, Ks_P0, Ks)
    
    # Conversion factor from total to SWS pH scale at zero pressure
    ktotal2SWS_P0 <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks_P0,Kff=Kff_P0)$ktotal2SWS

    # Conversion factor from SWS to chosen pH scale
    conv <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks,Kff=Kff)
    kSWS2chosen <- rep(1.,n)
    kSWS2chosen [pHscale == "T"] <- conv$kSWS2total [pHscale == "T"]
    kSWS2chosen [pHscale == "F"] <- conv$kSWS2free [pHscale == "F"]  

    K1 <- K1(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)   
    K2 <- K2(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)
    Kw <- Kw(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    K0 <- K0(S=SP, T=InsT, Patm=Patm, P=P, warn=warn)
    Kb <- Kb(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0, warn=warn)
    K1p <- K1p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    K2p <- K2p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    K3p <- K3p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    Ksi <- Ksi(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    Kspa <- Kspa(S=SP, T=InsT, P=P, warn=warn)
    Kspc <- Kspc(S=SP, T=InsT, P=P, warn=warn)

    rho <- rho(S=SP,T=InsT,P=P)

    # Compute potential K0 with S, potential temperature, and atmospheric pressure (usually 1 atm)
    K0pot <- K0(S=SP, T=theta(S=SP, T=InsT, P=P, Pref=0), Patm=Patm, P=0, warn=warn)
    
    #---------------------------------------------------------------------
    #--------------------    buffer effects    ---------------------------
    #---------------------------------------------------------------------
    
    DD=-((-Kb*BOR)/((h+Kb)*(h+Kb)))-(-Kw/((h)*(h)))+1;
    A= (2*K2*(2*CO3+HCO3)+h*(h+2*K2)*DD)/((h+2*K2)*(h+2*K2));
    B=( ( (2*CO3+HCO3) * h)/((h+2*K2)*K1) + (h/K1)* A );
    C= (-K2*(2*CO3+HCO3)+K2*(2*K2+h)*DD)/((h+2*K2)*(h+2*K2));
    PhiD=-1/(h*log(10) * ( B+A+C ) );
    BetaD=-h*log(10)*DIC/CO2*B*PhiD;
    
    
    Q=(h+2*K2);
    V=(Kb*BOR)/((h+Kb)*(h+Kb)) + Kw/(h*h)+1;
    
    DB=(( K2*(2*CO3+HCO3)+ Q*V *(h+K2)+(h/K1)*( (2*CO3+HCO3)*Q+2*K2*(2*CO3+HCO3)+h*Q*V))/Q)*(1/(Q-(h+K2+h*h/K1)))-((-Kb*BOR)/((h+Kb)*(h+Kb)))-(-Kw/((h)*(h)))+1;
    A= (2*K2*(2*CO3+HCO3)+h*(h+2*K2)*DB)/((h+2*K2)*(h+2*K2));
    B=( ( (2*CO3+HCO3) * h)/((h+2*K2)*K1) + (h/K1)* A );
    C= (-K2*(2*CO3+HCO3)+K2*(2*K2+h)*DB)/((h+2*K2)*(h+2*K2));
    PhiB=-1/(h*log(10) * ( B+A+C ) );
    BetaB=-h*log(10)*DIC/CO2*B*PhiB;
    
    DC=2*(( K2*(2*CO3+HCO3)+ Q*V *(h+K2)+(h/K1)*( (2*CO3+HCO3)*Q+2*K2*(2*CO3+HCO3)+h*Q*V))/Q)*(1/(Q-2*(h+K2+h*h/K1)))-((-Kb*BOR)/((h+Kb)*(h+Kb)))-(-Kw/((h)*(h)))+1;
    A= (2*K2*(2*CO3+HCO3)+h*(h+2*K2)*DC)/((h+2*K2)*(h+2*K2));
    B=( ( (2*CO3+HCO3) * h)/((h+2*K2)*K1) + (h/K1)* A );
    C= (-K2*(2*CO3+HCO3)+K2*(2*K2+h)*DC)/((h+2*K2)*(h+2*K2));
    PhiC=-1/(h*log(10) * ( B+A+C ) );
    BetaC=-h*log(10)*DIC/CO2*B*PhiC;
    
    D1=(K1*(K1*K2-h*h)*DIC)   /  ((h*h+h*K1+K1*K2)*(h*h+h*K1+K1*K2));
    D2=(-K1*K2*(2*h+K1)*DIC)  /  ((h*h+h*K1+K1*K2)*(h*h+h*K1+K1*K2));
    D=D1+2*D2;
    PhiH=1/ (h*log(10)* (D +(-Kb*BOR/((h+Kb)*(h+Kb)))  + (-Kw/(h*h))-1))  ; 
    
    Pi=(h*K1*(h+2*K2)*DIC)  /  ((h*h+h*K1+K1*K2)*(h*h+h*K1+K1*K2));

    # In situ values - use K0 computed with in situ T & in situ P (atm + hydrostatic)
    # PiH=((-h/K0)*log(10)*Pi)*PhiH;
    # PiB=CO2/(K0*DIC)*BetaB;
    # PiD=CO2/(K0*DIC)*BetaD;
    # PiC=CO2/(K0*DIC)*BetaC;

    # Potential values - use K0 computed with potential T & surface atm P only
    PiH=((-h/K0pot)*log(10)*Pi)*PhiH;
    PiB=CO2/(K0pot*DIC)*BetaB;
    PiD=CO2/(K0pot*DIC)*BetaD;
    PiC=CO2/(K0pot*DIC)*BetaC;

    col <- c("PhiD", "BetaD", "PiD", "PhiB", "BetaB", "PiB", "PhiC", "BetaC", "PiC", "PhiH", "PiH")
    res <- data.frame(PhiD,BetaD,PiD,PhiB,BetaB,PiB,PhiC,BetaC,PiC,PhiH,PiH)
    names(res) <- col 
    
    return(res)
  }

# Copyright (C) 2007 Karline Soetaert (K.Soetaert@nioo.knaw.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


"speciation" <- function(K1=K1(),K2=NULL,K3=NULL, # the dissociation constants
                         pH=8,               # the pH at which speciation is wanted 
                         conc=1)             # the concentration, mol/kg

{
univalent <- function(K1,H)
{ # concentration for univalent species (e.g. NH3,NH4+)
 C1<- H/(K1+H) # modificated by He\'loi\¨se Lavigne (C1=K1/(K1+H)-> C1=  H/(K1+H)) 2008
 C2<-1-C1
 return(list(C1=C1*conc,C2=C2*conc))
}

bivalent <- function(K1,K2,H)
{ # concentration for bivalent species(e.g. CO2,HCO3-,CO3--)
 den <-H*H + H*K1 + K1*K2
 C1  <-H*H/den
 C2  <-H*K1/den
 C3  <- 1-C1-C2
 return(list(C1=C1*conc,C2=C2*conc,C3=C3*conc))
}

trivalent <- function(K1,K2,K3,H)
{ # concentration for trivalent species(e.g. H3PO4,H2PO4-,HPO4--,PO4---)
 den <-H*H*H + H*H*K1 + H*K1*K2 + K1*K2*K3
 C1  <-H*H*H/den
 C2  <-H*H*K1/den
 C3  <-H*K1*K2/den
 C4  <- 1-C1-C2-C3
 return(list(C1=C1*conc,C2=C2*conc,C3=C3*conc,C4=C4*conc))
}

# estimates the speciation of the various ionic forms of a molecule as a function of pH 
 if (is.null(K1)) return()
 H   <- 10^(-pH)                  # proton concentration
 if (is.null(K2))      res<-univalent(as.double(K1),H)        
 else if (is.null(K3)) res<-bivalent (as.double(K1),as.double(K2),H)     
 else                  res<-trivalent(as.double(K1),as.double(K2),as.double(K3),H)  
 return(res)
}

# Copyright (C) 2021  Jean-Pierre Gattuso
# with a most valuable contribution of Bernard Gentili <gentili@obs-vlfr.fr>
# and valuable suggestions from Jean-Marie Epitalon <epitalon@lsce.saclay.cea.fr>
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
fCO2insi <- function(fCO2lab = 400, Tlab = 20, SST = 19){
  #Takahashi (1993) recommended by Pierrot et al. (2009)
  fCO2_insi <- fCO2lab * exp(0.0423 * (SST - Tlab))
return(fCO2_insi)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Heloise Lavigne
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"pHconv" <- function (flag=1, pH=8.100, S=35, T=25, P=0, ks="d")

{

n <- max(length(pH), length(flag), length(S), length(T), length(S), length(P), length(ks))

if(length(pH)!=n){pH <- rep(pH[1],n)}
if(length(flag)!=n){flag <- rep(flag[1],n)}
if(length(S)!=n){S <- rep(S[1],n)}
if(length(T)!=n){T <- rep(T[1],n)}
if(length(P)!=n){P <- rep(P[1],n)}
if(length(ks)!=n){ks <- rep(ks[1],n)}

pHconv <- rep(NA, n)
conv <- rep(NA, n)

for(i in (1:n)){  

	Ks = Ks(S=S[i], T=T[i], P=P[i], ks=ks[i])[1]                 # on free pH scale
        Cl = S[i] / 1.80655;      # Cl = chlorinity; S = salinity (per mille)
        ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
	Kf = Kf(S=S[i], T=T[i], P=P[i], pHscale="F")[1]  # on the free pH scale
        FT = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)

# flag 1 : seawater to total
if (flag[i]==1) {pHconv[i] <- pH[i] + log10(1 + ST/Ks + FT/Kf) - log10(1 + ST/Ks); conv[i] <- "seawater to total"} 
# flag 2 : free to total 
if (flag[i]==2) {pHconv[i] <- pH[i] - log10(1 + ST/Ks); conv[i] <- "free to total"} 
# flag 3 :  total to seawater
if (flag[i]==3) {pHconv[i] <- pH[i] - log10(1 + ST/Ks + FT/Kf) + log10(1 + ST/Ks); conv[i] <- "total to seawater"}
# flag 4 : total to free 
if (flag[i]==4) {pHconv[i] <- pH[i] + log10(1 + ST/Ks); conv[i] <- "total to free"} 
# flag 5 : seawater to free
if (flag[i]==5) {pHconv[i] <- pH[i] + log10(1 + ST/Ks + FT/Kf); conv[i] <- "seawater to free"}
# flag 6 : free to seawater
if (flag[i]==6) {pHconv[i] <- pH[i] - log10(1 + ST/Ks + FT/Kf); conv[i] <- "free to seater"}
}

attr(pHconv, "conversion") <- conv
return(pHconv)

}

# Copyright (C) 2014 James Orr
# This file is part of seacarb.
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Compute vapor pressure of seawater (atm) following preocedure from either Weiss and Price (1980) or Dickson et al. (2007)
# The latter also computed the vapor pressure of pure water
# Both should give the same result when rounded to the 4th digit after the decimal


vapress <- 
function(S=35, T=25, form="d2007"){

#RES <- data.frame()

#-------Constants----------------

  tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
  TK = T + tk;           # TK [K]; T[C]

  if (form=="wp1980") {
#   Formula from Weiss & Prince (1980, equation 10)
#   ln pH20 = 24.4543 - 67.4509 (100/T) - 4.8489 ln (T/100) - 0.000544 S

    vpsw <- exp(24.4543 - 67.4509*(100/TK) - 4.8489*log(TK/100) - 0.000544*S)
#   col <- c("VPseawater")
#    RES <- data.frame(vpsw)
    RES <- vpsw
  } 
  else if(form=="d2007") {

#   Compute vapor pressure of water and seawater following recommendations in best-practices guide (Dickson et al., 2007)
#   Constants to compute water & seawater vapor pressure [kPa]: Dickson et al. (2007) Best Practices Guide
    cv1 =  -7.85951783 
    cv2 =   1.84408259
    cv3 = -11.7866497
    cv4 =  22.6807411
    cv5 = -15.9618719
    cv6 =   1.80122502 
    co0 =   0.90799   
    co1 =  -0.08992  
    co2 =   0.18458 
    co3 =  -0.07395
    co4 =  -0.00221

#   Vapor pressure of water
    ztc = 647.096
    zpc = 22.064 / 101325.0e-6     #Convert 22.064 MPa to atmospheres)
  
    zrt = 1 - TK/ztc
    zrt05 = sqrt(zrt)
    zrt15 = zrt*zrt05
    zrt2  = zrt*zrt
    zrt3  = zrt*zrt2
    zrt35 = zrt3*zrt05
    zrt4  = zrt2*zrt2
    zrt75 = zrt4*zrt35
#   vpwat  = exp( -LOG(ztc) + (ztc/TK)* (cv1*zrt + cv2*zrt15 + cv3*zrt3 + cv4*zrt35 + cv5*zrt4 + cv6*zrt75) )
    vpwat  = zpc * exp((ztc/TK)* (cv1*zrt + cv2*zrt15 + cv3*zrt3 + cv4*zrt35 + cv5*zrt4 + cv6*zrt75) )

#   Vapor pressure of seawater
    zsumb = 31.998 * S / ( 1.0E+03 - 1.005*S )
    zsmh  = zsumb * 0.5
    zsmh2 = zsmh*zsmh
    zsmh3 = zsmh*zsmh2
    zsmh4 = zsmh2*zsmh2
    zosmo = co0 + co1*zsmh + co2*zsmh2 + co3*zsmh3 + co4*zsmh4
#   vpsw_kpa    = vpwat * exp(-0.018 * zosmo * zsumb)      # Vapor pressure for seawater [kPa] - NO (already in atm)
#   vpsw =  vpsw_kpa / 101.325                             # Vapor pressure for seawater [atm] - unnecessary
    vpsw = vpwat * exp(-0.018 * zosmo * zsumb)             # Vapor pressure for seawater [atm]

#   col <- c("VPwater", "VPseawater")
#   RES <- data.frame(vpwat, vpsw)
#    RES <- data.frame(vpsw)
    RES <- vpsw 
  }

return(RES)
}

# Copyright (C) 2007 Karline Soetaert (K.Soetaert@nioo.knaw.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


"bjerrum" <- function(K1=K1(),K2=NULL,K3=NULL,      # the dissociation constants
                    phmin=2, phmax=12, by=0.1, # the pH range and increment
                    conc=1,                    # the concentration, mol / kg
                    type="l",col="black",    # overruled default plotting options
                    ylab="Relative concentration (%)",
                    add=FALSE,               # false:start new, true: add to current
                    ...)                     # plotting options passed to matplot

{

# creates a bjerrum plot
 if (is.null(K1)) return()

 pH  <- seq (phmin,phmax,by=by)   # gradient of pH
 res <- speciation(K1, K2, K3, pH, conc)

 if (!add) matplot(pH, as.data.frame(res), ylab=ylab, type=type, col=col, ...)
 else matlines(pH, as.data.frame(res), type=type, col=col, ...)
}

# Copyright (C) 2008 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"pTA" <-
function(flag, sys=0, var1, var2, pCO2a, co3, hco3, S=35, T=20, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", eos="eos80", long=1.e20, lat=1.e20){
  # if the concentrations of total silicate and total phosphate are NA
  # they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  
  if (sys==0) {
		ci <- carb(flag=flag, var1=var1, var2=var2, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
		alkf <- ci$ALK + 2*co3 + hco3 # final alkalinity
		dicf <- ci$DIC + co3 + hco3	# final dic
		cf <- carb(flag=15, var1=alkf, var2=dicf, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
		co <- as.data.frame(c("pTA-closed-initial", rep("pTA-closed-final", nrow(cf))))
	}
	if (sys==1) {
		ci <- carb(flag=flag, var1=var1, var2=var2, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks,  pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
		alkc <- ci$ALK + 2*co3 + hco3 # total alkalinity before requilibration
		dicc <- ci$DIC + co3 + hco3	# dic before requilibration
		cc <- carb(flag=15, var1=alkc, var2=dicc, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
		alkf <- cc$ALK	# final total alkalinity
		cf <- carb(flag=24, var1=pCO2a, var2=alkf, S=S ,T=T, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
		co <- as.data.frame(c("pTA-open-initial", rep("pTA-open-final", nrow(cf))))
	}
out <- rbind(ci, cf)
out <- cbind(co, out)
names(out)[1] <- "comment"
return(out)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"K1" <-
function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale="x",ktotal2SWS_P0="x",warn="y")
{

    ##---------- check k1k2 parameter
    k1k2_valid_options <- c("cw", "l", "m02", "m06", "m10", "mp2", "p18", "r", "s20", "sb21", "w14", "x")
    test_valid <- is.element(k1k2, k1k2_valid_options)
    
    if (! all (test_valid))
    {
        # If k1k2 is a vector, find indexes of all invalid values in it
        invalids <- which(test_valid == FALSE)
        # Display the first invalid value
        stop ("Invalid parameter k1k2 = ", k1k2[invalids[1]])
    }
    
    ##-------- Create vectors for all input (if vectorial)

    nK <- max(length(S), length(T), length(P), length(k1k2), length(pHscale), length(kSWS2scale) ,length(ktotal2SWS_P0))

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(k1k2)!=nK){k1k2 <- rep(k1k2[1], nK)}
    if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}
    
    ##---------- pHsc : this vector is not the actual pHscale because it can change during processing
    pHsc <- rep(NA,nK)
    pHlabel <- rep(NA,nK)
    K1 <- rep(NA, nK)

    ##----------Check the validity of the method regarding the T/S range
    is_x <- k1k2 == 'x'
    is_outrange <- T>35 | T<2 | S<19 | S>43
    k1k2[is_x] <- 'l'  ## luecker by default
    k1k2[is_x & is_outrange] <- "w14"  # Waters et al. (2014) if outrange

    #-------Constants----------------

    #---- issues of equic----
    tk <- 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK <- T + tk;           # TC [C]; T[K]
    
    
    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #     Mehrbach et al (1973) refit by Lueker et al. (2000).
    #
    #(Lueker  et al., 2000 in Guide to the Best Practices for Ocean CO2 Measurements
    #   Dickson, Sabine and Christian , 2007, Chapter 5, p. 13)
    #
    #   pH-scale: 'total'. mol/kg-soln
    is_l <- k1k2 == "l"
    logK1lue <- (-3633.86)/TK[is_l] + 61.2172 - 9.67770*log(TK[is_l]) + 0.011555*S[is_l] - 0.0001152*S[is_l]*S[is_l]
    K1[is_l]<- 10^logK1lue
    pHsc[is_l] <- "T"


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #     Papadimitriou et al (2018)  in Geochimica et Cosmochimica Acta 220 (2018) 55–70
    #
    #   pH-scale: 'total'. mol/kg-soln
    is_p18 <- k1k2 == "p18"
    Sp18 <- S[is_p18]
    pK1 <- -176.48 + 6.14528*sqrt(Sp18) - 0.127714*Sp18 + 7.396e-5*Sp18*Sp18 +
            (9914.37 - 622.886*sqrt(Sp18) + 29.714*Sp18)/TK[is_p18] +
            (26.05129 - 0.666812*sqrt(Sp18)) * log(TK[is_p18])
    K1[is_p18]<- 10^(-pK1)
    pHsc[is_p18] <- "T"


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #     Sulpis et al (2020)  in Ocean Science, 16, 847–862, 2020
    #
    #   pH-scale: 'total'. mol/kg-soln
    is_s20 <- k1k2 == "s20"
    pK1 <- 8510.63/TK[is_s20] - 172.4493 + 26.32996*log(TK[is_s20]) - 0.011555*S[is_s20] + 0.0001152*S[is_s20]*S[is_s20]
    K1[is_s20]<- 10^(-pK1)
    pHsc[is_s20] <- "T"


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   (Roy et al., 1993 in Dickson and Goyet, 1994, Chapter 5, p. 14)
    #   pH-scale: 'total'. mol/kg-soln
    is_r <- k1k2 == "r"
    tmp1  <- 2.83655 - 2307.1266 / TK[is_r] - 1.5529413 * log(TK[is_r])
    tmp2  <- - (0.20760841 + 4.0484 / TK[is_r]) * sqrt(S[is_r])
    tmp3 <- 0.08468345 * S[is_r] - 0.00654208 * S[is_r] * sqrt(S[is_r])   
    tmp4 <- log(1 - 0.001005 * S[is_r])
    lnK1roy <- tmp1 + tmp2 + tmp3 + tmp4
    K1[is_r] <- exp(lnK1roy)
    pHsc[is_r] <- "T"


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   Cai and Wang (1998)
    #
    #   pH-scale: 'seawater'. mol/kg-soln  (not true, starts on NBS scale, need to convert to SWS scale)
    is_cw <- k1k2 == "cw"
    F1 <- 200.1/TK[is_cw] + 0.3220
    pK1 <- 3404.71/TK[is_cw] + 0.032786*TK[is_cw] - 14.8435 - 0.071692*F1*S[is_cw]^0.5 + 0.0021487*S[is_cw] #*F1
    K1[is_cw]  <- 10^(-pK1)         # this is on the NBS scale
    # Convert from NBS to SWS scale using combined activity coefficient fH 
    # Takahashi (1982, GEOSECS Pacific Expedition, Chap 3, p. 80), who say:
    # "fH is the total activity coeff., which includes contributions from HSO4- and HF [as well as H+].
    #  Culberson and Pytkowicz (28) determined fH as a function of temperature and salinity, and
    #  their results can be approximated by:"
    fH = (1.2948 - 0.002036*TK[is_cw] + (0.0004607 - 0.000001475*TK[is_cw])*S[is_cw]^2)
    K1[is_cw] <- K1[is_cw]/fH[is_cw]  # Convert from NBS to SWS scale
    pHsc[is_cw] <- "SWS"
    # CO2SYS-MATLAB and PyCO2SYS mention that this NBS -> SWS conversion is uncertain at low S due to junction potential
    # Culberson, C. H., & Pytkowicz, R. M. (1973). Ionization of water in seawater. Marine Chemistry, 1(4), 309-316.


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   Mojica Prieto and Millero (2002)
    #
    #   pH-scale: 'seawater'. mol/kg-soln
    is_mp2 <- k1k2 == "mp2"
    pK1 <- -43.6977 - 0.0129037 * S[is_mp2] + 1.364e-4 * S[is_mp2]*S[is_mp2] + 2885.378/TK[is_mp2] + 7.045159 * log(TK[is_mp2])
    K1[is_mp2]<- 10^(-pK1)
    pHsc[is_mp2] <- "SWS"

    
    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    # Millero et al., 2002. Deep-Sea Res. I (49) 1705-1723.
    # Calculated from overdetermined WOCE-era field measurements
    # This is from page 1715
    #
    #   pH-scale: 'seawater'. mol/kg-soln
    is_m02 <- k1k2 == "m02"
    pK1 =  6.359 - 0.00664 * S[is_m02] - 0.01322 * T[is_m02] + 4.989e-5 * T[is_m02]*T[is_m02]
    K1[is_m02]<- 10^(-pK1)
    pHsc[is_m02] <- "SWS"
    
        
    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   Millero et al. 2006 Marine Chemistry
    #   pH-scale: 'SWS scale'. mol/kg-soln
    is_m06 <- k1k2 == "m06"
    pK1o <- 6320.813/TK[is_m06] + 19.568224*log(TK[is_m06]) -126.34048
    A1 <- 13.4191*S[is_m06]^(0.5) + 0.0331*S[is_m06] - (5.33e-5)*S[is_m06]^2
    B1 <- -530.123*S[is_m06]^(0.5) - 6.103*S[is_m06]
    C1 <- -2.06950*S[is_m06]^(0.5)
    pK1 <- pK1o + A1 + B1/TK[is_m06] + C1*log(TK[is_m06])
    K1[is_m06] <- 10^(-pK1)
    pHsc[is_m06] <- "SWS"


    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   Millero 2010 Marine and Fresh water research

    is_m10 <- k1k2 == "m10"
    
    #   pH-scale: 'SWS scale'. mol/kg-soln
    is_m10_SWS <- is_m10 & (pHscale=="SWS" | P>0)
    A1 <- 13.4038*S[is_m10_SWS]^(0.5) + 0.03206*S[is_m10_SWS] - (5.242e-5)*S[is_m10_SWS]^2
    B1 <- -530.659*S[is_m10_SWS]^(0.5) - 5.8210*S[is_m10_SWS]
    C1 <- -2.0664*S[is_m10_SWS]^(0.5)
    pK1o <- 6320.813/TK[is_m10_SWS] + 19.568224*log(TK[is_m10_SWS]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_m10_SWS] + C1*log(TK[is_m10_SWS])
    K1[is_m10_SWS] <- 10^(-pK1)   # K1 according to Millero et al. 2010 at Seawater scale
    pHsc[is_m10_SWS] <- "SWS" 

    #   pH-scale: 'Total scale'. mol/kg-soln
    is_m10_T <- is_m10 & (pHscale=="T" & P==0)
    A1 <- 13.4051*S[is_m10_T]^(0.5) + 0.03185*S[is_m10_T] - (5.218e-5)*S[is_m10_T]^2
    B1 <- -531.095*S[is_m10_T]^(0.5) - 5.7789*S[is_m10_T]
    C1 <- -2.0663*S[is_m10_T]^(0.5)
    pK1o <- 6320.813/TK[is_m10_T] + 19.568224*log(TK[is_m10_T]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_m10_T] + C1*log(TK[is_m10_T])
    K1[is_m10_T] <- 10^(-pK1)   # K1 according to Millero et al. 2010 at Total scale 
    pHsc[is_m10_T] <- "T"

    #   pH-scale: 'Free scale'. mol/kg-soln
    is_m10_F <- is_m10 & (pHscale=="F" & P==0)
    A1 <- 5.09247*S[is_m10_F]^(0.5) + 0.05574*S[is_m10_F] - (9.279e-5)*S[is_m10_F]^2
    B1 <- -189.879*S[is_m10_F]^(0.5) - 11.3108*S[is_m10_F]
    C1 <- -0.8080*S[is_m10_F]^(0.5)
    pK1o <- 6320.813/TK[is_m10_F] + 19.568224*log(TK[is_m10_F]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_m10_F] + C1*log(TK[is_m10_F])
    K1[is_m10_F] <- 10^(-pK1)  # K1 according to Millero et al. 2010 at Total scale
    pHsc[is_m10_F] <- "F"

    # --------------------- K1 ---------------------------------------
    #   first acidity constant:
    #   [H^+] [HCO_3^-] / [CO2] = K_1
    #
    #   Waters, Millero, Woosley (Mar. Chem., 165, 66-67, 2014)

    # Option k1k2= "sb21" refers to Shockman & Byrne (2021) for K2  and Waters et al.(2014) for K1
    is_w14 <- (k1k2 == "w14") | (k1k2 == "sb21")

    #   pH-scale: 'SWS scale'. mol/kg-soln
    is_w14_SWS <- is_w14 & (pHscale=="SWS" | P>0)
    A1 <- 13.409160*S[is_w14_SWS]^(0.5) + 0.031646*S[is_w14_SWS] - (5.1895e-5)*S[is_w14_SWS]^2
    B1 <- -531.3642*S[is_w14_SWS]^(0.5) - 5.713*S[is_w14_SWS]
    C1 <- -2.0669166*S[is_w14_SWS]^(0.5)
    pK1o <- 6320.813/TK[is_w14_SWS] + 19.568224*log(TK[is_w14_SWS]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_w14_SWS] + C1*log(TK[is_w14_SWS])
    K1[is_w14_SWS] <- 10^(-pK1)   # K1 according to Millero et al. 2010 at Seawater scale
    pHsc[is_w14_SWS] <- "SWS" 

    #   pH-scale: 'Total scale'. mol/kg-soln
    is_w14_T <- is_w14 & (pHscale=="T" & P==0)
    A1 <- 13.568513*S[is_w14_T]^(0.5) + 0.031645*S[is_w14_T] - (5.3834e-5)*S[is_w14_T]^2
    B1 <- -539.2304*S[is_w14_T]^(0.5) - 5.635*S[is_w14_T]
    C1 <- -2.0901396*S[is_w14_T]^(0.5)
    pK1o <- 6320.813/TK[is_w14_T] + 19.568224*log(TK[is_w14_T]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_w14_T] + C1*log(TK[is_w14_T])
    K1[is_w14_T] <- 10^(-pK1)   # K1 according to Millero et al. 2010 at Total scale 
    pHsc[is_w14_T] <- "T"

    #   pH-scale: 'Free scale'. mol/kg-soln
    is_w14_F <- is_w14 & (pHscale=="F" & P==0)
    A1 <- 5.592953*S[is_w14_F]^(0.5) + 0.028845*S[is_w14_F] - (6.388e-5)*S[is_w14_F]^2
    B1 <- -225.7489*S[is_w14_F]^(0.5) - 4.761*S[is_w14_F]
    C1 <- -0.8715109*S[is_w14_F]^(0.5)
    pK1o <- 6320.813/TK[is_w14_F] + 19.568224*log(TK[is_w14_F]) -126.34048
    pK1 <- pK1o + A1 + B1/TK[is_w14_F] + C1*log(TK[is_w14_F])
    K1[is_w14_F] <- 10^(-pK1)  # K1 according to Millero et al. 2010 at Total scale
    pHsc[is_w14_F] <- "F"

    ##----------------- Conversion from total to SWS scale
    ##                  if pressure correction needed
    ##                  or pH scale conversion required anyway
    ##                 (in which case SWS may be an intermediate stage of conversion)
    convert <- (pHsc == "T") & ((P > 0) | (pHscale != pHsc))
    if (any (convert))
    {
        ##------------- Convert from total to SWS scale
        # if correction factor (from Total scale to seawater at P=0) not given
        if (missing(ktotal2SWS_P0) || ktotal2SWS_P0[1] == "x")
        {
            # Compute it
            ktotal2SWS_P0  <- rep(1.0,nK)
            ktotal2SWS_P0 <- kconv(S=S[convert], T=T[convert], P=0)$ktotal2SWS
        }
        else
        {
            # Check its length
            if(length(ktotal2SWS_P0)!=nK) ktotal2SWS_P0 <- rep(ktotal2SWS_P0[1], nK)
            # Filter
            ktotal2SWS_P0 <- ktotal2SWS_P0[convert]
        }
        K1[convert] <- K1[convert] * ktotal2SWS_P0
        pHsc[convert] <- "SWS"

        # Just computed constant is on Free scale only if required scale is free scale
        # and if no pressure correction needed  
        # --> No need to convert from free to other scale
        # --> No need to determine conversion factor from free to SWS scale
    }

    # ------------------- Pressure effect --------------------------------
    i_press <- which (P > 0)
    if (length(i_press) > 0)
    {
        # Call Pcorrect() on SWS scale
	# last 2 argupments are conversion factors: they are set to one since no PH scale conversion needed
        K1[i_press] <- Pcorrect(Kvalue=K1[i_press], Ktype="K1", T=T[i_press], 
             S=S[i_press], P=P[i_press], pHscale=pHsc[i_press], 1., 1., warn=warn)
    }


    ## --------------- Last conversion in the required pHscale ----------------

    convert <- (pHscale != pHsc)
    # At this stage, if any conversion is needed, it is from SWS scale
    # Which pH scales are required ?
    is_total <- convert & (pHscale=="T")
    is_free  <- convert & (pHscale=="F")

    # if any pH scale correction required (from SWS scale)
    if (any(convert))
    {
        # if pH scale correction factor not given
        if (missing(kSWS2scale) || kSWS2scale[1] == "x")
        {
            # Compute it
            kSWS2scale <- rep(1.0,nK)
	    if (any(is_total)){ kSWS2scale[is_total] <- kconv(S=S[is_total], T=T[is_total], P=P[is_total])$kSWS2total }
            if (any(is_free)){  kSWS2scale[is_free]  <- kconv(S=S[is_free],  T=T[is_free],  P=P[is_free])$kSWS2free }
        }
        else
        {
            # Check its length
            if (length(kSWS2scale)!=nK) kSWS2scale <- rep(kSWS2scale[1], nK)
        }
        # Apply pH scale correction
        K1[convert] <- K1[convert] * kSWS2scale[convert]
    }

    # Return full name of pH scale
    is_total <- (pHscale=="T")
    is_free  <- (pHscale=="F")
    pHlabel[is_total] <- "total scale"
    pHlabel[is_free]  <- "free scale"
    pHlabel[!is_total & !is_free] <- "seawater scale"


    ##------------Warnings

    is_w <- warn == "y"

    if (   any (is_w & is_l & (T>35 | T<2 | S<19 | S>43))  || any (is_w & is_r & (T<0 | T>45 | S<5 | S>45)) 
        || any (is_w & is_mp2 & (T<0 | T>45 | S<5 | S>42)) || any (is_w & is_m02 & (T<(-1.6) | T>35 | S<34 | S>37))
        || any (is_w & is_m06 & (T<1 | T>50 | S<0.1 | S>50)) 
        || any (is_w & (is_m10 | is_w14) & (T<0 | T>50 | S<1 | S>50))
        || any (is_w & is_cw & (T<0.2 | T>30 | S<0.0 | S>40)) 
        || any (is_w & is_p18 & (T<(-6) | T>25 | S<33 | S>100)) 
        || any (is_w & is_s20 & (T<(-1.7) | T>31.8 | S<30.7 | S>37.6)) )
        warning("S and/or T is outside the range of validity of the formulation chosen for K1.")

    if (any(is_w & (T>50 | S>50))) 
        warning("S and/or T is outside the range of validity of the formulations available for K1 in seacarb.")

    ##---------------Attributes
    method <- rep(NA, nK)
    method[is_cw]  <- "Cai and Wang (1998)"
    method[is_mp2] <- "Mojica Prieto et al. (2002)"
    method[is_m02] <- "Millero et al. (2002)"
    method[is_m06] <- "Millero et al. (2006)"
    method[is_m10] <- "Millero (2010)"
    method[is_w14] <- "Waters et al. (2014)"
    method[is_r]   <- "Roy et al. (1993)"
    method[is_p18] <- "Papadimitriou et al. (2018)"
    method[is_s20] <- "Sulpis et al. (2020)"
    method[! (is_cw | is_mp2 | is_m02 | is_m06 | is_m10 | is_w14 | is_r | is_p18 | is_s20) ] <- "Luecker et al. (2000)"

    attr(K1,"unit") <- "mol/kg-soln"
    attr(K1,"pH scale") <- pHlabel
    attr(K1, "method") <- method
    return(K1)
}

# Copyright (C) 2009 Jean-Pierre Gattuso
# Copyright (C) 2018 Jens Daniel Mueller, update
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"tris" <-
function(S=35, T=25, b=0.04, k="d98", warn="y")
  {
  #-------Harmonize input vector length-----
  
  n <- max(length(S), length(T), length(k), length(b))
  
  if(length(S)!=n){S <- rep(S[1],n)}
  if(length(T)!=n){T <- rep(T[1],n)}
  if(length(b)!=n){b <- rep(b[1],n)}
  if(length(k)!=n){k <- rep(k[1],n)}


  #-------Initialise output vector------- 
  
  tris <- rep(NA, n)
  method <- rep(NA, n)
  
  #-------Change temperature scale from [deg C] to [K] ----
  
  TK = T + 273.15;           # T [C]; TK [K]
  
  #-------Calculate tris buffer pH based on input values-------
  
  #--------------------------------------------------------------
  #-------tris characterization by DelValls and Dickson 1998----
    #
    #       doi: 10.1016/S0967-0637(98)00019-3
    #       Also Dickson et al. (2007)
    #       pH of TRIS buffered artifical seawater solutions
    #
    #       pH-scale: total scale
    #        
    #       correct for T range : 0 - 45 °C
    #       correct for S range : 20 - 40
    #       correct for equmolal TRIS/TRISH+ molality b: 0.04 mol/kg-H20
  #--------------------------------------------------------------  
  
  tris_d98 <- ((11911.08-18.2499*S-0.039336*S^2)*1/(TK))-
	  366.27059+0.53993607*S+0.00016329*S^2+
	  ((64.52243-0.084041*S)*log(TK))-
	  (0.11149858*(TK))
  
	#--------------------------------------------------------------
	#-------tris characterization by Mueller et al 2018-----------
	#
	#       doi: 10.3389/fmars.2018.00176    
	#     
	#       pH of TRIS buffered artifical seawater solutions
	#       extend to low salinities and different TRIS/TRISH+ molalities
	#
	#       pH-scale: total scale
	#        
	#       correct for T range : 5 - 45 °C
	#       correct for S range : 5 - 40
	#       correct for equmolal TRIS/TRISH+ molalities b: 0.01 - 0.04 mol/kg-H20 for S in 5-20
	#       correct for equmolal TRIS/TRISH+ molalities b: 0.04 mol/kg-H20 for S in 20-40
	#--------------------------------------------------------------
	
  tris_m18 <-
	  -327.3307 -                 
	  2.400270 * S +    
	  8.124630e-2 * S^2 -   
	  9.635344e-4 * S^3 -   
	  
	  9.103207e-2 * TK -
	  1.963311e-3 * S * TK +
	  6.430229e-5 * S^2 * TK -
	  7.510992e-7 * S^3 * TK +
	  
	  56.92797 * log(TK) +
	  5.235889e-1 * S * log(TK) -
	  1.7602e-2 * S^2 * log(TK) +
	  2.082387e-4 * S^3 * log(TK) +
	  
	  11382.97 * (1/TK) -
	  
	  2.417045 * b +
	  7.645221e-2 * b * S +
	  1.122392e-2 * b * TK -
	  3.248381e-4 * b * S * TK -
	  
	  4.161537 * b^2 +
	  6.143395e-2 * b^2 * S
	
  #-------------------------------------------------------------------
	#-------Choose between tris characterizations (=method)----------
	
  is_d98 <- (k=='d98')
  is_m18 <- (k=='m18')
	
  tris[is_d98] <- tris_d98[is_d98]
  tris[is_m18] <- tris_m18[is_m18]
	
	#-------Assign method -------------------
	
	method[is_d98] <- "DelValls and Dickson (1998)"
  method[is_m18] <- "Mueller et al. (2018)"
	
	#-------Set warnings-----------
	is_w <- warn == "y"
	if (any(is_w & is_d98 & (T>45 | T<0 | S>40 | S<20 | b!=0.04)))
	  {warning("S, T, and/or b is outside the range of validity for the TRIS buffer pH formulation by DelValls and Dickson (1998).")}

	if (any(is_w & is_m18 & (T>45 | T<5 | S>40 | b>0.04 | b<0.01)))
	  {warning("S, T, and/or b is outside the range of validity for the TRIS buffer pH formulation by Mueller et al. (2018).")}
	
	#-------Assign attributes and define return value-----------
	
	attr(tris, "method") = method
	attr(tris, "pH scale") = "total scale"
	attr(tris,"unit") <- "mol/kg-soln"
	return(tris)

}

# Copyright (C) 2014 James Orr
# This file is part of seacarb.
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Convert pCO2 (uatm) to xCO2 (ppm) 

p2xCO2 <- function(S=35, T=25, Patm=1.0, pCO2=400){
  pH20 <- vapress(T=T, S=S, form="d2007")
  xCO2 <- pCO2 / (Patm - pH20) 
return(xCO2)
}

# Copyright (C) 2008 Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye and Jean-Pierre Gattuso  
# with a most valuable contribution of Bernard Gentili
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
# This function is the common implementation of  three public funtions : carb(), carbb() and carbfull()
# As such, it accepts as input the parameters of all three functions.
#
# The last parameter, "envir", is an optional environment. When given, it is the environment of the caller funtion.
# This environment may contain redefinition of some public seacarb functions such as K0(), K1(), K2(), ....

calculate_carb<-
function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, NH4t=0, HSt=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", badd=0,
         warn="y", eos="eos80", long=1.e20, lat=1.e20, fullresult=FALSE, envir=NULL){
    n <- max(length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), length(ks), length(b))
    if(length(flag)!=n){flag <- rep(flag[1],n)}
    if(length(var1)!=n){var1 <- rep(var1[1],n)}
    if(length(var2)!=n){var2 <- rep(var2[1],n)}
    if(length(S)!=n){S <- rep(S[1],n)}
    if(length(T)!=n){T <- rep(T[1],n)}
    if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
    if(length(P)!=n){P <- rep(P[1],n)}
    if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
    if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
    if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
    if(length(kf)!=n){kf <- rep(kf[1],n)}
    if(length(ks)!=n){ks <- rep(ks[1],n)}
    if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
    if(length(b)!=n){b <- rep(b[1],n)}
    if(length(gas)!=n){gas <- rep(gas[1],n)}
    if (fullresult) {
        if(length(NH4t)!=n){NH4t <- rep(NH4t[1],n)}
        if(length(HSt)!=n){HSt <- rep(HSt[1],n)}
    } else {
        NH4t <- rep(0,n)
        HSt <- rep(0,n)
    }
    
    # if the concentrations of total silicate, total phosphate, total nitrite,
    # total ammonium, and total hydrogen sulfide are NA, they are set at 0
    Sit[is.na(Sit)] <- 0
    Pt[is.na(Pt)] <- 0
    NH4t[is.na(NH4t)] <- 0
    HSt[is.na(HSt)] <- 0
    
    # Only two options for eos
    if (eos != "teos10" && eos != "eos80")
        stop ("invalid parameter eos: ", eos)
    
    # if use of EOS-10 standard
    if (eos == "teos10")
    {
        # Must convert temperature and salinity from TEOS-10 to EOS-80
        # convert temperature: from Conservative (CT) to in-situ temperature
        # and salinity from Absolute to Practical (SP)
        STeos <- teos2eos_geo (S, T, P, long, lat)
        InsT <- STeos$T
        SP <- STeos$SP
    }
    else
    {
        InsT <- T
        SP <- S
    }
 
    #-------Constants----------------   
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])  
    TK = InsT + tk;           # TK [K]; T[C]
    
    #---- issues de equic----
    Cl = SP / 1.80655;             # Cl = chlorinity; S = salinity (per mille)
    ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
    FLUO = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
 
    # get functions that compute K0, K1, K2, Kb, Kw, Kspa, Kspc and Boron concentration
    # --------------------------------------------------------------------------------------------------------------------
    
    if (! is.null(envir)) 
    {
        # These functions may, in certain cases, have a replacement function (when computing derivatives, namely)
        # We get the original one or, when it exists, the replacement one by accessing them through the "environment" given as input parameter
        
        K0_fct <-get("K0", envir=envir)
        K1_fct <-get("K1", envir=envir)
        K2_fct <-get("K2", envir=envir)
        Kw_fct <-get("Kw", envir=envir) 
        Kb_fct <-get("Kb", envir=envir)
        Kspa_fct <-get("Kspa", envir=envir)
        Kspc_fct <-get("Kspc", envir=envir)
        bor_fct <-get("bor", envir=envir)
    }
    else
    {
        # Get the original functions from Seacarb
        K0_fct <-K0
        K1_fct <-K1
        K2_fct <-K2
        Kw_fct <-Kw
        Kb_fct <-Kb
        Kspa_fct <-Kspa
        Kspc_fct <-Kspc
        bor_fct <-bor
    }
   
    
    #---------------------------------------------------------------------
    #-------------- compute K's and Boron ---------------------
    #---------------------------------------------------------------------
    
    BOR = bor_fct(S=S , b=b) + badd;         # (mol/kg) total boron + boron added
    
    # Ks (free pH scale) at zero pressure and given pressure
    Ks_P0 <- Ks(S=SP, T=InsT, P=0, ks=ks, warn=warn)
    Ks    <- Ks(S=SP, T=InsT, P=P, ks=ks, warn=warn)
    
    # Kf on free pH scale
    Kff_P0 <- Kf(S=SP, T=InsT, P=0, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
    Kff <- Kf(S=SP, T=InsT, P=P, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
    # Kf on given pH scale
    Kf <- Kf(S=SP, T=InsT, P=P, pHscale=pHscale, kf=kf, Ks_P0, Ks, warn=warn)
    
    # Conversion factor from total to SWS pH scale at zero pressure
    ktotal2SWS_P0 <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks_P0,Kff=Kff_P0)$ktotal2SWS

    # Conversion factor from SWS to chosen pH scale
    conv <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks,Kff=Kff, warn=warn)
    kSWS2chosen <- rep(1.,n)
    kSWS2chosen [pHscale == "T"] <- conv$kSWS2total [pHscale == "T"]
    kSWS2chosen [pHscale == "F"] <- conv$kSWS2free [pHscale == "F"]  

    K1 <- K1_fct(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)   
    K2 <- K2_fct(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn)
    Kw <- Kw_fct(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    Kb <- Kb_fct(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0, warn=warn)
    K1p <- K1p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    K2p <- K2p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    K3p <- K3p(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    Ksi <- Ksi(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn)
    Kspa <- Kspa_fct(S=SP, T=InsT, P=P, warn=warn)
    Kspc <- Kspc_fct(S=SP, T=InsT, P=P, warn=warn)
    if (fullresult) {
        Kn <- Kn(S=S, T=T, P=P, pHscale=pHscale)
        Khs <- Khs(S=S, T=T, P=P, pHscale=pHscale)
        K2si <- K2si(S=S, T=T, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0)
    } else {
        Kn <- rep(0,n)
        Khs <- rep(0,n)
        K2si <- rep(0,n)
    }
    
    rho <- rho(S=SP,T=InsT,P=P)

    # Compute "standard" K0 with S, in situ T, and atmospheric pressure
    K0 <- K0_fct(S=SP, T=InsT, Patm=Patm, P=0, warn=warn)                         
    # Compute potential K0 with S, potential temperature, and atmospheric pressure (usually 1 atm)
    K0pot <- K0_fct(S=SP, T=theta(S=SP, T=InsT, P=P, Pref=0), Patm=Patm, P=0, warn=warn)
    # Compute in situ K0 with S, in situ temperature, and total pressure pressure (atmospheric + hydrostatic)
    K0insitu <- K0_fct(S=SP, T=InsT, Patm=Patm, P=P, warn=warn)
 
    #------------------------------------------------------------------#
    #------------------------------------------------------------------#
    #                            VARIABLES                             #
    #------------------------------------------------------------------#
    #------------------------------------------------------------------#
    
    # flag = 1      pH-CO2 given
    # flag = 2      CO2-HCO3 given
    # flag = 3      CO2-CO3 given
    # flag = 4      CO2-ALK given
    # flag = 5      CO2-DIC given
    # flag = 6      pH and HCO3 given
    # flag = 7      pH and CO3 given
    # flag = 8      pH and ALK given
    # flag = 9      pH and DIC given
    # flag = 10     HCO3 and CO3 given
    # flag = 11     HCO3 and ALK given
    # flag = 12     HCO3 and DIC given
    # flag = 13     CO3 and ALK given
    # flag = 14     CO3 and DIC given
    # flag = 15     ALK and DIC given
    # flag = 21     pCO2-pH given
    # flag = 22     pCO2-HCO3 given
    # flag = 23     pCO2-CO3 given
    # flag = 24     pCO2-ALK given
    # flag = 25     pCO2-DIC given
    
    # Initialise output vectors
    H   <- rep(NA, n)
    PH   <- rep(NA, n)
    CO2  <- rep(NA, n)
    pCO2 <- rep(NA, n)
    fCO2 <- rep(NA, n)
    HCO3 <- rep(NA, n)
    CO3  <- rep(NA, n)
    DIC  <- rep(NA, n)
    ALK  <- rep(NA, n)

    # ------------ case 1.) PH and CO2 given ----
    # Indices of flag elements where flag = 1
    i_flag_1 <- which (flag == 1)    
    PH[i_flag_1]   <- var1[i_flag_1]
    CO2[i_flag_1]  <- var2[i_flag_1]
    h <- 10^(-PH[i_flag_1])
    fCO2[i_flag_1] <- CO2[i_flag_1] / K0[i_flag_1]
    HCO3[i_flag_1] <- (K1[i_flag_1] * CO2[i_flag_1]) / h
    CO3[i_flag_1]  <- (K2[i_flag_1] * HCO3[i_flag_1]) / h
    DIC[i_flag_1]  <- CO2[i_flag_1] + HCO3[i_flag_1] + CO3[i_flag_1]
    H[i_flag_1] <- h
    
    # ------------ case 2.) CO2 and HCO3 given ----
    # Indices of flag elements where flag = 2
    i_flag_2 <- which (flag == 2)     
    CO2[i_flag_2]  <- var1[i_flag_2]
    HCO3[i_flag_2] <- var2[i_flag_2]
    fCO2[i_flag_2] <- CO2[i_flag_2] / K0[i_flag_2]
    h <- K0[i_flag_2] * K1[i_flag_2] * fCO2[i_flag_2] / HCO3[i_flag_2]
    CO3[i_flag_2]  <- K0[i_flag_2] * K1[i_flag_2] * K2[i_flag_2] * fCO2[i_flag_2] / (h^2)
    DIC[i_flag_2]  <- CO2[i_flag_2] + HCO3[i_flag_2] + CO3[i_flag_2]
    PH[i_flag_2]   <- -log10(h)
    H[i_flag_2] <- h

    # ------------ case 3.) CO2 and CO3 given ----
    # Indices of flag elements where flag = 3
    i_flag_3 <- which (flag == 3)     
    CO2[i_flag_3]  <- var1[i_flag_3]
    CO3[i_flag_3]  <- var2[i_flag_3]
    fCO2[i_flag_3] <- CO2[i_flag_3] / K0[i_flag_3]
    k1co2 <- K1[i_flag_3] * CO2[i_flag_3]
    h <- sqrt((K2[i_flag_3]*k1co2) / CO3[i_flag_3])
    HCO3[i_flag_3] <- k1co2 / h
    DIC[i_flag_3]  <- CO2[i_flag_3] + HCO3[i_flag_3] + CO3[i_flag_3]
    PH[i_flag_3]   <- -log10(h)
    H[i_flag_3] <- h

    # ------------ case 4.) CO2 and ALK given ----
    # Indices of flag elements where flag = 4
    i_flag_4 <- which (flag == 4)     
    CO2[i_flag_4] <- var1[i_flag_4]
    ALK[i_flag_4] <- var2[i_flag_4]
    
    # Calculate [H+] from [CO2] and total alk
    h <- rep(NA, length(i_flag_4))
    j <- 1
    for(i in (i_flag_4))
    {
        # Create a list containing all the dissociation constants
        dissoc = list(K1_DIC=K1[i], K2_DIC=K2[i], K_BT=Kb[i], 
                        K1_PO4 = K1p[i], K2_PO4 = K2p[i], K3_PO4 = K3p[i], K_Sil = Ksi[i],
                        K2_Sil = K2si[i], K_NH4 = Kn[i], K_H2S = Khs[i],
                        K_HSO4 = Ks[i], K_HF = Kff[i], K_H2O = Kw[i])

        # use SolveSAPHE v2.0
        result <- solve_pH_from_AT(ALK[i], CO2[i], BOR[i], Pt[i], Sit[i], NH4t[i], HSt[i], 
                                    ST[i], FLUO[i], pHscale[i], "CO2", FALSE, dissoc)
        h[j] <- result[1]
        j <- j + 1
    }
 
    fCO2[i_flag_4] <- CO2[i_flag_4] / K0[i_flag_4]
    HCO3[i_flag_4] <- K1[i_flag_4]*CO2[i_flag_4]/h
    CO3[i_flag_4]  <- K2[i_flag_4]*HCO3[i_flag_4]/h
    PH[i_flag_4]   <- -log10(h)
    H[i_flag_4] <- h
    DIC[i_flag_4]  <- CO2[i_flag_4] + HCO3[i_flag_4] + CO3[i_flag_4]


    # ------------ case 5.) CO2 and DIC given ----
    # Indices of flag elements where flag = 5
    i_flag_5 <- which (flag == 5)     
    CO2[i_flag_5]  <- var1[i_flag_5]
    DIC[i_flag_5]  <- var2[i_flag_5]
    fCO2[i_flag_5] <- CO2[i_flag_5] / K0[i_flag_5]
    a <- K1[i_flag_5] * K2[i_flag_5] * CO2[i_flag_5]
    b <- K1[i_flag_5] * CO2[i_flag_5]
    c <- CO2[i_flag_5] - DIC[i_flag_5]
    D <- b*b - 4*a*c
    h <- (2*a) / (sqrt(D)-b)
    HCO3[i_flag_5] <- K0[i_flag_5] * K1[i_flag_5] * fCO2[i_flag_5] / h
    CO3[i_flag_5] <- DIC[i_flag_5] - CO2[i_flag_5] - HCO3[i_flag_5]
    PH[i_flag_5] <- -log10(h)
    H[i_flag_5] <- h

    # ------------ case 6.) PH and HCO3 given ----
    # Indices of flag elements where flag = 6
    i_flag_6 <- which (flag == 6)     
    PH[i_flag_6] <- var1[i_flag_6]
    HCO3[i_flag_6] <- var2[i_flag_6]
    h <- 10^(-PH[i_flag_6])
    CO2[i_flag_6] <- (HCO3[i_flag_6] * h)/K1[i_flag_6]
    CO3[i_flag_6] <- K2[i_flag_6] * HCO3[i_flag_6] /h
    DIC[i_flag_6] <- CO2[i_flag_6] + HCO3[i_flag_6] + CO3[i_flag_6]
    fCO2[i_flag_6] <- CO2[i_flag_6]/K0[i_flag_6]
    H[i_flag_6] <- h

    # ------------ case 7.) PH and CO3 given ---- 
    # Indices of flag elements where flag = 7
    i_flag_7 <- which (flag == 7)     
    PH[i_flag_7] <- var1[i_flag_7]
    CO3[i_flag_7] <- var2[i_flag_7]
    h <- 10^(-PH[i_flag_7])
    HCO3[i_flag_7] <- CO3[i_flag_7] * h/K2[i_flag_7]
    CO2[i_flag_7]  <- HCO3[i_flag_7] * h/K1[i_flag_7]
    fCO2[i_flag_7] <- CO2[i_flag_7]/K0[i_flag_7]
    DIC[i_flag_7]  <- CO2[i_flag_7] + HCO3[i_flag_7] + CO3[i_flag_7]
    H[i_flag_7] <- h

    # ------------ case 8.) PH and ALK given ----
    # Indices of flag elements where flag = 8
    i_flag_8 <- which(flag == 8)     
    PH[i_flag_8]  <- var1[i_flag_8]
    ALK[i_flag_8] <- var2[i_flag_8] 
    h <- 10^(-PH[i_flag_8])
    H[i_flag_8] <- h

    ## calculate Hfree anf Htot
    hfree <- rep(NA, length(i_flag_8))
    htot  <- rep(NA, length(i_flag_8))    
    sc <- pHscale[i_flag_8] 
    st <- ST[i_flag_8]
    ks  <- Ks[i_flag_8]
    fluo <- FLUO[i_flag_8]
    kff  <- Kff[i_flag_8]
    
    # Where pHscale=="F", pHscale = free scale
    i_sc_F <- which (sc == "F") 
    hfree[i_sc_F] <- h[i_sc_F]  
    htot[i_sc_F]  <- h[i_sc_F] / (1+st[i_sc_F]/ks[i_sc_F]) 
    # Where pHscale=="T", pHscale = total scale
    i_sc_T <- which (sc == "T") 
    hfree[i_sc_T] <- h[i_sc_T] * (1+st[i_sc_T]/ks[i_sc_T])
    htot[i_sc_T]  <- h[i_sc_T]
    # Where pHscale=="SWS", pHscale = SW scale
    i_sc_S <- which (sc == "SWS") 
    hfree[i_sc_S] <- h[i_sc_S] * (1 + st[i_sc_S]/ks[i_sc_S] + fluo[i_sc_S]/kff[i_sc_S])
    htot[i_sc_S]  <- hfree[i_sc_S] / (1+st[i_sc_S]/ks[i_sc_S])

    # Calculate some invariable components of total alkalinity
    boh4   <- BOR[i_flag_8] / (1+h/Kb[i_flag_8])
    oh     <- Kw[i_flag_8]/h
    temp   <- (h^3 + K1p[i_flag_8]*h^2 + K1p[i_flag_8]*K2p[i_flag_8]*h +
              K1p[i_flag_8]*K2p[i_flag_8]*K3p[i_flag_8]) 
    h3po4  <- Pt[i_flag_8]*h^3 / temp
    hpo4   <- Pt[i_flag_8]*K1p[i_flag_8]*K2p[i_flag_8]*h / temp
    po4    <- Pt[i_flag_8]*K1p[i_flag_8]*K2p[i_flag_8]*K3p[i_flag_8] / temp
    hso4   <- st/(1+ks/hfree)
    hf     <- fluo / (1+Kf[i_flag_8]/htot)

    if (fullresult) {
        # adapted to include second dissociation constant of silicate
        siooh3 <- Sit[i_flag_8]*h*Ksi[i_flag_8] / 
        (h*h + Ksi[i_flag_8]*h + Ksi[i_flag_8]*K2si[i_flag_8])
        sio2oh2 <- Sit[i_flag_8]*Ksi[i_flag_8]*K2si[i_flag_8] / 
        (h*h + Ksi[i_flag_8]*h + Ksi[i_flag_8]*K2si[i_flag_8])
        nh3 <- NH4t[i_flag_8]/(1+h/Kn[i_flag_8])
        hs <- HSt[i_flag_8]/(1+h/Khs[i_flag_8])

        # Sum of these components (partial alkalinity)
        alk_p <- boh4+oh+hpo4+2*po4+siooh3+2*sio2oh2+nh3+hs-hfree-hso4-hf-h3po4
    }
    else {
        siooh3 <- Sit[i_flag_8] / (1+h/Ksi[i_flag_8])
        # Sum of these components (partial alkalinity)
        alk_p <- boh4+oh+hpo4+2*po4+siooh3-hfree-hso4-hf-h3po4
    }
    
    # As [HCO3-] and  [CO3--] are both linearly dependent on DIC
    #    hco3  = co2*K1/h = DIC/(1+K1/h+K1*K2/(h^2)) * K1/h
    #   2*co3  = 2*hco3*K2/h = DIC/(1+K1/h+K1*K2/(h^2)) * 2*K1*K2/h^2
    # carbonate alk ([HCO3-] + 2*[CO3--]) is also linearly dependent on DIC
    # Calculate DIC from carbonate alk (total alk - partial alk)
    temp <- h*h + K1[i_flag_8]*h + K1[i_flag_8]*K2[i_flag_8]
    DIC[i_flag_8] <- (ALK[i_flag_8]-alk_p) * temp / 
         (h*K1[i_flag_8] + 2*K1[i_flag_8]*K2[i_flag_8])

    CO2[i_flag_8]  <- DIC[i_flag_8] / 
        (1 + K1[i_flag_8]/h + K1[i_flag_8]*K2[i_flag_8]/(h^2))
    HCO3[i_flag_8] <- CO2[i_flag_8]*K1[i_flag_8]/h
    CO3[i_flag_8]  <- HCO3[i_flag_8]*K2[i_flag_8]/h
    fCO2[i_flag_8] <- CO2[i_flag_8]/K0[i_flag_8]
    
    # ------------ case 9.) PH and DIC given ----
    # Indices of flag elements where flag = 9
    i_flag_9 <- which (flag == 9)     
    PH[i_flag_9]  <- var1[i_flag_9]
    DIC[i_flag_9] <- var2[i_flag_9]
    h <- 10^(-PH[i_flag_9])
    H[i_flag_9] <- h
    temp <- h*h + K1[i_flag_9]*h + K1[i_flag_9]*K2[i_flag_9]
    HCO3[i_flag_9] <- (DIC[i_flag_9]*K1[i_flag_9]*h) / temp
    CO3[i_flag_9]  <- (DIC[i_flag_9]*K1[i_flag_9]*K2[i_flag_9]) / temp
    CO2[i_flag_9]  <- h*HCO3[i_flag_9]/K1[i_flag_9]
    fCO2[i_flag_9] <- CO2[i_flag_9]/K0[i_flag_9]
    
    # ------------ case 10.) HCO3 and CO3 given ----
    # Indices of flag elements where flag = 10
    i_flag_10 <- which (flag == 10)     
    HCO3[i_flag_10] <- var1[i_flag_10]
    CO3[i_flag_10]  <- var2[i_flag_10]
    h <- K2[i_flag_10]*HCO3[i_flag_10]/CO3[i_flag_10]
    CO2[i_flag_10]  <- h*HCO3[i_flag_10]/K1[i_flag_10]
    DIC[i_flag_10]  <- CO2[i_flag_10] + HCO3[i_flag_10] + CO3[i_flag_10]
    fCO2[i_flag_10] <- CO2[i_flag_10]/K0[i_flag_10]
    PH[i_flag_10] <- -log10(h)
    H[i_flag_10] <- h

    # ------------ case 11.) HCO3 and ALK given ----
    # Indices of flag elements where flag = 11
    i_flag_11 <- which (flag == 11)     
    HCO3[i_flag_11] <- var1[i_flag_11]
    ALK[i_flag_11]  <- var2[i_flag_11]

    # Calculate [H+] from [HCO3] and total alk
    h <- rep(NA, length(i_flag_11))
    j <- 1 
    for(i in (i_flag_11))
    {
        # Create a list containing all the dissociation constants
        dissoc = list(K1_DIC=K1[i], K2_DIC=K2[i], K_BT=Kb[i], 
                        K1_PO4 = K1p[i], K2_PO4 = K2p[i], K3_PO4 = K3p[i], K_Sil = Ksi[i], 
                        K2_Sil = K2si[i], K_NH4 = Kn[i], K_H2S = Khs[i],
                        K_HSO4 = Ks[i], K_HF = Kff[i], K_H2O = Kw[i])

        # Use SolveSAPHE PACKAGE
        result <- solve_pH_from_AT(ALK[i], HCO3[i], BOR[i], Pt[i], Sit[i], NH4t[i], HSt[i],
                                    ST[i], FLUO[i], pHscale[i], "HCO3", FALSE, dissoc)
        h[j] <- result[1]
        j <- j + 1
    }
    
    CO2[i_flag_11] <- h*HCO3[i_flag_11]/K1[i_flag_11]
    CO3[i_flag_11] <- K2[i_flag_11]*HCO3[i_flag_11]/h
    DIC[i_flag_11] <- CO2[i_flag_11] + HCO3[i_flag_11] + CO3[i_flag_11]
    PH[i_flag_11] <- -log10(h)
    H[i_flag_11] <- h
    fCO2[i_flag_11] <- CO2[i_flag_11]/K0[i_flag_11]

    # ------------ case 12.) HCO3 and DIC given
    # Indices of flag elements where flag = 12
    i_flag_12 <- which (flag == 12)     
    HCO3[i_flag_12] <- var1[i_flag_12]
    DIC[i_flag_12]  <- var2[i_flag_12]
    a <- HCO3[i_flag_12]
    b <- K1[i_flag_12]*(HCO3[i_flag_12]-DIC[i_flag_12])
    c <- K1[i_flag_12]*K2[i_flag_12]*HCO3[i_flag_12]
    D <- b*b - 4*a*c
    h <- (-b-sqrt(D))/(2*a)
    CO2[i_flag_12]  <- h*HCO3[i_flag_12]/K1[i_flag_12]
    CO3[i_flag_12]  <- K2[i_flag_12]*HCO3[i_flag_12]/h
    fCO2[i_flag_12] <- CO2[i_flag_12]/K0[i_flag_12]
    PH[i_flag_12] <- -log10(h)
    H[i_flag_12] <- h

    # ------------ case 13.) CO3 and ALK given ----
    # Indices of flag elements where flag = 13
    i_flag_13 <- which (flag == 13)     
    CO3[i_flag_13] <- var1[i_flag_13]
    ALK[i_flag_13] <- var2[i_flag_13]

    h <- rep(NA, length(i_flag_13))
    j <- 1 
    for(i in (i_flag_13))
    {
        # Create a list containing all the dissociation constants
        dissoc = list(K1_DIC=K1[i], K2_DIC=K2[i], K_BT=Kb[i], 
                        K1_PO4 = K1p[i], K2_PO4 = K2p[i], K3_PO4 = K3p[i], K_Sil = Ksi[i], 
                        K2_Sil = K2si[i], K_NH4 = Kn[i], K_H2S = Khs[i],
                        K_HSO4 = Ks[i], K_HF = Kff[i], K_H2O = Kw[i])

        # Use solveSAPHE package
        result <- solve_pH_from_AT(ALK[i], CO3[i], BOR[i], Pt[i], Sit[i], NH4t[i], HSt[i],
                                    ST[i], FLUO[i], pHscale[i], "CO3", FALSE, dissoc)

        if (result[2] == -1) # only one solution for pH
        {                
            h[j] <- result[1]
        }
        else    # two solutions for pH
        {
            warning ("Given Alk and [CO3--], there were 2 solutions for pH. The one closest to pH7 was chosen !")
            # look for the solution closest to ph7
            h1 <- result[1]
            h2 <- result[2]
            if (abs(log10(h1) + 7) < abs(log10(h2) + 7)) {
                i_close7 <- 1
            } else {
                i_close7 <- 2
            }
            h[j] <- result[i_close7]
        }
        j <- j + 1
    }
    
    HCO3[i_flag_13] <- h*CO3[i_flag_13]/K2[i_flag_13]
    CO2[i_flag_13]  <- h*HCO3[i_flag_13]/K1[i_flag_13]
    fCO2[i_flag_13] <- CO2[i_flag_13]/K0[i_flag_13]
    DIC[i_flag_13]  <- HCO3[i_flag_13] + CO2[i_flag_13] + CO3[i_flag_13]
    PH[i_flag_13] <- -log10(h)
    H[i_flag_13] <- h

    # ------------ case 14.) CO3 and DIC given ----
    # Indices of flag elements where flag = 14
    i_flag_14 <- which (flag == 14)     
    CO3[i_flag_14] <- var1[i_flag_14]
    DIC[i_flag_14] <- var2[i_flag_14]

    a <- CO3[i_flag_14]
    b <- K1[i_flag_14]*CO3[i_flag_14]
    c <- K1[i_flag_14] * K2[i_flag_14] * (CO3[i_flag_14]-DIC[i_flag_14])
    D <- b*b - 4*a*c
    h <- (-b+sqrt(D))/(2*a)
    HCO3[i_flag_14] <- h*CO3[i_flag_14]/K2[i_flag_14]
    CO2[i_flag_14]  <- h*HCO3[i_flag_14]/K1[i_flag_14]
    fCO2[i_flag_14] <- CO2[i_flag_14]/K0[i_flag_14]
    PH[i_flag_14] <- -log10(h)
    H[i_flag_14] <- h
    
    # ------------ case 15.) ALK and DIC given ----
    # Indices of flag elements where flag = 15
    i_flag_15 <- which(flag == 15)     
    ALK[i_flag_15] <- var1[i_flag_15]
    DIC[i_flag_15] <- var2[i_flag_15]

    h <- rep(NA, length(i_flag_15))
    j <- 1 
    for(i in (i_flag_15))
    {
        # Create a list containing all the dissociation constants
        dissoc = list(K1_DIC=K1[i], K2_DIC=K2[i], K_BT=Kb[i], 
                        K1_PO4 = K1p[i], K2_PO4 = K2p[i], K3_PO4 = K3p[i], K_Sil = Ksi[i], 
                        K2_Sil = K2si[i], K_NH4 = Kn[i], K_H2S = Khs[i],
                        K_HSO4 = Ks[i], K_HF = Kff[i], K_H2O = Kw[i])

        # use SolveSAPHE v2.0
        result <- solve_pH_from_AT(ALK[i], DIC[i], BOR[i], Pt[i], Sit[i], NH4t[i], HSt[i],
                                    ST[i], FLUO[i], pHscale[i], "DIC", FALSE, dissoc)
        h[j] <- result[1]
        j <- j + 1
    }

    temp <- h*h + K1[i_flag_15]*h + K1[i_flag_15]*K2[i_flag_15]
    HCO3[i_flag_15] <- (DIC[i_flag_15]*K1[i_flag_15]*h) / temp 
    CO3[i_flag_15]  <- (DIC[i_flag_15]*K1[i_flag_15]*K2[i_flag_15]) / temp
    CO2[i_flag_15]  <- h*HCO3[i_flag_15]/K1[i_flag_15]
    fCO2[i_flag_15] <- CO2[i_flag_15]/K0[i_flag_15]
    PH[i_flag_15] <- -log10(h)
    H[i_flag_15] <- h
    
    #Initialize new 'potential' & "insitu' arrays (this are overwritten below with correct values)
    fCO2pot <- fCO2
    pCO2pot <- pCO2
    fCO2insitu <- fCO2
    pCO2insitu <- pCO2

    # ------------ Compute pCO2 for cases 1 to 15 (pCO2, pCO2insitu, & pCO2pot are NOT input variables) 
    # Indices of flag elements where 1 <= flag <= 15
    i_flag <- which (flag >= 1 & flag <= 15)
    # 1) Classic pCO2: compute from in situ T and surface P (Patm)
    tk <- TK[i_flag]
    B  <- -1636.75 + 12.0408*tk  - 0.0327957*(tk*tk)   + 0.0000316528*(tk*tk*tk);
    fugfac <- exp((Patm[i_flag])*(B + 2*((1-fCO2[i_flag])^2)*(57.7-0.118*tk))/(82.05736*tk))
    pCO2[i_flag]    <- fCO2[i_flag] / fugfac
    # 2) In-situ pCO2: compute from in situ T and total P (Patm + Phydrostatic)
    fCO2insitu[i_flag] <- fCO2[i_flag] * K0[i_flag] / K0insitu[i_flag]
    xCO2approx <- pCO2[i_flag]       #Results virtually insensitive to this (do not use fCO2insitu)
    fugfacinsitu <- exp((Patm[i_flag] + P[i_flag]/1.01325)*(B + 2*((1-xCO2approx)^2)*(57.7-0.118*tk))/(82.05736*tk))
    pCO2insitu[i_flag]    <- fCO2insitu[i_flag] / fugfacinsitu
    # 3) Potential pCO2: compute from potential T & surface P (Patm only)  
    tkp <- theta(S=SP[i_flag], T=InsT[i_flag], P=P[i_flag], Pref=0) + 273.15       #Potential temperature in Kelvin 
    Bpot   <- -1636.75 + 12.0408*tkp - 0.0327957*(tkp*tkp) + 0.0000316528*(tkp*tkp*tkp);
    fCO2pot[i_flag] <- fCO2[i_flag] * K0[i_flag] / K0pot[i_flag]
    fugfacpot <- exp((Patm[i_flag])*(Bpot + 2*((1-fCO2pot[i_flag])^2)*(57.7-0.118*tkp))/(82.05736*tkp))
    pCO2pot[i_flag] <- fCO2pot[i_flag] / fugfacpot

    # ------------ Compute fCO2 for cases 21 to 25 (pCO2, pCO2pot or pCO2insitu is an input variable)
    # ------------ Notice that there are three cases : insitu, potential or standard
    
    # Indices of flag elements where 21 <= flag <= 25
    i_flag <- which (flag >= 21 & flag <= 25)
    # gas option for those elements
    gas_option <- gas[i_flag]

    # ------------ Process "insitu" case

    # select in "i_flag" those indices that have "insitu" option
    i_flag_insitu <- i_flag[gas_option == "insitu"]

    #if there is any such case
    if (any(i_flag_insitu))
    {
        tk    <- TK[i_flag_insitu]
        tkp   <- theta(S=SP[i_flag_insitu], T=InsT[i_flag_insitu], P=P[i_flag_insitu], Pref=0) + 273.15       #Potential temperature in Kelvin
        B     <- -1636.75 + 12.0408*tk  - 0.0327957*(tk*tk)   + 0.0000316528*(tk*tk*tk);
        Bpot  <- -1636.75 + 12.0408*tkp - 0.0327957*(tkp*tkp) + 0.0000316528*(tkp*tkp*tkp);

        # From in situ pCO2 (in situ T, atm + hydro P), compute potential and standard fCO2 and pCO2
        # a) define input as pCO2insitu & compute fCO2insitu
        pCO2insitu[i_flag_insitu] <- var1[i_flag_insitu] * 1e-6
        # xCO2approx <- pCO2insitu[i_flag_insitu]  (this would be a very gross overestimate at say 5000 m)
        xCO2approx <- 0.    #This poor approximation has virtually no effect on computed result
        fugfacinsitu <- exp((Patm[i_flag_insitu]+P[i_flag_insitu]/1.01325)*(B    + 2*((1-xCO2approx)^2)*(57.7-0.118*tk)) /(82.05736*tk))
        fCO2insitu[i_flag_insitu] <- pCO2insitu[i_flag_insitu] * fugfacinsitu
        # b) compute fCO2pot & fCO2
        fCO2pot[i_flag_insitu]    <- fCO2insitu[i_flag_insitu] * K0insitu[i_flag_insitu] / K0pot[i_flag_insitu]
        fCO2[i_flag_insitu]       <- fCO2insitu[i_flag_insitu] * K0insitu[i_flag_insitu] / K0[i_flag_insitu]
        # c) compute pCO2pot & pCO2
        fugfac       <- exp((Patm[i_flag_insitu])                  *(B    + 2*((1-fCO2[i_flag_insitu])^2)      *(57.7-0.118*tk)) /(82.05736*tk))
        fugfacpot    <- exp((Patm[i_flag_insitu])                  *(Bpot + 2*((1-fCO2pot[i_flag_insitu])^2)      *(57.7-0.118*tkp))/(82.05736*tkp))
        pCO2pot[i_flag_insitu] <- fCO2pot[i_flag_insitu] / fugfacpot
        pCO2[i_flag_insitu]    <- fCO2[i_flag_insitu]    / fugfac
    }

    # ------------ Process "potential" case

    # select in "i_flag" those indices that have "potential" option
    i_flag_potential <- i_flag[gas_option == "potential"]

    #if there is any such case
    if (any(i_flag_potential))
    {
        tk    <- TK[i_flag_potential]
        tkp   <- theta(S=SP[i_flag_potential], T=InsT[i_flag_potential], P=P[i_flag_potential], Pref=0) + 273.15       #Potential temperature in Kelvin
        B     <- -1636.75 + 12.0408*tk  - 0.0327957*(tk*tk)   + 0.0000316528*(tk*tk*tk);
        Bpot  <- -1636.75 + 12.0408*tkp - 0.0327957*(tkp*tkp) + 0.0000316528*(tkp*tkp*tkp);

        # From potential pCO2 (potential T, atm P), compute standard and in situ fCO2 and pCO2
        # a) define input as pCO2pot & compute fCO2pot
        pCO2pot[i_flag_potential] <- var1[i_flag_potential] * 1e-6
        fugfacpot <- exp((Patm[i_flag_potential]                     )*(Bpot + 2*((1-pCO2pot[i_flag_potential])^2)*(57.7-0.118*tkp))/(82.05736*tkp))
        fCO2pot[i_flag_potential] <- pCO2pot[i_flag_potential] * fugfacpot
        # b) compute fCO2 & fCO2insitu
        fCO2[i_flag_potential]       <- fCO2pot[i_flag_potential] * K0pot[i_flag_potential] / K0[i_flag_potential]
        fCO2insitu[i_flag_potential] <- fCO2pot[i_flag_potential] * K0pot[i_flag_potential] / K0insitu[i_flag_potential]
        # c) Compute pCO2 & pCO2insitu
        fugfac       <- exp((Patm[i_flag_potential]                  )*(B    + 2*((1-fCO2[i_flag_potential])^2)*(57.7-0.118*tk))/(82.05736*tk))
        fugfacinsitu <- exp((Patm[i_flag_potential]+P[i_flag_potential]/1.01325)*(B    + 2*((1-fCO2[i_flag_potential])^2)*(57.7-0.118*tk))/(82.05736*tk))
        pCO2[i_flag_potential]      <- fCO2[i_flag_potential]      / fugfac
        pCO2insitu[i_flag_potential] <- fCO2insitu[i_flag_potential] / fugfacinsitu
    }

    # ------------ Process "standard" case

    # select in "i_flag" those indices that have "standard" option
    i_flag_standard <- i_flag[gas_option == "standard"]

    #if there is any such case
    if (any(i_flag_standard))
    {
        tk    <- TK[i_flag_standard]
        tkp   <- theta(S=SP[i_flag_standard], T=InsT[i_flag_standard], P=P[i_flag_standard], Pref=0) + 273.15       #Potential temperature in Kelvin
        B     <- -1636.75 + 12.0408*tk  - 0.0327957*(tk*tk)   + 0.0000316528*(tk*tk*tk);
        Bpot  <- -1636.75 + 12.0408*tkp - 0.0327957*(tkp*tkp) + 0.0000316528*(tkp*tkp*tkp);

        # From standard pCO2 (in situ T, atm P), compute potential and in situ fCO2 and pCO2
        # a) define input as pCO2 & compute fCO2
        pCO2[i_flag_standard] <- var1[i_flag_standard] * 1e-6
        fugfac <- exp((Patm[i_flag_standard]                        )*(B    + 2*((1-pCO2[i_flag_standard])^2)*(57.7-0.118*tk))/(82.05736*tk))
        fCO2[i_flag_standard] <- pCO2[i_flag_standard] * fugfac
        # b) compute fCO2pot & fCO2insitu
        fCO2pot[i_flag_standard]    <- fCO2[i_flag_standard] * K0[i_flag_standard] / K0pot[i_flag_standard]
        fCO2insitu[i_flag_standard] <- fCO2[i_flag_standard] * K0[i_flag_standard] / K0insitu[i_flag_standard]
        # c) Compute pCO2pot & pCO2insitu
        fugfacpot    <- exp((Patm[i_flag_standard]                  )*(Bpot + 2*((1-pCO2[i_flag_standard])^2)*(57.7-0.118*tkp))/(82.05736*tkp))
        fugfacinsitu <- exp((Patm[i_flag_standard]+P[i_flag_standard]/1.01325)*(B    + 2*((1-pCO2[i_flag_standard])^2)*(57.7-0.118*tk))/(82.05736*tk))
        pCO2pot[i_flag_standard]   <- fCO2pot[i_flag_standard]   / fugfacpot
        pCO2insitu[i_flag_standard] <- fCO2insitu[i_flag_standard] / fugfacinsitu
    }
    

    # ------------ case 21.) PH and pCO2 given ----
    # Indices of flag elements where flag = 21
    i_flag_21 <- which (flag == 21)
    PH[i_flag_21] <- var2[i_flag_21]
    h <- 10^(-PH[i_flag_21])
    H[i_flag_21] <- h
    CO2[i_flag_21]  <- K0[i_flag_21]*fCO2[i_flag_21]
    HCO3[i_flag_21] <- K1[i_flag_21]*CO2[i_flag_21]/h
    CO3[i_flag_21]  <- K2[i_flag_21]*HCO3[i_flag_21]/h
    DIC[i_flag_21]  <- CO2[i_flag_21] + HCO3[i_flag_21] + CO3[i_flag_21]

    # ------------ case 22.) HCO3 and pCO2 given ----
    # Indices of flag elements where flag = 22
    i_flag_22 <- which (flag == 22)
    HCO3[i_flag_22] <- var2[i_flag_22]
    CO2[i_flag_22]  <- fCO2[i_flag_22]*K0[i_flag_22]
    h <- CO2[i_flag_22]*K1[i_flag_22]/HCO3[i_flag_22]
    CO3[i_flag_22]  <- HCO3[i_flag_22]*K2[i_flag_22]/h
    DIC[i_flag_22]  <- CO2[i_flag_22] + HCO3[i_flag_22] + CO3[i_flag_22]
    PH[i_flag_22] <- -log10(h)
    H[i_flag_22] <- h

    # ------------ case 23.) CO3 and pCO2 given ----
    # Indices of flag elements where flag = 23
    i_flag_23 <- which (flag == 23)
    CO3[i_flag_23] <- var2[i_flag_23]
    h <- sqrt(K0[i_flag_23]*K1[i_flag_23]*K2[i_flag_23]*fCO2[i_flag_23]/CO3[i_flag_23])
    HCO3[i_flag_23] <- h*CO3[i_flag_23]/K2[i_flag_23]
    CO2[i_flag_23]  <- h*HCO3[i_flag_23]/K1[i_flag_23]
    DIC[i_flag_23] <- CO2[i_flag_23] + HCO3[i_flag_23] + CO3[i_flag_23]
    PH[i_flag_23] <- -log10(h)
    H[i_flag_23] <- h

    # ------------ case 24.) ALK and pCO2 given ----
    # Indices of flag elements where flag = 24
    i_flag_24 <- which (flag == 24)
    ALK[i_flag_24] <- var2[i_flag_24]
    CO2[i_flag_24] <- fCO2[i_flag_24]*K0[i_flag_24]

    # From this line on, this case is similar to case 4
    # Calculate [H+] from [CO2] and total alk
    h <- rep(NA, length(i_flag_24))
    j <- 1 
    for(i in (i_flag_24))
    {
        # Create a list containing all the dissociation constants
        dissoc = list(K1_DIC=K1[i], K2_DIC=K2[i], K_BT=Kb[i], 
                        K1_PO4 = K1p[i], K2_PO4 = K2p[i], K3_PO4 = K3p[i], K_Sil = Ksi[i], 
                        K2_Sil = K2si[i], K_NH4 = Kn[i], K_H2S = Khs[i],
                        K_HSO4 = Ks[i], K_HF = Kff[i], K_H2O = Kw[i])

        # use SolveSAPHE v2.0
        result <- solve_pH_from_AT(ALK[i], CO2[i], BOR[i], Pt[i], Sit[i], NH4t[i], HSt[i],
                                    ST[i], FLUO[i], pHscale[i], "CO2", FALSE, dissoc)
        h[j] <- result[1]
        j <- j + 1
    }
    
    HCO3[i_flag_24] <- K1[i_flag_24]*CO2[i_flag_24]/h
    CO3[i_flag_24]  <- K2[i_flag_24]*HCO3[i_flag_24]/h
    PH[i_flag_24]   <- -log10(h)
    H[i_flag_24] <- h
    DIC[i_flag_24]  <- CO2[i_flag_24] + HCO3[i_flag_24] + CO3[i_flag_24]

    # ------------ case 25.) DIC and pCO2 given ----
    # Indices of flag elements where flag = 25
    i_flag_25 <- which (flag == 25)
    DIC[i_flag_25] <- var2[i_flag_25]
    CO2[i_flag_25] <- K0[i_flag_25]*fCO2[i_flag_25]
    # Though case 25 is the same as case 5, computations are made in a different way
    K <- K1[i_flag_25]/K2[i_flag_25]
    b <- K*K0[i_flag_25]*fCO2[i_flag_25]
    c <- (K*K0[i_flag_25]*fCO2[i_flag_25]) * 
         (K0[i_flag_25]*fCO2[i_flag_25]-DIC[i_flag_25])
    D <- b*b - 4*c
    HCO3[i_flag_25] <- (1/2)*(-b + sqrt(D))
    CO3[i_flag_25]  <- DIC[i_flag_25] - CO2[i_flag_25] - HCO3[i_flag_25]
    h <- K1[i_flag_25]*CO2[i_flag_25]/HCO3[i_flag_25]
    PH[i_flag_25] <- -log10(h)
    H[i_flag_25] <- h

    # ------------ CALCULATION OF ALK in cases ----
    cases <- c(1, 2, 3, 5, 6, 7, 9, 10, 12, 14, 21, 22, 23, 25)
    # Indices of flag elements in these cases
    i_flag <- which (flag %in% cases)
    h <- H[i_flag]

    # HCO3[i_flag] <- DIC[i_flag]*h*K1[i_flag]/(h*h + K1[i_flag]*h + K1[i_flag]*K2[i_flag])
    # CO3[i_flag]  <- HCO3[i_flag] * K2[i_flag] / h
    boh4 <- BOR[i_flag]/(1+h/Kb[i_flag])
    oh <- Kw[i_flag]/h 
    temp <- h^3 + K1p[i_flag]*h^2 + K1p[i_flag]*K2p[i_flag]*h + K1p[i_flag]*K2p[i_flag]*K3p[i_flag]
    h3po4 <- Pt[i_flag]*(h^3) / temp
    hpo4  <- Pt[i_flag]*K1p[i_flag]*K2p[i_flag]*h / temp
    po4   <- Pt[i_flag]*K1p[i_flag]*K2p[i_flag]*K3p[i_flag] / temp

    if (fullresult) {
        # adapted to include second dissociation constant of silicate
        siooh3 <- Sit[i_flag]*h*Ksi[i_flag] / 
            (h*h + Ksi[i_flag]*h + Ksi[i_flag]*K2si[i_flag])
        sio2oh2 <- Sit[i_flag]*Ksi[i_flag]*K2si[i_flag] / 
            (h*h + Ksi[i_flag]*h + Ksi[i_flag]*K2si[i_flag])
        nh3 <- NH4t[i_flag]/(1+h/Kn[i_flag])
        hs <- HSt[i_flag]/(1+h/Khs[i_flag])
    }
    else {
        siooh3 <- Sit[i_flag]/(1+h/Ksi[i_flag])
    }
    
    ## calculate Hfree anf Htot
    hfree <- rep(NA, length(i_flag))
    htot  <- rep(NA, length(i_flag))    
    sc <- pHscale[i_flag] 
    st <- ST[i_flag]
    ks  <- Ks[i_flag]
    fluo <- FLUO[i_flag]
    kff  <- Kff[i_flag]
    
    # Where pHscale=="F", pHscale = free scale
    i_sc_F <- which (sc == "F") 
    hfree[i_sc_F] <- h[i_sc_F]  
    htot[i_sc_F]  <- h[i_sc_F] / (1+st[i_sc_F]/ks[i_sc_F]) 
    # Where pHscale=="T", pHscale = total scale
    i_sc_T <- which (sc == "T") 
    hfree[i_sc_T] <- h[i_sc_T] * (1+st[i_sc_T]/ks[i_sc_T])
    htot[i_sc_T]  <- h[i_sc_T]
    # Where pHscale=="SWS", pHscale = SW scale
    i_sc_S <- which (sc == "SWS") 
    hfree[i_sc_S] <- h[i_sc_S] * (1 + st[i_sc_S]/ks[i_sc_S] + fluo[i_sc_S]/kff[i_sc_S])
    htot[i_sc_S]  <- hfree[i_sc_S] / (1+st[i_sc_S]/ks[i_sc_S])

    hso4 <- st/(1+ks/hfree)
    hf   <- fluo/(1+Kf[i_flag]/htot)

    if (fullresult) {
        ALK[i_flag]  <- HCO3[i_flag] + 2*CO3[i_flag] + boh4+oh+hpo4+2*po4+siooh3+2*sio2oh2+nh3+hs-hfree-hso4-hf-h3po4
    }
    else {
        ALK[i_flag]  <- HCO3[i_flag] + 2*CO3[i_flag] + boh4+oh+hpo4+2*po4+siooh3-hfree-hso4-hf-h3po4
    }

    ##########################################################
    # CALCULATION OF ARAGONITE AND CALCITE SATURATION STATE  #
    ##########################################################

    Ca = (0.02128/40.078) * SP/1.80655  #Improved formula from Dickson et al. (2007) as discussed in Orr & Epitalon (2014, revised)
    Oa  <- (Ca*CO3)/Kspa
    Oc  <- (Ca*CO3)/Kspc

    #PCO2 and fCO2 - convert from atm to microatm
    pCO2       <- pCO2*1e6
    fCO2       <- fCO2*1e6    
    pCO2pot    <- pCO2pot*1e6
    fCO2pot    <- fCO2pot*1e6    
    pCO2insitu <- pCO2insitu*1e6
    fCO2insitu <- fCO2insitu*1e6    

    if (fullresult) {
        ####################################################
        # CALCULATION OF FULL ACID-BASE SPECIATION         #
        # not included: HNO3, NO3, HNO2, NO2, S(2-), H2SO4 #
        ####################################################
        NH4     <- (H) / (H + Kn) * NH4t      
        NH3     <- (Kn) / (H + Kn) * NH4t
        BOH3    <- (H) / (H + Kb) * BOR
        BOH4    <- (Kb) / (H + Kb) * BOR
        H3PO4   <- (H^3) / (H^3 + K1p * H^2 + K1p * K2p * H + K1p * K2p * K3p) * Pt
        H2PO4   <- (K1p * H^2) / (H^3 + K1p * H^2 + K1p * K2p * H + K1p * K2p * K3p) * Pt
        HPO4    <- (K1p * K2p * H) / (H^3 + K1p * H^2 + K1p * K2p * H + K1p * K2p * K3p) * Pt
        PO4     <- (K1p * K2p * K3p) / (H^3 + K1p * H^2 + K1p * K2p * H + K1p * K2p * K3p) * Pt
        H2S     <- (H) / (H + Khs) * HSt
        HS      <- (Khs) / (H + Khs) * HSt
        SiOH4   <- (H^2) / (H^2 + Ksi * H + Ksi * K2si) * Sit
        SiOOH3  <- (Ksi * H) / (H^2 + Ksi * H + Ksi * K2si) * Sit
        SiO2OH2 <- (Ksi * K2si) / (H^2 + Ksi * H + Ksi * K2si) * Sit
        HF      <- (H) / (H + Kf) * FLUO
        F       <- (Kf) / (H + Kf) * FLUO
        HSO4    <- (H) / (H + Ks) * ST
        SO4     <- (Ks) / (H + Ks) * ST
        OH      <- Kw/H
        
        RES <- data.frame(flag, S, T, Patm, P, PH, CO2, fCO2, pCO2, fCO2pot, pCO2pot, fCO2insitu, pCO2insitu, HCO3, CO3, DIC, ALK, Oa, Oc,
                        NH4, NH3, BOH3, BOH4, H3PO4, H2PO4, HPO4, PO4, H2S, HS, SiOH4, SiOOH3, SiO2OH2, HF, F, HSO4, SO4, H, OH, NH4t, BOR, Pt, HSt, Sit, FLUO, ST)
        names(RES) <- c("flag", "S", "T", "Patm", "P", "pH", "CO2", "fCO2", "pCO2", "fCO2pot", "pCO2pot", "fCO2insitu", "pCO2insitu", "HCO3", "CO3", "DIC", "ALK", "OmegaAragonite", "OmegaCalcite",
                        "NH4", "NH3", "BOH3", "BOH4", "H3PO4", "H2PO4", "HPO4", "PO4", "H2S", "HS", "SiOH4", "SiOOH3", "SiO2OH2", "HF", "F", "HSO4", "SO4", "H", "OH", "NH4t", "BOR", "Pt", "HSt", "Sit",
                        "FLUO", "ST")
    }
    else {
        RES <- data.frame(flag, S, T, Patm, P, PH, CO2, fCO2, pCO2, fCO2pot, pCO2pot, fCO2insitu, pCO2insitu, HCO3, CO3, DIC, ALK, Oa, Oc)
        names(RES) <- c("flag", "S", "T", "Patm", "P", "pH", "CO2", "fCO2", "pCO2", "fCO2pot", "pCO2pot", "fCO2insitu", "pCO2insitu", "HCO3", "CO3", "DIC", "ALK", "OmegaAragonite", "OmegaCalcite")
    }
    return(RES)
}

# Copyright (C) 2009 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"amp" <- 
function(S=35,T=25){
	amp = (111.35+5.44875*S)*1/(T+273.15)+41.6775-0.015683*S-6.20815*log(T+273.15)-log10(1-0.00106*S)
	attr(amp,"unit") <- "mol/kg"
	return(amp)
}

# Copyright (C) 2015 James C. Orr
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"theta" <- 
function (S=35, T=25, P=0, Pref=0)

#--------------------------------------------------------------
# Potential temperature
#--------------------------------------------------------------

{
    nK <- max(length(S), length(T), length(P), length(Pref))

    ##-------- If any input argument is a vector, make vectors out of others having only a single entry

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(P)!=nK){Pref <- rep(Pref[1], nK)}

    #The 'oce' package's 'swTheta' routine is used to compute potential temperature

    #Convert pressure in bar (seacarb default) to decibar (oce swTheta default)
    pressure = 10*P
    referencePressure = 10*Pref

#   Compute potential temperature on ITS 90 scale (with routine that requires IPTS 68 scale)
#   - IPTS 68 = International Practical Temperature Scale of 1968
#   - ITS 90  = International Temperature Scale of 1990
#   (for T scale conversion: see Dickson et al., 2007, Chap. 5, p. 7, including footnote)
#   -------------------------------------------------------------------------------------
#   a) Compute "in-situ Temperature" (ITS 90) from "in situ Temperature" (IPTS 68)
    T68 <- (T - 0.0002) / 0.99975

#   b) Compute "potential temperature" (IPTS68) from "in situ temperature" (IPTS 68)
    theta68 <- swTheta(salinity=S, temperature=T68, pressure=pressure, referencePressure=referencePressure, "unesco")

#   c) Compute potential temp (ITS 90) from potential temp (IPTS 68)
    theta90 <- 0.99975*theta68 + 0.0002

#   Note: parts (a) and (c) above are tiny corrections (often neglected)
#         part  (b) is a big correction for deep waters (but zero at surface)

    attr(theta90,"unit")     = "degrees C (ITS 90)"
    return(theta90)
}  ## END theta

# Copyright (C) 2008 Karline Soetaert (K.Soetaert@nioo.knaw.nl) and Heloise Lavigne
# Revised by James Orr, 2012-01-17
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


#--------------------------------------------------------------
# Conversion factors for converting dissociation constants 
# from total pH scale to free pH scale (ktotal2free) 
# from total pH scale to seawater pH scale (ktotal2SWS)
# and from free pH scale to seawater scale (kfree2SWS)
# kfree = ktotal * ktotal2free
# kSWS  = ktotal * ktotal2SWS
# kSWS  = kfree  * kfree2SWS
# kfree = kSWS * kSWS2free
# ktotal = kfree * kfree2total
# ktotal = kSWS * kSWS2total
#--------------------------------------------------------------


"kconv" <- function (S=35,T=25,P=0,kf='x',Ks=0,Kff=0i,warn="y")

{
    #nK <- max(length(S), length(T), length(P), length(kf), length(Ks), length(Kff))
    #if(length(S)!=nK){S <- rep(S[1], nK)}
    #if(length(T)!=nK){T <- rep(T[1], nK)}
    #if(length(P)!=nK){P <- rep(P[1], nK)}
    #if(length(kf)!=nK){kf <- rep(kf[1], nK)}
    #if(length(Ks)!=nK){Ks <- rep(Ks[1], nK)}
    #if(length(Kff)!=nK){Kff <- rep(Kff[1], nK)}
    
    if ( missing(kf) && exists("kfg",envir=parent.frame()) )
       {
	 kf <- get("kfg", envir = parent.frame())
       }

	#--------------------------------------------------------------
    # CONVERT equilibrium constants to free scale:
	#--------------------------------------------------------------
	#------------------ Ks ----------------------------------------
	#       Dickson and Goyet (1994), Chapter 5, p.13
	#       (required for total2free)
	#       Equilibrium constant for HSO4- = H+ + SO4--
	#
	#       K_S  = [H+]free [SO4--] / [HSO4-]
	#       pH-scale: free scale !!!
	#
	#       the term log(1-0.001005*S) converts from
	#       mol/kg H2O to mol/kg soln
	
    # if Ks not given
    if (missing(Ks))
	{
	    Ks = Ks(S=S, T=T, P=P, warn=warn)                 # on free pH scale
	}
        Cl = S / 1.80655              # Cl = chlorinity; S = salinity (per mille)
        ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
	total2free  = 1/(1+ST/Ks)      # Kfree = Ktotal*total2free
	total2free <- as.numeric(total2free)	

	#---------------------------------------------------------------------
	# --------------------- Kf  ------------------------------------------
	#  Kf = [H+][F-]/[HF]
	#
	#   (Dickson and Riley, 1979 in Dickson and Goyet,
	#   1994, Chapter 5, p. 14)
	#   pH-scale: 'free'

    # if Kf on free pH scale is not given as an argument 
	if (missing(Kff))
	{
	    Kff = Kf(S=S, T=T, P=P, kf=kf, pHscale="F", warn=warn)
	}


	#------- sws2free -----------------------------------------------
	#
	#       convert from pH_sws ('seawater scale`) to pH ('free`):
	#       pH_sws = pH_free - log(1+S_T/K_S(S,T)+F_T/K_F(S,T))
        Cl = S / 1.80655            # Cl = chlorinity; S = salinity (per mille)
        FT = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
	free2SWS  = 1+ST/Ks+FT/Kff         # Kfree = Ksws*sws2free
	free2SWS <- as.numeric(free2SWS)
	total2SWS = total2free * free2SWS # KSWS = Ktotal*total2SWS
	total2SWS <- as.numeric(total2SWS)

    return (list(ktotal2SWS=total2SWS, ktotal2free=total2free,kfree2SWS=free2SWS, kfree2total=(1/total2free), kSWS2total=(1/total2SWS), kSWS2free=(1/free2SWS)))
}

# Copyright (C) 2024 James Orr
# This file is part of seacarb.
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Convert pHSWS to pHNBS using the total hydrogen ion activity coefficient, which depends on T and S

pHsws2nbs <- function(pHSWS, S=35, T=25){
    # --------------------------------------------------------------------------------------------------------------
    # phSWS = -log10([H+]sws)  #pH on the SWS scale  (-log10 of the concentration)
    # pHNBS = -log10(ah)       #pH on the NBS scale  (-log10 of the activity, not the concentration)
    # 
    # Note:
    # -----
    # ah = fH * [H+]sws
    # where ah  is the activity of hydrogen ion,
    #       fH  is the total activity coefficient, and
    #       [H+]sws = [H+] + [HSO4-] + [HF], or in other words "the hydrogen ion conccentration on the seawater scale"
    # --------------------------------------------------------------------------------------------------------------
    #
    # The SWS-to-NBS conversion is done with the total activity coefficient fH (combined activity coeff for H+, HSO4-, and HF)
    # from Takahashi (1982) based on data from Culberson and Pytkowicz (1973). The approach is old and full of uncertainty.
    # Newer approaches are more complicated (Pitzer equations) and big uncertainties remain (Marion et al., 2011; Pilson, 2013).
    #
    # Culberson, CH, & Pytkowicz, RM (1973). Ionization of water in seawater. Marine Chemistry, 1(4), 309-316.
    #
    # Marion GM, Millero FJ, Camoes MF, Spitzer P, Feistel R, Chen CTA. 2011. pH of seawater. Marine Chemistry 126: 89-96
    #
    # Pilson MEQ. 2013. An introduction to the chemistry of the sea, 2 edn. Cambridge, UK: Cambridge University Press.
    #
    # Takahashi, T. et al (1982). Carbonate chemistry. GEOSECS Pacific Expedition, Volume 3, Hydrographic Data 1973-1974, 77-83.

    hSWS  <- 10^(-pHSWS)
    ah    <-  fH(S=S, T=T) * hSWS
    pHNBS <- -log10(ah)

return(pHNBS)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Heloise Lavigne and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"Ks" <-
function(S=35,T=25,P=0, ks="d",warn="y"){

    nK <- max(length(S), length(T), length(P), length(ks))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(ks)!=nK){ks <- rep(ks[1], nK)}

    #-------Constants----------------

    #---- issues de equic----
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK = T + tk;           # T [C]; TK [K]
    Cl = S / 1.80655;      # Cl = chlorinity; S = salinity (per mille)
    iom0 = 19.924*S/(1000-1.005*S);
    ST = 0.14 * Cl/96.062  # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)


	#--------------------------------------------------------------
	#------------------ Ks ----------------------------------------
	#       Dickson (1990)
	#     
	#       Equilibrium constant for HSO4- = H+ + SO4--
	#
	#       K_S  = [H+]free [SO4--] / [HSO4-]
	#       pH-scale: free scale !!!
	#
	#       the term log(1-0.001005*S) converts from
	#       mol/kg H2O to mol/kg soln
	
	
	tmp1 = -4276.1 / TK + 141.328 -23.093*log(TK);
	tmp2 = +(-13856 / TK + 324.57 - 47.986 * log(TK))*sqrt(iom0);
	tmp3 = +(35474 / TK - 771.54 + 114.723 * log(TK))*iom0;
	tmp4 = -2698 / TK *(sqrt(iom0)*iom0) + 1776 / TK *iom0 *iom0;
	                                       
	
	lnKs = tmp1 + tmp2 + tmp3 + tmp4 + log(1-0.001005*S);
		
	Ks_d = exp(lnKs);

	#--------------------------------------------------------------
	#------------------ Ks ----------------------------------------
	#       Khoo et al. 1977
	#     
	#       Equilibrium constant for HSO4- = H+ + SO4--
	#
	#       K_S  = [H+]free [SO4--] / [HSO4-]
	#       pH-scale: free scale !!!
	#        
	#      correct for T range : 5 - 40°C
	#      correct for S range : 20 - 45

    I35 <- 0.7227
    I <- I35*(27.570*S)/(1000-1.0016*S)   #formal ionic strength
    logbeta <- 647.59/TK - 6.3451 + 0.019085*TK - 0.5208*I^(0.5)
    beta <- 10^(logbeta)
    Ks_k <- 1/beta


    #-------------------------------------------------------------------
    #--------------- choice between the formulations -------------------
    is_k <- (ks=='k')    # everything that is not "k" is "d" by default
    Ks <- Ks_d
    Ks[is_k] <- Ks_k[is_k]

    method <- rep(NA, nK)
    method[!is_k] <- "Dickson (1990)"
    method[is_k] <- "Khoo et al. (1977)"
    
    # ------------------- Pressure effect --------------------------------

    if (any(P != 0))
        Ks <- Pcorrect(Kvalue=Ks, Ktype="Ks", T=T, S=S, P=P, pHscale="F", warn=warn)

    ##------------Warnings

    is_w <- warn == "y"
    
    if (any (is_w & is_k & (T<5 | T>40 | S<20 | S>45)))
        {warning("S and/or T is outside the range of validity of the formulation chosen for Ks.")}
	
    if (any (is_w & (T>45 | S>45 | T<0 | S<5)))
        {warning("S and/or T is outside the range of validity of the formulations available for Ks in seacarb.")}

    ##------------Attributes

    attr(Ks,"unit") = "mol/kg-soln"	
    attr(Ks,"pH scale") = "free scale"
    attr(Ks, "method") = method
    return(Ks)
}

# Copyright (C) 2007 Karline Soetaert (K.Soetaert@nioo.knaw.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"Khs"              <- function (S=35,T=25,P=0,pHscale="T",warn="y")

#--------------------------------------------------------------
# Dissociation constant of H2S
#--------------------------------------------------------------

{

nK <- max(length(S), length(T), length(P), length(pHscale))

##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

if(length(S)!=nK){S <- rep(S[1], nK)}
if(length(T)!=nK){T <- rep(T[1], nK)}
if(length(P)!=nK){P <- rep(P[1], nK)}
if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}


tk = 273.15           # [K] (for conversion [deg C] <-> [K])
TC = T + tk           # TC [C]; T[K]

	#--------------------------------------------------------------
  # Dissociation constant of H2S on total scale - Millero 1995
 	#--------------------------------------------------------------
lnK <- 225.838 + 0.3449*sqrt(S) - 0.0274*S -13275.3/TC  -34.6435*log(TC)

Khs      <- exp(lnK)

# ---- Conversion from Total scale to seawater scale before pressure corrections
factor <- kconv(S=S, T=T, P=rep(0,nK))$ktotal2SWS
Khs <- Khs * factor

# ----------------- Pressure Correction ------------------	
Khs <- Pcorrect(Kvalue=Khs, Ktype="Khs", T=T, S=S, P=P, pHscale="SWS", warn=warn)


###----------------pH scale corrections
factor <- rep(NA,nK)
pHsc <- rep(NA,nK)
for(i in (1:nK)){   
 if(pHscale[i]=="T"){factor[i] <- kconv(S=S[i], T=T[i], P=P[i])$kSWS2total ; pHsc[i] <- "total scale"}
 if(pHscale[i]=="F"){factor[i] <- kconv(S=S[i], T=T[i], P=P[i])$kSWS2free ; pHsc[i] <- "free scale"}
 if(pHscale[i]=="SWS"){factor[i] <- 1 ; pHsc[i] <- "seawater scale"}
Khs[i] <- Khs[i]*factor[i]
}

##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>45 | T<0 | S<0 | S>45)) ) {warning("S and/or T is outside the range of validity of the formulation chosen for Khs.")}

  attr(Khs,"unit")     = "mol/kg-soln"
  attr(Khs,"pH scale") = pHsc
  return(Khs)

}  ## END Khs

# Copyright (C) 2003 Jean-Pierre Gattuso and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"Kspc" <-
function(S=35,T=25,P=0,warn="y"){
    
    nK <- max(length(S), length(T), length(P))
    
    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)
    
    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    
    
    #---- issues de equic----
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TC = T + tk;           # TC [C]; T[K]

    # --------------------- Kspc (calcite) ----------------------------
    #
    # apparent solubility product of calcite
    #
    #  Kspc = [Ca2+]T [CO32-]T
    #
    #  where $[]_T$ refers to the equilibrium total 
    # (free + complexed) ion concentration.
    #
    #  Mucci 1983 mol/kg-soln
    
    tmp1 = -171.9065-0.077993*TC+2839.319/TC+71.595*log10(TC);
    tmp2 = +(-0.77712+0.0028426*TC+178.34/TC)*sqrt(S);
    tmp3 = -0.07711*S+0.0041249*S^1.5;
    log10Kspc = tmp1 + tmp2 + tmp3;
    
    Kspc = 10^(log10Kspc);
    	
    # ----------------- Pressure Correction ------------------
    	
    Kspc <- Pcorrect(Kvalue=Kspc, Ktype="Kspc", T=T, S=S, P=P, warn=warn)

		
		##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>40 | S>44 | T<5 | S<5)) )
        warning("S and/or T is outside the range of validity of the formulation available for Kspc in seacarb.")

    
    attr(Kspc,"unit") = "mol/kg"
    return(Kspc)
}

# Copyright (C) 2003 Jean-Pierre Gattuso and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"rho" <-
function(S=35,T=25,P=0){	

#       Convert temperature (C) on today's "ITS 90" scale to older "IPTS 68" scale
#       (see Dickson et al., Best Practices Guide, 2007, Chap. 5, p. 7, including footnote)
#       IPTS 68 = International Practical Temperature Scale of 1968
#       ITS 90  = International Temperature Scale of 1990
#       Note: Density calculation below requires temperature on the IPTS 68 scale
        T68 <- (T - 0.0002) / 0.99975

	#------------ Density of pure water
	rhow = 999.842594 + 6.793952e-2*T68 -9.095290e-3*T68^2 + 1.001685e-4*T68^3 -1.120083e-6*T68^4 + 6.536332e-9*T68^5;
	
	#------------ Density of seawater at 1 atm, P=0
	A = 8.24493e-1 - 4.0899e-3*T68 + 7.6438e-5*T68^2 - 8.2467e-7*T68^3 + 5.3875e-9*T68^4;
	B = -5.72466e-3 + 1.0227e-4*T68 - 1.6546e-6*T68^2; 
	C = 4.8314e-4;   
	
	rho0 = rhow + A*S + B*S^(3/2) + C*S^2;

	#-------------- Secant bulk modulus of pure water 
	#
	# The secant bulk modulus is the average change in pressure 
	# divided by the total change in volume per unit of initial volume.
	Ksbmw = 19652.21 + 148.4206*T68 - 2.327105*T68^2 + 1.360477e-2*T68^3 - 5.155288e-5*T68^4
	
	#-------------- Secant bulk modulus of seawater at 1 atm
	Ksbm0 = Ksbmw + S*( 54.6746 - 0.603459*T68 + 1.09987e-2*T68^2 - 6.1670e-5*T68^3) + S^(3/2)*( 7.944e-2 + 1.6483e-2*T68 - 5.3009e-4*T68^2);
	
	#-------------- Secant bulk modulus of seawater at S,T,P		
	Ksbm = Ksbm0 + P*( 3.239908 + 1.43713e-3*T68 + 1.16092e-4*T68^2 - 5.77905e-7*T68^3) + P*S*( 2.2838e-3 - 1.0981e-5*T68 - 1.6078e-6*T68^2) + P*S^(3/2)*1.91075e-4 + P*P*(8.50935e-5 - 6.12293e-6*T68 + 5.2787e-8*T68^2) + P^2*S*(-9.9348e-7 + 2.0816e-8*T68 + 9.1697e-10*T68^2);
	 	
	#------------- Density of seawater at S,T,P
	rho = rho0/(1-P/Ksbm);
	attr(rho,"unit") = "(kg/m3)"
    return(rho)
	}

# Copyright (C) 2003 Jean-Pierre Gattuso and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"pHinsi" <- function(pH=8.2, ALK=2.4e-3, Tinsi=20, Tlab=25, Pinsi=0, S=35, Pt = 0, Sit = 0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", eos="eos80", long=1.e20, lat=1.e20) {
  # if the concentrations of total silicate and total phosphate are NA
  # they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  
  # Only two options for eos
  if (eos != "teos10" && eos != "eos80")
      stop ("invalid parameter eos: ", eos)
  
  # if use of EOS-10 standard
  if (eos == "teos10")
  {
      # Must convert salinity from TEOS-10 to EOS-80
      # convert salinity from Absolute to Practical (SP)
      STeos <- teos2eos_geo (S, Tlab, P=0, long, lat)
      SP <- STeos$SP
  }
  else
  {
      SP <- S
  }

  # according to Hunter (1998)
	#si TA=0 alors calculer TA: TA=660+47.6 * S
	dat1 <- carb(flag = 8, var1 = pH, var2 = ALK, S = SP, T = Tlab, P = 0, Pt = Pt, Sit = Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b)
	DIC <- dat1$DIC
	dat2 <- carb(flag = 15, var1 = ALK, var2 = DIC, S = SP, T = Tinsi, P = Pinsi, Pt = Pt, Sit = Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b)
	ph_insi <- dat2$pH
	#utiliser DIC et TA pour calculer pH in situ (flag=15)
	#attr(bor,"unit") <- "mol/kg"
	return(ph_insi)
}

# Copyright (C) 2010  Héloïse Lavigne and Jean-Pierre Gattuso
# with a most valuable contribution of Bernard Gentili <gentili@obs-vlfr.fr>
# and valuable suggestions from Jean-Marie Epitalon <epitalon@lsce.saclay.cea.fr>
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
f2pCO2 <- function(T=25, Patm=1, P=0, fCO2){
tk <- 273.15;           # [K] (for conversion [deg C] <-> [K])
TK <- T + tk;           # TK [K]; T[C]
Phydro_atm = P / 1.01325  # convert hydrostatic pressure from bar to atm (1.01325 bar / atm)
Ptot = Patm + Phydro_atm  # total pressure (in atm) = atmospheric pressure + hydrostatic pressure

# Original "seacarb" f2pCO2 calculation:
# B <- (-1636.75+12.0408*TK-0.0327957*(TK*TK)+0.0000316528*(TK*TK*TK))*1e-6
# pCO2 <-  fCO2*(1/exp((1*100000)*(B+2*(57.7-0.118*TK)*1e-6)/(8.314*TK)))
# Above calculation:
# - uses incorrect R (wrong units, incompatible with pressure in atm)
# - neglects a term "x2" (see below)
# - assumes pressure is always 1 atm (wrong for subsurface)

# Corrected approach to compute the fugacity coefficient, with which we convert pCO2 <-> fCO2
  Phydro_atm = P / 1.01325  # convert hydrostatic pressure from bar to atm (1.01325 bar / atm)
  Ptot = Patm + Phydro_atm  # total pressure (in atm)
  R = 82.05736              # (cm3 * atm) / (mol * K)  CODATA (2006)

# To compute fugcoeff, we need 3 other terms (B, Del, xc2) in addition to 3 others above (TK, Ptot, R)
  B   <- -1636.75 + 12.0408*TK - 0.0327957*TK^2 + 0.0000316528*TK^3
  Del <- 57.7-0.118*TK

# "x2" term often neglected (assumed = 1) in applications of Weiss's (1974) equation 9
# x2 = 1 - x1 = 1 - xCO2 (it is close to 1, but not quite)
# Let's assume that xCO2 = fCO2. Resulting fugcoeff is identical to 8th digit after the decimal.
  xCO2approx <- fCO2
  xc2 <- (1 - xCO2approx*1e-6)^2 

  fugcoeff = exp( Ptot*(B + 2*xc2*Del)/(82.057*TK) )
  pCO2 <- fCO2 / fugcoeff

return(pCO2)
}

# Copyright (C) 2003 Jean-Pierre Gattuso and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"Kspa" <-
function(S=35,T=25,P=0,warn="y"){

nK <- max(length(S), length(T), length(P))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)
    
    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    
    	
    #---- issues de equic----
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TC = T + tk;           # TC [C]; T[K]
    
    # --------------------- Kspa (aragonite) ----------------------------
    #
    # apparent solubility product of aragonite
    #
    #  Kspa = [Ca2+]T [CO32-]T
    #
    #  where $[]_T$ refers to the equilibrium total 
    # (free + complexed) ion concentration.
    #
    #  Mucci 1983 mol/kg-soln
    
    tmp1 = -171.945-0.077993*TC+2903.293/TC+71.595*log10(TC);
    tmp2 = +(-0.068393+0.0017276*TC+88.135/TC)*sqrt(S);
    tmp3 = -0.10018*S+0.0059415*S^1.5;
    log10Kspa = tmp1 + tmp2 + tmp3;
    
    Kspa = 10^(log10Kspa);

    # ----------------- Pressure Correction ------------------
    	
    Kspa <- Pcorrect(Kvalue=Kspa, Ktype="Kspa", T=T, S=S, P=P, warn=warn)
		
    ##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>40 | S>44 | T<5 | S<5)) )
        warning("S and/or T is outside the range of validity of the formulation available for Kspa in seacarb.")
		
    attr(Kspa,"unit") = "mol/kg"
    return(Kspa)
}

# Copyright (C) 2016 Mathilde Hagens (M.Hagens@uu.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

buffergen <- function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, k1k2="x", kf="x", ks="d", pHscale="T", b="u74", 
                      gas="potential", NH4t=0, HSt=0) {
  
  n <- max(length(flag), length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), 
           length(ks), length(b), length(gas), length(NH4t), length(HSt))
  if(length(flag)!=n){flag <- rep(flag[1],n)}
  if(length(var1)!=n){var1 <- rep(var1[1],n)}
  if(length(var2)!=n){var2 <- rep(var2[1],n)}
  if(length(S)!=n){S <- rep(S[1],n)}
  if(length(T)!=n){T <- rep(T[1],n)}
  if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
  if(length(P)!=n){P <- rep(P[1],n)}
  if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
  if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
  if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
  if(length(kf)!=n){kf <- rep(kf[1],n)}
  if(length(ks)!=n){ks <- rep(ks[1],n)}
  if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
  if(length(b)!=n){b <- rep(b[1],n)}
  if(length(gas)!=n){gas <- rep(gas[1],n)}
  if(length(NH4t)!=n){NH4t <- rep(NH4t[1],n)}
  if(length(HSt)!=n){HSt <- rep(HSt[1],n)}
  
  # if the concentrations of total silicate, total phosphate, total nitrite,
  # total ammonium, and total hydrogen sulfide are NA, they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  NH4t[is.na(NH4t)] <- 0
  HSt[is.na(HSt)] <- 0
  
  # Calculate acid-base speciation
  Carbfull <- carbfull(flag=flag, var1=var1, var2=var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, gas=gas, NH4t=NH4t, HSt=HSt)

  #---------------------------------------------------------------------
  #--------------------    buffer effects    ---------------------------
  #---------------------------------------------------------------------
  
  # Define a function to calculate the buffer factors
  bufferfunc <- function(Carbfull, species=species){
  with(as.list(Carbfull),{

    # Calculation of acid-base fractions
    n1  <- NH4 / NH4t
    n2  <- NH3 / NH4t
    b1  <- BOH3 / BOR
    b2  <- BOH4 / BOR
    c1  <- CO2 / DIC
    c2  <- HCO3 / DIC
    c3  <- CO3 / DIC
    p1  <- H3PO4 / Pt
    p2  <- H2PO4 / Pt
    p3  <- HPO4 / Pt
    p4  <- PO4 / Pt
    s1  <- H2S / HSt
    s2  <- HS / HSt
    si1 <- SiOH4 / Sit
    si2 <- SiOOH3 / Sit
    si3 <- SiO2OH2 / Sit
    f1  <- HF / FLUO
    f2  <- F / FLUO
    so1 <- HSO4 / ST
    so2 <- SO4 / ST
    
    # Ceation of output vectors
    dALK.dH <- dtotX.dH <- dALK.dX <- dtotX.dX <- dALK.dpH <- dtotX.dpH <- 
      dH.dALK <- dX.dALK <- dX.dtotX <- dpH.dALK <- dH.dtotX <- dpH.dtotX <-
      rep(NA, length.out = length(species))
    
    # Assign names to output vectors
    names(dALK.dH) <- names(dtotX.dH) <- names(dALK.dX) <- 
      names(dtotX.dX) <- names(dALK.dpH) <- names(dtotX.dpH) <-
      names(dH.dALK) <- names(dX.dALK) <- names(dX.dtotX) <-
      names(dpH.dALK) <- names(dH.dtotX) <- names(dpH.dtotX) <- species
    
    # Define function for calculating ALK.X
    ALK.species <- function(X.name)
    {
      if(X.name == "CO2" || X.name == "HCO3" || X.name == "CO3" || 
         X.name == "DIC") {HCO3 + 2*CO3} else
           if(X.name == "NH3" || X.name == "NH4" || X.name == "NH4t") {
             NH3} else
               if(X.name == "H2PO4" || X.name == "H3PO4" || 
                  X.name == "HPO4" || X.name == "PO4" || 
                  X.name == "Pt") {-H3PO4 + HPO4 + 2*PO4} else
                    if(X.name == "H2S" || X.name == "HS" || X.name == "HSt") {HS} else
                        if(X.name == "SiOH4" || X.name == "SiOOH3" || 
                           X.name == "SiO2OH2" || X.name == "Sit") {
                          SiOOH3 + 2*SiO2OH2} else
                            if(X.name == "BOH3" || X.name == "BOH4" || X.name == "BOR") {
                              BOH4} else
                                 if(X.name == "F" || X.name == "HF" || X.name == "FLUO") {
                                   -HF} else
                                     if(X.name == "SO4" || X.name == "HSO4" || 
                                        X.name == "ST") {-HSO4} else
                                         NULL
    }
    
    # Define function for determining N
    N.function <- function(X.name)
    {
      if(X.name == "CO2" || X.name == "BOH3" || X.name == "NH4" || 
         X.name == "H2PO4" || X.name == "H2S" || X.name == "SiOH4" || X.name == "F" || 
         X.name == "SO4" || X.name == "DIC" || X.name == "BOR" || 
         X.name == "NH4t" || X.name == "Pt" || 
         X.name == "HSt" || X.name == "Sit" || 
         X.name == "FLUO" || X.name == "ST") {N <- 0} else
           
           if(X.name == "HCO3" || X.name == "NH3" || 
              X.name == "HPO4" || X.name == "HS" ||
              X.name == "SiOOH3" || X.name == "BOH4") {N <- 1} else
                
                if(X.name == "CO3" || X.name == "PO4" || X.name == "SiO2OH2") {N <- 2} else
                     
                     if(X.name == "H3PO4" || X.name == "HF" || X.name == "HSO4") {
                       n <- -1} else
                         
                         NULL
    }
    
    # Define function for calculating dALKdH  
    dALKdH.function <- function(X.name)
    {
      N <- N.function(X.name)
      ALK.X <- ALK.species(X.name)
      if(DIC!=0) {
        if(X.name == "CO2" || X.name == "HCO3" || X.name == "CO3" || 
           X.name == "DIC") {
          dALKdH.CO2 <- (-1/H)*(-N*ALK.X + HCO3 + 4*CO3)}  else {
            dALKdH.CO2 <- (-1/H)*(HCO3*(c1-c3) + 2*CO3*(2*c1+c2))}
      } else dALKdH.CO2 <- 0
      if(NH4t!=0) {
        if(X.name == "NH3" || X.name == "NH4" || X.name == "NH4t") {
          dALKdH.NH4 <- (-1/H)*(-N*ALK.X + NH3)} else {
            dALKdH.NH4 <- (-1/H)*(NH3*n1)}
      } else dALKdH.NH4 <- 0
      if(Pt!=0) {
        if(X.name == "H2PO4" || X.name == "H3PO4" || X.name == "HPO4" || 
           X.name == "PO4" || X.name == "Pt") {
          dALKdH.H2PO4 <- (-1/H)*(-N*ALK.X + H3PO4 + HPO4 + 4*PO4)}  else {
            dALKdH.H2PO4 <- (-1/H)*(-H3PO4*(-p2-2*p3-3*p4) + 
                                     HPO4*(2*p1+p2-p4) + 2*PO4*(3*p1+2*p2+p3))}
      } else dALKdH.H2PO4 <- 0
      if(HSt!=0) {
        if(X.name == "H2S" || X.name == "HS" || X.name == "HSt") {
          dALKdH.H2S <- (-1/H)*(-N*ALK.X + HS)}  else {
            dALKdH.H2S <- (-1/H)*(HS*s1)} 
      } else dALKdH.H2S <- 0
      if(Sit!=0) {
        if(X.name == "SiOH4" || X.name == "SiOOH3" || 
           X.name == "SiO2OH2" || X.name == "Sit") {
          dALKdH.SiOH4 <- (-1/H)*(-N*ALK.X + SiOOH3 + 4*SiO2OH2)} else {
            dALKdH.SiOH4 <- (-1/H)*(SiOOH3*(si1-si3) + 
                                     2*SiO2OH2*(2*si1+si2))} 
      } else dALKdH.SiOH4 <- 0
      if(BOR!=0) {
        if(X.name == "BOH3" || X.name == "BOH4" || X.name == "BOR") {
          dALKdH.BOH3 <- (-1/H)*(-N*ALK.X + BOH4)} else {
            dALKdH.BOH3 <-(-1/H)*(BOH4*b1)}
      } else dALKdH.BOH3 <- 0
      if(FLUO!=0) {
        if(X.name == "F" || X.name == "HF" || X.name == "FLUO") {
          dALKdH.F <- (-1/H)*(-N*ALK.X + HF)} else {
            dALKdH.F <- (-1/H)*(-HF*f2)}
      } else dALKdH.F <- 0  
      if(ST!=0) {
        if(X.name == "SO4" || X.name == "HSO4" || X.name == "ST") {
          dALKdH.SO4 <- (-1/H)*(-N*ALK.X + HSO4)} else {
            dALKdH.SO4 <- (-1/H)*(-HSO4*so2)}
      } else dALKdH.SO4 <- 0  
      dALKdH.H <- (-1/H)*(OH+H)       # Internal enhancement of buffering
      return(dALKdH.CO2 + dALKdH.NH4 + dALKdH.H2PO4 + dALKdH.H2S + dALKdH.SiOH4 + dALKdH.BOH3 + 
               dALKdH.F + dALKdH.SO4 + dALKdH.H)
    }

    # Define function for determining totX
    totX.function <- function(X.name)
    {
      if(X.name == "CO2" || X.name == "HCO3" || X.name == "CO3" || 
         X.name == "DIC") {DIC} else
           if(X.name == "NH3" || X.name == "NH4" || X.name == "NH4t") {
             NH4t} else
               if(X.name == "H2PO4" || X.name == "H3PO4" || 
                  X.name == "HPO4" || X.name == "PO4" || 
                  X.name == "Pt") {Pt} else
                    if(X.name == "H2S" || X.name == "HS" || X.name == "HSt") {HSt} else
                      if(X.name == "SiOH4" || X.name == "SiOOH3" || X.name == "SiO2OH2" || 
                         X.name == "Sit") {Sit} else
                           if(X.name == "BOH3" || X.name == "BOH4" || X.name == "BOR") {
                             BOR} else
                               if(X.name == "F" || X.name == "HF" || X.name == "FLUO") {
                                 FLUO} else
                                   if(X.name == "SO4" || X.name == "HSO4" || 
                                      X.name == "ST") {ST} else
                                        NULL
    }
    
    # Define function to calculate beta.H
    beta.H_func <- function()
    {
      if(DIC!=0) {dALKdH.CO2 <- (-1/H)*(HCO3*(c1-c3) + 2*CO3*(2*c1+c2))} else {
        dALKdH.CO2 <- 0}
      if(NH4t!=0) {dALKdH.NH4 <- (-1/H)*(NH3*n1)} else {dALKdH.NH4 <- 0}
      if(Pt!=0) {dALKdH.H2PO4 <- (-1/H)*(-H3PO4*(-p2-2*p3-3*p4) + 
                                        HPO4*(2*p1+p2-p4) + 2*PO4*(3*p1+2*p2+p3))} else {
        dALKdH.H2PO4 <- 0}
      if(HSt!=0) {dALKdH.H2S <- (-1/H)*(HS*s1)} else {dALKdH.H2S <- 0}
      if(Sit!=0) {dALKdH.SiOH4 <- (-1/H)*(SiOOH3*(si1-si3) + 
                                         2*SiO2OH2*(2*si1+si2))} else {dALKdH.SiOH4 <- 0}
      if(BOR!=0) {dALKdH.BOH3 <-(-1/H)*(BOH4*b1)} else {dALKdH.BOH3 <- 0}
      if(FLUO!=0) {dALKdH.F <- (-1/H)*(-HF*f2)} else {dALKdH.F <- 0}  
      if(ST!=0) {dALKdH.SO4 <- (-1/H)*(-HSO4*so2)} else {dALKdH.SO4 <- 0}  
      dALKdH.H <- (-1/H)*(OH+H)       # Internal enhancement of buffering
      return(dALKdH.CO2 + dALKdH.NH4 + dALKdH.H2PO4 + dALKdH.H2S + dALKdH.SiOH4 + dALKdH.BOH3 + 
               dALKdH.F + dALKdH.SO4 + dALKdH.H)
    }
    
    # Define a function to calculate the Revelle factor 
    Revelle_func <- function()
    {
      X.name   <- "CO2"
      X        <- get(X.name)
      N        <- N.function(X.name)
      ALK.X    <- ALK.species(X.name)
      totX     <- totX.function(X.name)
      
      dALK.dH   <- dALKdH.function(X.name)
      dX.dtotX <- X * H * dALK.dH / (ALK.X^2 + (H*dALK.dH - N*ALK.X)*totX)
      RF <- as.numeric(DIC / CO2 * dX.dtotX)
      
      return(RF)
    }
    
    # Loop that calculates output species per variable
    for (i in 1:length(species))
    {
      
      # In case a change in the total concentration is specified, 
      # change X to the reference species for ALK
      X.name <- species[i]
      
      if(X.name == "DIC") X <- get("CO2") else                      
        if(X.name == "NH4t") X <- get("NH4") else                     
          if(X.name == "Pt") X <- get("H2PO4") else
            if(X.name == "HSt") X <- get("H2S") else 
              if(X.name == "Sit") X <- get("SiOH4") else
                if(X.name == "BOR") X <- get("BOH3") else
                  if(X.name == "FLUO") X <- get("F") else
                    if(X.name == "ST") X <- get("SO4") else
                      X <- get(species[i])
                        
      N              <- N.function(X.name)
      ALK.X          <- ALK.species(X.name)
      totX           <- totX.function(X.name)
      
      if(totX!=0) {
      # Calculation of factors in Table 3a
      dALK.dH[i]     <- dALKdH.function(X.name)
      dtotX.dH[i]    <- (N*totX - ALK.X) / H
      dALK.dX[i]     <- ALK.X / X
      dtotX.dX[i]    <- totX / X
      
      # Transformation from dH to dpH
      dH.dpH         <- -log(10)*H
      
      dALK.dpH[i]    <- dH.dpH * dALK.dH[i]
      dtotX.dpH[i]   <- dH.dpH * dtotX.dH[i]
      
      # Calculation of factors in Table 3b
      dH.dALK[i]     <- totX*H  / (ALK.X^2 + (H*dALK.dH[i] - N*ALK.X)*totX)
      dH.dtotX[i]    <- -ALK.X*H / (ALK.X^2 + (H*dALK.dH[i] - N*ALK.X)*totX)
      dX.dALK[i]     <- -X * (N*totX - ALK.X)  / (ALK.X^2 + (H*dALK.dH[i] - 
                                                             N*ALK.X)*totX)
      dX.dtotX[i]    <- X * H * dALK.dH[i] / (ALK.X^2 + (H*dALK.dH[i] - 
                                                         N*ALK.X)*totX)
      
      # Transformation from dH to dpH
      dpH.dH         <- 1/(-log(10)*H)
      
      dpH.dALK[i]     <- dpH.dH * dH.dALK[i]
      dpH.dtotX[i]   <- dpH.dH * dH.dtotX[i]
      }
    }  
    
    # Calculate the Revelle factor
    RF <- Revelle_func()
    
    # Calculate beta.H
    beta.H <- as.numeric(beta.H_func())
      
    # Combine the results in a list
    result <- list(Carbfull=as.matrix(Carbfull), dALK.dH=dALK.dH, dtotX.dH=dtotX.dH, dALK.dX=dALK.dX, 
                   dtotX.dX=dtotX.dX, dALK.dpH=dALK.dpH, dtotX.dpH=dtotX.dpH, dH.dALK=dH.dALK, 
                   dH.dtotX=dH.dtotX, dX.dALK=dX.dALK, dX.dtotX=dX.dtotX, dpH.dALK=dpH.dALK, 
                   dpH.dtotX=dpH.dtotX, beta.H=beta.H, RF=RF)
    
    # Return result
    return(result)
  })
}
  
  # Define all species for which bufferfunc needs to be calculated
  species <- names(Carbfull)[c(7,14:15,20:36,39:40,16,41:45)]
  
  # Create an empty output list
  output <- list()
  
  # Loop buffer function over results of Carbfull
  for(j in 1:dim(Carbfull)[1]){
  
    # Run function
    result <- bufferfunc(Carbfull[j,], species)
  
    # Transfer the result to an output
    for(k in 1:length(result)){
      if(j==1) {
        output[[k]] <- result[[k]]
        names(output)[k] <- names(result)[k]} else {
          output[[k]] <- rbind(output[[k]],result[[k]])}
    }
    }
  
  # Return statement
  return(output)
  
}

# Copyright (C) 2009 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"pH" <- 
function(Ex=-67,Etris=-72.4,S=35,T=25){
	R  =  8.31447215 # J K–1 mol–1, Dickson et al. (2007)
	F  =  96485.339924 # C mol–1, Dickson et al. (2007)
	pH = tris(S=S, T=T) + (Etris/1000-Ex/1000)/(R*(T + 273.15)*log(10)/F)
	attr(pH,"unit") <- "mol/kg"
	attr(pH, "pH scale") <- "total scale"
	return(pH)
}

# Copyright (C) 2008 Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye and Jean-Pierre Gattuso  
# with a most valuable contribution of Bernard Gentili
#
# This file is part of seacarb.
#/
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
sp2sa_chem <- function (SP, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
#
# This function converts from Practical (SP) to Absolute Salinity (SA)
# based on Total Alkalinity, Dissolved Inorganic Carbon 
# and Nitrate and Silicate concentrations.
#
# It is made of two parts :
#   1) conversion from Practical to Reference Salinity (major part)
#   2) conversion from Reference to Absolute Salinity (Absolute Anomaly, minor part)
#
{
    # DIC and TA values of standard sea water
    TA_stdsw  <- 0.0023  * SP /35
    DIC_stdsw <- 0.00208 * SP /35
    
    # Absolute salinity anomaly (g/kg)
    sal_anomaly = 55.6 * (TA - TA_stdsw) + 4.7 * (DIC - DIC_stdsw) + 38.9 * NO3 + 50.7 * SIOH4
    
    SR <- SP * 35.16504 / 35
    SA <- SR + sal_anomaly
    return(SA)
}


sp2sa_geo <- function (SP, P=0, long=1.e20, lat=1.e20)
#
# This function converts from Practical (SP) to Absolute Salinity (SA)
# based on Pressure P in dbar and geographic location : 
# latitude in decimal degrees [-90 ... 90] 
# and longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]
#
# Latitude and longitude are optional. 
# If any is omitted, an arbitrary geographic point is chosen : central equatorial Atlantic.
# Note that this implies an error on computed SA ranging from 0 up to 0.02 g/kg
{
    if ((length(lat) == 1 && lat == 1.e20) || (length(long) == 1 && long == 1.e20))
    {
        lat <- 0; long <- -25;
    }
    return (gsw_SA_from_SP(SP,P,long,lat))
}


sa2sp_chem <- function (SA, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
#
# This function does the reverse of above function SA_from_SP_chem()
# It converts from Absolute (SA) to Practical Salinity (SP)
# based on chemical composition of sea water : Total Alkalinity, Dissolved Inorganic Carbon 
# and Nitrate and Silicate concentrations.
#
# Reverse conversion (SA from SP) follows this equation :
#  SP = 1/r * (SA - sa_anomaly)
#    with
#        r = 35.16504 / 35
#  SP = 1/r * (SA - 55.6 * (TA - x*SP) - 4.7 * (DIC - y*SP) - nutrients)
#    with
#       x  = 0.00230 / 35
#       y  = 0.00208 / 35
#       nutrients = 38.9 * NO3 + 50.7 * SIOH4
#
#  SP * (1 - 55.6 * x/r - 4.7 * y/r) = 1/r * (SA - 55.6 * TA - 4.7 * DIC - nutrients)
#
#  SP * (1 - 55.6 * 0.00230 / 35.16504 - 4.7 * 0.00208 / 35.16504) = ....
#
#  SP * 0.9960854303 = 1/r * (SA - 55.6 * TA - 4.7 * DIC - nutrients)
#
#  SP = 0.999218211671 * (SA - 55.6 * TA - 4.7 * DIC - nutrients)
{
    SP <- 0.999218211671 * (SA - 55.6 * TA - 4.7 * DIC - 38.9 * NO3 - 50.7 * SIOH4)
    return(SP)
}


sa2sp_geo <- function (SA, P=0, long=1.e20, lat=1.e20)
#
# This function converts from Absolute (SA) to Practical Salinity (SP)
# based on Pressure P in dbar and geographic location : 
#   - latitude in decimal degrees [-90 ... 90] 
#   - longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]
#
# Latitude and longitude are optional. 
# If any is omitted, an arbitrary geographic point is chosen : central equatorial Atlantic.
# Note that this implies an error on computed SP ranging from 0 up to 0.02 g/kg
{
    if ((length(lat) == 1 && lat == 1.e20) || (length(long) == 1 && long == 1.e20))
    {
        lat <- 0; long <- -25;
    }
    return (gsw_SP_from_SA(SA,P,long,lat))
}


eos2teos_geo <- function (SP, T, P=0, long=1.e20, lat=1.e20)
#
# This function converts Temperature and Salinity from the EOS-80 to the TEOS-10 standard.
# 
# Input parameters :
#   - SP       :   Practical salinity in psu
#   - T        :   In situ Temperature in degrees C
#   - P        :   Sea Pressure in dbar
#   - long     :   longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]
#   - lat      :   latitude in decimal degrees [-90 ... 90] 
#
# Latitude and longitude are optional. 
# If any is omitted, an arbitrary geographic point is chosen : central equatorial Atlantic.
# Note that this implies an error on computed SP ranging from 0 up to 0.02 g/kg
#
# Output : a data frame with the following colums
#    - CT      :   Conservative Temperature
#    - SA      :   Absolute Salinity
#
{
    if ((length(lat) == 1 && lat == 1.e20) || (length(long) == 1 && long == 1.e20))
    {
        lat <- 0; long <- -25;
    }
    # convert salinity
    SA <- gsw_SA_from_SP(SP, P, long, lat)
    # convert temperature
    CT <- gsw_CT_from_t (SA, T, P)
    
    return (as.data.frame(cbind(CT, SA)))
}

teos2eos_geo <- function (SA, CT, P=0, long=1.e20, lat=1.e20)
#
# This function converts Temperature and Salinity from the TEOS-10 to the EOS-80 standard.
# 
# Input parameters :
#   - SA       :   Absolute salinity in g/kg
#   - CT       :   Conservative Temperature in degrees C
#   - P        :   Sea Pressure in dbar
#   - long     :   longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]
#   - lat      :   latitude in decimal degrees [-90 ... 90] 
#
# Latitude and longitude are optional. 
# If any is omitted, an arbitrary geographic point is chosen : central equatorial Atlantic.
# Note that this implies an error on computed SP ranging from 0 up to 0.02 g/kg
#
# Output : a data frame with the following colums
#    - T       :   in situ Temperature
#    - SP      :   Practical Salinity
#
{
    if ((length(lat) == 1 && lat == 1.e20) || (length(long) == 1 && long == 1.e20))
    {
        lat <- 0; long <- -25;
    }
    # convert temperature
    T <- gsw_t_from_CT (SA, CT, P)
    # convert salinity
    SP <- gsw_SP_from_SA(SA, P, long, lat)
    
    return (as.data.frame(cbind(T, SP)))
}


eos2teos_chem <- function (SP, T, P=0, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
#
# This function from Temperature and Salinity from EOS-80 to TEOS-10 standard.
# 
# Input parameters :
#   - SP       :   Practical salinity in psu
#   - T        :   in situ Temperature in degrees C
#   - TA       :   Total Alkalinity in mol/kgSW
#   - DIC      :   Dissolved Inorganic Carbon in mol/kgSW
#   - NO3      :   Nitrate concentration in mol/kgSW
#   - SIOH4    :   Total Silicate in mol/kgSW
#
# Output : a data frame with the following colums
#    - CT      :   Conservative Temperature
#    - SA      :   Absolute Salinity
#
{
    # convert salinity
    SA <- sp2sa_chem (SP, TA=TA, DIC=DIC, NO3=NO3, SIOH4=SIOH4)
    # convert temperature
    CT <- gsw_CT_from_t (SA, T, P)
    
    return (as.data.frame(cbind(CT, SA)))
}


teos2eos_chem <- function (SA, CT, P=0, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
#
# This function converts Temperature and Salinity from the TEOS-10 to the EOS-80 standard.
# 
# Input parameters :
#   - SA       :   Absolute salinity in g/kg
#   - CT       :   Conservative Temperature in degrees C
#   - TA       :   Total Alkalinity in mol/kgSW
#   - DIC      :   Dissolved Inorganic Carbon in mol/kgSW
#   - NO3      :   Nitrate concentration in mol/kgSW
#   - SIOH4    :   Total Silicate in mol/kgSW
#
# Output : a data frame with the following colums
#    - T       :   in situ Temperature
#    - SP      :   Practical Salinity
#
{
    # convert temperature
    T <- gsw_t_from_CT (SA, CT, P)
    # convert salinity
    SP <- sa2sp_chem (SA, TA=TA, DIC=DIC, NO3=NO3, SIOH4=SIOH4)
    
    return (as.data.frame(cbind(T, SP)))
}

# Copyright (C) 2007 Karline Soetaert (K.Soetaert@nioo.knaw.nl)
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


"Kn" <- 
function (S=35, T=25, P=0, pHscale="T",warn="y")

#--------------------------------------------------------------
# Dissociation constant of ammonium 
#--------------------------------------------------------------

{

nK <- max(length(S), length(T), length(P), length(pHscale))

##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

if(length(S)!=nK){S <- rep(S[1], nK)}
if(length(T)!=nK){T <- rep(T[1], nK)}
if(length(P)!=nK){P <- rep(P[1], nK)}
if(length(pHscale)!=nK){pHscale <- rep(pHscale[1],nK)}

#-------Constantes----------------

tk = 273.15           # [K] (for conversion [deg C] <-> [K])
TC = T + tk           # TC [C]; T[K]

#--------------------------------------------------------------
# Dissociation constant of ammonium on seawater scale - Millero 1995
#--------------------------------------------------------------
lnK <- -6285.33/TC+0.0001635*TC-0.25444+(0.46532-123.7184/TC)* sqrt(S)+
      (-0.01992+3.17556/TC)*S
Kn <- exp(lnK)

# ----------------- Pressure Correction ------------------	
Kn <- Pcorrect(Kvalue=Kn, Ktype="Kn", T=T, S=S, P=P, pHscale="SWS", warn=warn)


###----------------pH scale corrections
factor <- rep(NA,nK)
pHsc <- rep(NA,nK)
for(i in (1:nK)){   
 if(pHscale[i]=="T"){factor[i] <- kconv(S=S[i], T=T[i], P=P[i])$kSWS2total ; pHsc[i] <- "total scale"}
 if(pHscale[i]=="F"){factor[i] <- kconv(S=S[i], T=T[i], P=P[i])$kSWS2free ; pHsc[i] <- "free scale"}
 if(pHscale[i]=="SWS"){factor[i] <- 1 ; pHsc[i] <- "seawater scale"}
Kn[i] <- Kn[i]*factor[i]
}

##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>45 | T<0 | S<0 | S>45)) ) {warning("S and/or T is outside the range of validity of the formulation chosen for Kn.")}

  attr(Kn,"pH scale") = pHsc
  attr(Kn,"unit")     = "mol/kg-soln"
  return(Kn)

}  ## END Kn

# Copyright (C) 2024 James Orr
# This file is part of seacarb.
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# Compute total hydrogen ion activity coefficient to convert between SWS and NBS scales

fH <- function(S=35, T=25){
    # --------------------------------------------------------------------------------------------------------------
    # ah = fH * [H+]sws
    # where ah  is the activity of hydrogen ion,
    #       fH  is the total activity coefficient, and
    #       [H+]sws = [H+] + [HSO4-] + [HF], or in other words "the hydrogen ion conccentration on the seawater scale"
    #
    # Note:
    # -----
    # pHnbs = -log10(ah)       #pH on the NBS scale
    # phsws = -log10([H+]sws)  #pH on the SWS scale
    # --------------------------------------------------------------------------------------------------------------

    # The activity coefficient (fH) is used to convert from H+ conccentration on SWS scale 
    # to H+ activity (ah), as used for NBS scale and vice versa. 
    # Here, fH is taken from Takahashi et al (1982, GEOSECS Pacific Expedition, Chap 3, p. 80) who say:
    # "fH is the total activity coeff., which includes contributions from HSO4- and HF [as well as H+].
    #  Culberson & Pytkowicz (28) determined fH as a function of temperature and salinity, and
    #  their results can be approximated by:"

    tk <- 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK <- T + tk;           # T [C]; T [K]
 
    #f = (1.2948 - 0.002036*TK + (0.0004607 - 0.000001475*TK)*S^2)       #CO2SYS.m code
    f  =  1.2948 - 0.002036*TK + 4.607e-4 * S^2  - 1.475e-6 * S^2 * TK   #eqn just as in Takahashi et al.

    # The approach used to compute fH is old. Its use to convert between the NBS and SWS scales is
    # full of uncertainty.  Newer approaches are more complicated (Pitzer equations) but big uncertainties
    # remain (Marion et al., 2011; Pilson, 2013).
    #
    # Culberson, CH, & Pytkowicz, RM (1973). Ionization of water in seawater. Marine Chemistry, 1(4), 309-316.
    #
    # Marion GM, Millero FJ, Camoes MF, Spitzer P, Feistel R, Chen CTA. 2011. pH of seawater. Marine Chemistry 126: 89-96
    #
    # Pilson MEQ. 2013. An introduction to the chemistry of the sea, 2 edn. Cambridge, UK: Cambridge University Press.
    #
    # Takahashi, T. et al (1982). Carbonate chemistry. GEOSECS Pacific Expedition, Volume 3, Hydrographic Data 1973-1974, 77-83.

return(f)
}

# Copyright (C) 2008 Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye and Jean-Pierre Gattuso  
# with a most valuable contribution of Bernard Gentili
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
carb<-
function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", warn="y", eos="eos80", long=1.e20, lat=1.e20){

   # Pass to calculate_carb() all received parameters plus enclosing environment of carb function
    RES <- calculate_carb(flag, var1, var2, S, T, Patm, P, Pt, Sit, NH4t=0, HSt=0, k1k2, kf, ks, pHscale, b, gas, badd=0, warn, eos, long, lat, fullresult=FALSE, environment(carb))
    return(RES)
}

# derivnum()
# This subroutine computes partial derivatives of output carbonate variables 
# with respect to input variables (two), plus nutrients (two), temperature and salinity,
# and dissociation constants.
#
# It uses the central differences, which consists of  
# - introducing a small perturbation delta (plus or minus) in one input
# - and computing the induced delta in output variables
#
# The ratio between delta and input value is chosen as follows: 
#    for perturbing input pair :  a constant which depends on the type of input pair of variables.
#    for perturbing nutrients  :  1.e-6
#    for perturbing T and S    :  1.e-3
#
# Input parameters :
#   - varid  :  (scalar) identifier of variable with respect to which derivative is requested
#                = variable length, case insensitive, character code
#                  case '1' or 'var1'  :  Variable 1 of the input pair (This is TAlk if flag is 15)
#                  case '2' or 'var2'  :  Variable 2 of the input pair (This is DIC  if flag is 15)
#                  case 'sil', 'silt', 'tsil' or 'silicate'      : Total dissolved inorganic silicon concentration
#                  case 'phos', 'phost', 'tphos' or 'phosphate'  : Total dissolved inorganic phosphorus concentration
#                  case 't', 'temp' or 'temperature' : temperature
#                  case 's', 'sal' or 'salinity'     : salinity
#                  case 'K0','K1','K2','Kb','Kw','Kspa' or 'Kspc' : dissociation constant
#                  case 'bor' : total boron
#
#   - others     :  same à input of subroutine  carb() : scalar or vectors
#
# Returns one set of partial derivatives
derivnum <- 
function(varid, flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, 
         k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", warn="y",  eos="eos80", long=1.e20, lat=1.e20)
{
    # Input conditionning
    # -------------------
    
    n <- max(length(flag), length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), length(ks), length(b))
    if(length(flag)!=n){flag <- rep(flag[1],n)}
    if(length(var1)!=n){var1 <- rep(var1[1],n)}
    if(length(var2)!=n){var2 <- rep(var2[1],n)}
    if(length(S)!=n){S <- rep(S[1],n)}
    if(length(T)!=n){T <- rep(T[1],n)}
    if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
    if(length(P)!=n){P <- rep(P[1],n)}
    if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
    if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
    if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
    if(length(kf)!=n){kf <- rep(kf[1],n)}
    if(length(ks)!=n){ks <- rep(ks[1],n)}
    if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
    if(length(b)!=n){b <- rep(b[1],n)}

    # Only two options for eos
    if (eos != "teos10" && eos != "eos80")
        stop ("invalid parameter eos: ", eos)

    varid <-toupper(varid)
    
    # Constants: names of variables, sorted by flag number
    varnames  = rbind (
        c("H", "CO2"),            # flag = 1    
        c("CO2", "HCO3"),         # flag = 2    
        c("CO2", "CO3"),          # flag = 3    
        c("CO2", "ALK"),          # flag = 4    
        c("CO2", "DIC"),          # flag = 5    
        c("H", "HCO3"),           # flag = 6    
        c("H", "CO3"),            # flag = 7    
        c("H", "ALK"),            # flag = 8    
        c("H", "DIC"),            # flag = 9    
        c("HCO3", "CO3"),         # flag = 10   
        c("HCO3", "ALK"),         # flag = 11   
        c("HCO3", "DIC"),         # flag = 12   
        c("CO3", "ALK"),          # flag = 13   
        c("CO3", "DIC"),          # flag = 14   
        c("ALK", "DIC"),          # flag = 15 
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("pCO2", "H"),           # flag = 21   
        c("pCO2", "HCO3"),        # flag = 22   
        c("pCO2", "CO3"),         # flag = 23   
        c("pCO2", "ALK"),         # flag = 24   
        c("pCO2", "DIC")          # flag = 25   
    )
    var1name <- varnames[flag,1]
    var2name <- varnames[flag,2]
    
    # Relative deltas (perturbations) sorted by flag number
    # default value is 1.e-3
    const_deltas = rbind (
        c(1.e-8, 1.e-3),         # flag = 1    
        c(1.e-3, 1.e-3),         # flag = 2    
        c(1.e-3, 1.e-3),         # flag = 3    
        c(1.e-6, 1.e-6),         # flag = 4    
        c(1.e-5, 1.e-5),         # flag = 5    
        c(1.e-8, 1.e-3),         # flag = 6    
        c(1.e-8, 1.e-3),         # flag = 7    
        c(1.e-1, 1.e-5),         # flag = 8    
        c(1.e-1, 1.e-1),         # flag = 9    
        c(1.e-3, 1.e-3),         # flag = 10   
        c(1.e-3, 1.e-3),         # flag = 11   
        c(1.e-3, 1.e-3),         # flag = 12   
        c(1.e-3, 1.e-3),         # flag = 13   
        c(1.e-3, 1.e-3),         # flag = 14   
        c(1.e-6, 1.e-6),         # flag = 15   
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(1.e-3, 1.e-8),         # flag = 21   
        c(1.e-3, 1.e-3),         # flag = 22   
        c(1.e-3, 1.e-3),         # flag = 23   
        c(1.e-6, 1.e-6),         # flag = 24   
        c(1.e-5, 1.e-5)          # flag = 25   
    )

    # Replace all values of const_deltas (above) with 1e-6 for consistency with CO2SYS-MATLAB
    const_deltas = rbind (
        c(1.e-6, 1.e-6),         # flag = 1    
        c(1.e-6, 1.e-6),         # flag = 2    
        c(1.e-6, 1.e-6),         # flag = 3    
        c(1.e-6, 1.e-6),         # flag = 4    
        c(1.e-6, 1.e-6),         # flag = 5    
        c(1.e-6, 1.e-6),         # flag = 6    
        c(1.e-6, 1.e-6),         # flag = 7    
        c(1.e-6, 1.e-6),         # flag = 8    
        c(1.e-6, 1.e-6),         # flag = 9    
        c(1.e-6, 1.e-6),         # flag = 10   
        c(1.e-6, 1.e-6),         # flag = 11   
        c(1.e-6, 1.e-6),         # flag = 12   
        c(1.e-6, 1.e-6),         # flag = 13   
        c(1.e-6, 1.e-6),         # flag = 14   
        c(1.e-6, 1.e-6),         # flag = 15   
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(1.e-6, 1.e-6),         # flag = 21   
        c(1.e-6, 1.e-6),         # flag = 22   
        c(1.e-6, 1.e-6),         # flag = 23   
        c(1.e-6, 1.e-6),         # flag = 24   
        c(1.e-6, 1.e-6)          # flag = 25   
    )
    
    # Reference 'concentrations': sorted by flag number
    const_refs  = rbind (
        c(1e-8, 10e-6),          # flag = 1    
        c(10e-6, 1800e-6),       # flag = 2    
        c(10e-6, 100e-6),        # flag = 3    
        c(10e-6, 2300e-6),       # flag = 4    
        c(10e-6, 2000e-6),       # flag = 5    
        c(1e-8, 1800e-6),        # flag = 6    
        c(1e-8, 100e-6),         # flag = 7    
        c(1e-8, 2300e-6),        # flag = 8    
        c(1e-8, 2000e-6),        # flag = 9    
        c(1800e-6, 100e-6),      # flag = 10   
        c(1800e-6, 2300e-6),     # flag = 11   
        c(1800e-6, 2000e-6),     # flag = 12   
        c(100e-6, 2300e-6),      # flag = 13   
        c(100e-6, 2000e-6),      # flag = 14   
        c(2300e-6, 2000e-6),     # flag = 15 
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(0, 0),
        c(400, 1e-8),            # flag = 21   
        c(400, 1800e-6),         # flag = 22   
        c(400, 100e-6),          # flag = 23   
        c(400, 2300e-6),         # flag = 24   
        c(400, 2000e-6)          # flag = 25   
    )

    # Compute two slightly different values for input
    # -----------------------------------------------

    # Default values : not perturbed
    # no change on var1 and var2
    var11 = var12 = var1
    var21 = var22 = var2
    # no change on Sil total and Phos total
    Sit1 = Sit2 = Sit
    Pt1  = Pt2  = Pt
    # no change on T and S
    T1 = T2 = T
    S1 = S2 = S

    # Initialise absolute deltas
    abs_dx <- rep(NA, n)

    # Uppercase names of dissociation constants and 'total boron'
    K_id <- c('K0', 'K1', 'K2', 'KB', 'KW', 'KSPA', 'KSPC', 'BOR')
    # Flag for dissociation constant as perturbed variable 
    flag_dissoc_K = varid %in% K_id
    
    # if perturbed variable is a dissociation constant
    if (flag_dissoc_K)
    {
        # Approximate values for K0, K1, K2, Kb, Kw, Kspa, Kspc, and bor
        # They will be used to compute an absolute perturbation value on these constants
        K <- c(0.034, 1.2e-06, 8.3e-10, 2.1e-09, 3.1e-14, 6.7e-07, 4.3e-07, 4.156877e-4)
        # Choose value of absolute perturbation
        index <- which (K_id == varid)
        perturbation = K[index] * 1.e-3   # 0.1 percent of Kx value
        abs_dx = 2 * perturbation
    }
    # Var1 is perturbed
    else if (varid %in% c('1', 'VAR1'))
    {
        # Define relative deltas
        delta = const_deltas[flag, 1]
        xref  = const_refs[flag, 1]

        # where input variable is [H+] concentration
        FH = (var1name == 'H')
        # Change slightly [H+]
        H_var1 <- 10^(-var1[FH])
        H1 = H_var1 - xref[FH]*delta[FH]
        H2 = H_var1 + xref[FH]*delta[FH]
        var11[FH] = -log10(H1)
        var12[FH] = -log10(H2)
        abs_dx[FH] = H2 - H1

        # where it is not
        GH = ! FH
        # Change slightly var1
        var11[GH] = var1[GH] - xref[GH]*delta[GH]
        var12[GH] = var1[GH] + xref[GH]*delta[GH]
        abs_dx[GH] = var12[GH] - var11[GH]
    }
    else if (varid %in% c('2', 'VAR2'))    # var2 (second variable of input pair) is perturbed
    {
        # Define relative deltas
        delta = const_deltas[flag, 2]
        xref  = const_refs[flag, 2]

        # where second input variable is [H+] concentration
        FH = (var2name == 'H')
        # Change slightly H+]
        H_var2 <- 10^(-var2[FH])
        H1 = H_var2 - xref[FH]*delta[FH]
        H2 = H_var2 + xref[FH]*delta[FH]
        var21[FH] = -log10(H1)
        var22[FH] = -log10(H2)
        abs_dx[FH] = H2 - H1

        # where it is not
        GH = ! FH
        # Change slightly var2
        var21[GH] = var2[GH] - xref[GH]*delta[GH]
        var22[GH] = var2[GH] + xref[GH]*delta[GH]
        abs_dx[GH] = var22[GH] - var21[GH]
    }
    else if (varid %in% c('SIL', 'TSIL', 'SILT', 'SILICATE', 'SIT'))    # Sil total
    {
        # Define relative delta
        delta  = 1.e-3
        Sitref = 7.5e-6 # mol/kg (Global surface mean)

        center_value = Sit
        # Change slightly total silicon
        Sit1 = Sit - Sitref*delta
        Sit2 = Sit + Sitref*delta
        abs_dx = Sit2 - Sit1
    }
    else if (varid %in% c('PHOS', 'TPHOS', 'PHOST', 'PHOSPHATE', 'PT'))    # Phos total
    {
        # Define relative delta
        delta = 1.e-3
        Ptref = 0.5e-6   # mol/kg (Global surface mean)

        center_value = Pt
        # Change slightly total phosphorus
        Pt1 = Pt - Ptref*delta
        Pt2 = Pt + Ptref*delta
        abs_dx = Pt2 - Pt1
    }
    else if (varid %in% c('T', 'TEMP', 'TEMPERATURE'))    # Temperature
    {
        # Define relative delta
        delta = 1.e-4
        Tref = 18   # degrees Celcius
        
        center_value = T
        # Change slightly temperature
        T1 = T - Tref*delta
        T2 = T + Tref*delta
        abs_dx = T2 - T1
    }
    else if (varid %in% c('S', 'SAL', 'SALINITY'))   # Salinity
    {
        # Define relative delta
        delta = 1.e-4
        Sref = 35

        center_value = S
        # Change slightly salinity
        S1 = S - Sref*delta
        S2 = S + Sref*delta
        abs_dx = S2 - S1
    }
    else
    {
        stop ("Invalid input parameter: ", varid)
    }

    # Define local functions K0, K1, K2, ...
    # --------------------------------------

    # These function will replace original K0, K1, K2.... functions, at some computation time
    # They are intended to return slightly modified (perturbed) values of dissociation constants
    #
    # Their input (beside their regular parameters):
    #
    #   perturbation :  absolute value of (tiny) perturbation
    #   sign_factor  :  factor -1 or +1 gives the sign or perturbation
    if (varid == 'K0') 
    {
        K0 <- function(S=35,T=25,P=0,Patm=1,...)
       #K0 <- function(S=35,T=25,P=0,Patm=1,warn='n')
        {
            # Call original K0 function
            out <- seacarb::K0(S, T, P, Patm=Patm)
            #out <- K0(S, T, P, Patm=Patm, warn=warn)
            # perturb value of K0
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'K1') 
    {
        K1 <-function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0,...)
       #K1 <-function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0,warn='n')
        {
            # Call original K1 function
            out <- seacarb::K1(S, T, P, k1k2=k1k2, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0)
           #out <- K1(S, T, P, k1k2=k1k2, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0, warn=warn)
            # perturb value of K1
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'K2') 
    {
        K2 <- function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0,...)
       #K2 <- function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0,warn='n')
        {
            # Call original K2 function
            out <- seacarb::K2(S, T, P, k1k2=k1k2, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0)
           #out <- seacarb::K2(S, T, P, k1k2=k1k2, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0, warn=warn)
            # perturb value of K2
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'KW') 
    {
        Kw <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0,...)
       #Kw <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0, warn='n')
        {
            # Call original Kw function
            out <- seacarb::Kw(S, T, P, pHscale=pHscale, kSWS2scale=kSWS2scale)
           #out <- Kw(S, T, P, pHscale=pHscale, kSWS2scale=kSWS2scale, warn=warn)
            # perturb value of Kw
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'KB')
    {
        Kb <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0,...)
       #Kb <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0, warn='n')
        {
            # Call original Kb function
            out <- seacarb::Kb(S, T, P, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0)
           #out <- Kb(S, T, P, pHscale=pHscale, kSWS2scale=kSWS2scale, ktotal2SWS_P0=ktotal2SWS_P0, warn=warn)
            # perturb value of Kb
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'KSPA')
    {
        Kspa <- function(S=35,T=25,P=0,...)
       #Kspa <- function(S=35,T=25,P=0, warn='n')
        {
            # Call original Kspa function
            out <- seacarb::Kspa(S, T, P)
           #out <- Kspa(S, T, P, warn=warn)
            # perturb value of Kspa
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'KSPC')
    {
        Kspc <- function(S=35,T=25,P=0,...)
       #Kspc <- function(S=35,T=25,P=0, warn='n')
        {
            # Call original Kspc function
            out <- seacarb::Kspc(S, T, P)
           #out <- Kspc(S, T, P, warn=warn)
            # perturb value of Kspc
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    else if (varid == 'BOR')
    {
       #Create a new function "bor" that will replace  "seacarb::bor" at call time
        bor <- function(S=35,b="u74")
        {
            # Call original bor function
            out <- seacarb::bor(S,b)
            # perturb value of bor
            out = out + sign_factor * perturbation  # sign_factor is +1 or -1
            return (out)
        }
    }
    
    # PERTURBATION:
    # -------------
    
    # if perturbed variable is a dissociation constant
    if (flag_dissoc_K)
    {
        # Save then change execution environment of function carb()
        saved_env <- environment(carb)
        # Using environment(NULL) will hide original seacarb::K0, seacarb::K1, ... function (depending on "varid")
        environment(carb) <- environment(NULL)
        # Note :  environment(NULL) is the environment created when the function derivnum() is executed
        #         It then contains all locally defined functions, like K0, K1, K2, ...
        #         Then, when calculate_carb() is executed (called by carb() below)
        #         and looks for a name (variable or function) that is not a local variable,
        #         it will look into evironment(NULL), which has been attached to carb(), 
        #         before looking into global or seacarb module environment.
        
        # Point 1: (one dissociation constant is somewhat smaller)
        sign_factor = -1.0
        cdel1 <- carb(flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)

        # Point 2: (one dissociation constant is somewhat bigger)
        sign_factor = 1.0
        cdel2 <- carb(flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)

        # Restore environment of carb()
        environment(carb) <- saved_env
    }
    else
    {
        # Point 1: (one of var1, var2, T or S is somewhat smaller)
        cdel1 <- carb(flag, var11, var21, S=S1, T=T1, Patm=Patm, P=P, Pt=Pt1, Sit=Sit1, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)

        # Point 2: (one of var1, var2, T or S is somewhat bigger)
        cdel2 <- carb(flag, var12, var22, S=S2, T=T2, Patm=Patm, P=P, Pt=Pt2, Sit=Sit2, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
    }

    # Compute [H+] concentration and add it to data-frame
    H <- 10^(-cdel1$pH)
    cdel1 <- cbind(H,cdel1)
    H <- 10^(-cdel2$pH)
    cdel2 <- cbind(H,cdel2)

    # Centered difference
    dy <-(cdel2 - cdel1)

    # Drop unnecessary columns
    # ------------------------
    
    # list fixed columns to drop
    drops=c('flag', 'S', 'T', 'P', 'Patm', 'fCO2pot', 'pCO2pot', 'fCO2insitu', 'pCO2insitu')
    # add input variable 1 to list
    if (all(var1name == var1name[1]))
    {
        drops = c(drops, var1name)
        # if to drop H, drop also pH
        if (var1name[1] == 'H')
            drops = c(drops, 'pH')
    }
    # add input variable 2 to list
    if (all(var2name == var2name[1]))
    {
        drops = c(drops, var2name)
        # if to drop H, drop also pH
        if (var2name[1] == 'H')
            drops = c(drops, 'pH')
    }
    # Drop columns
    dy <- dy[,!(names(dy) %in% drops)]

    # Compute ratio dy/dx
    dydx <- dy / abs_dx

    # There are 10 or  8 output variables 
    # return partial derivatives as a 8x1 array
    return (dydx)
}

# Copyright (C) 2007 Karline Soetaert (K.Soetaert@nioo.knaw.nl)
# Revised by James Orr, 2012-01-17
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"Ksi" <- 
function (S=35,T=25,P=0,pHscale="T",kSWS2scale="x",warn="y")

#--------------------------------------------------------------
# Dissociation constant for Si(OH)4
#--------------------------------------------------------------

{

    nK <- max(length(S), length(T), length(P), length(pHscale), length(kSWS2scale))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}

	    
    #-------Constantes----------------


    tk = 273.15               # [K] (for conversion [deg C] <-> [K])
    TC = T + tk               # TC [C]; T[K]

    #--------------------------------------------------------------
    # Dissociation constant of Si(OH)4 on total scale - DOE 1994 - correct for mol/(kg-H2O)->mol/kg soln
    #--------------------------------------------------------------
    Io  <- 19.924*S/(1000-1.005*S)                     # ionic strenght, mol/kg-H2O     # ionic strength
	#	*** J. Orr (15 Jan 2013): Formulation changed to be on the SWS scale (without later conversion)
	                                                     

	# From J. C. Orr on 15 Jan 2013:
	# The formulation below was a modified version of Millero (1995) where Dickson et al. (2007) subtracted 0.015
        # from Millero's original constant (117.40) to give 117.385 (the 2nd term above). BUT Dickson's reason for that 
        # operation was to "convert--approximately--from theSWS pH scale (including HF) used by Millero (1995) to the 'total' 
        # scale ...". 
        # This subtraction of 0.015 to switch from the SWS to Total scale is not good for 2 reasons:
        # (1) The 0.015 value is inexact (not constant), e.g., it is 0.022 at T=25, S=35, P=0;
	# (2) It makes no sense to switch to the Total scale when just below you switch back to the SWS scale.
        # The best solution is to reestablish the original equation (SWS scale) and delete the subsequent scale conversion.

    #lnK <- 117.385 + 3.5913*sqrt(Io) - 1.5998*Io + 0.07871*Io*Io +
    lnK <- 117.40  + 3.5913*sqrt(Io) - 1.5998*Io + 0.07871*Io*Io +
	  (-8904.2 - 458.79*sqrt(Io) + 188.74*Io - 12.1652*Io*Io)/TC +
	  (-19.334)*log(TC)+log(1-0.001005*S)

    Ksi <- exp(lnK)

    # ---- Conversion from Total scale to seawater scale before pressure corrections
    #  *** JCO: This is no longer necessary: with original formulation (Millero, 1995), Kw is on "seawater scale"!
    # factor <- kconv(S=S, T=T, P=rep(0,nK))$ktotal2SWS
    # Ksi <- Ksi * factor

    # If needed, pressure correction
    if (any (P != 0))
        Ksi <- Pcorrect(Kvalue=Ksi, Ktype="Ksi", T=T, S=S, P=P, pHscale="SWS", 1., 1., warn=warn)


    ###----------------pH scale corrections
    
    # Which pH scales are required ?
    is_total <- pHscale=="T"
    is_free   <- pHscale=="F"

    # if any pH scale correction required (from SWS scale)
    if (any(is_total) || any(is_free))
    {
        # if pH scale correction factor not given
        if (missing(kSWS2scale) || kSWS2scale[1] == "x")
        {
            # Compute it
            kSWS2scale <- rep(1.0,nK)
            if (any(is_total))
                kSWS2scale[is_total] <- kconv(S=S[is_total], T=T[is_total], P=P[is_total])$kSWS2total
            if (any(is_free))
                kSWS2scale[is_free]  <- kconv(S=S[is_free], T=T[is_free], P=P[is_free])$kSWS2free
        }
        else
            # Check its length
            if(length(kSWS2scale)!=nK){kSWS2scale <- rep(kSWS2scale[1], nK)}
        # Apply pH scale correction
    	Ksi <- Ksi*kSWS2scale
    }

    # Return full name of pH scale
    pHsc <- rep(NA,nK)
    pHsc[is_total] <- "total scale"
    pHsc[is_free]  <- "free scale"
    pHsc[!is_total & !is_free] <- "seawater scale"

    ##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>45 | S>45 | T<0 | S<0)) ) {warning("S and/or T is outside the range of validity of the formulation available for Ksi in seacarb.")}

    attr(Ksi,"unit")     = "mol/kg-soln"
    attr(Ksi,"pH scale") = pHsc
    return(Ksi)
}  ## END Ksi

# Copyright (C) 2008 Jean-Pierre Gattuso and Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#

"Kb" <-
function(S=35,T=25,P=0,pHscale="T",kSWS2scale="x",ktotal2SWS_P0="x",warn="y"){

    nK <- max(length(S), length(T), length(P), length(pHscale), length(kSWS2scale) ,length(ktotal2SWS_P0))

    ##-------- Create vectors for all input (if vectorial)

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}
            
    #-------Constants----------------

    #---- issues of equic----
    tk <- 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK <- T + tk;           # T [C]; TK [K]

            
    #---------------------------------------------------------------------
    # --------------------- Kb  --------------------------------------------
    #  Kbor = [H+][B(OH)4-]/[B(OH)3]
    #
    #   (Dickson, 1990 in Guide to Best Practices in Ocean CO2 Measurements 2007)
    #   pH-scale: 'total'. mol/kg-soln
    
    
    tmp1 =  (-8966.90-2890.53*sqrt(S)-77.942*S+1.728*S^(3/2)-0.0996*S*S);
    tmp2 =   +148.0248+137.1942*sqrt(S)+1.62142*S;
    tmp3 = +(-24.4344-25.085*sqrt(S)-0.2474*S)*log(TK);
    
    lnKb = tmp1 / TK + tmp2 + tmp3 + 0.053105*sqrt(S)*TK;
    Kb <- exp(lnKb)

    ## ---- Conversion from Total scale to seawater scale before pressure corrections
    
    # if correction factor (from Total scale to seawater at P=0) not given
    if (missing(ktotal2SWS_P0) || ktotal2SWS_P0[1] == "x")
    {
        # Compute it
        ktotal2SWS_P0 <- kconv(S=S, T=T, P=0)$ktotal2SWS
    }
    else
        # Check its length
        if(length(ktotal2SWS_P0)!=nK) ktotal2SWS_P0 <- rep(ktotal2SWS_P0[1], nK)
    Kb <- Kb * ktotal2SWS_P0

    # -------------------Correct for Pressure effect, if needed ---------------------------
    if (any (P != 0))
        Kb <- Pcorrect(Kvalue=Kb, Ktype="Kb", T=T, S=S, P=P, pHscale="SWS", 1., 1., warn=warn)

    ###----------------pH scale corrections

    # Which pH scales are required ?
    is_total <- pHscale=="T"
    is_free   <- pHscale=="F"

    # if any pH scale correction required (from SWS scale)
    if (any(is_total) || any(is_free))
    {
        # if pH scale correction factor not given
        if (missing(kSWS2scale) || kSWS2scale[1] == "x")
        {
            # Compute it
            kSWS2scale <- rep(1.0,nK)
            if (any(is_total)) {kSWS2scale[is_total] <- kconv(S=S[is_total], T=T[is_total], P=P[is_total])$kSWS2total}
            if (any(is_free))  {kSWS2scale[is_free]  <- kconv(S=S[is_free], T=T[is_free], P=P[is_free])$kSWS2free}
        }
        else
            # Check its length
            if(length(kSWS2scale)!=nK){kSWS2scale <- rep(kSWS2scale[1], nK)}
        # Apply pH scale correction
        Kb <- Kb*kSWS2scale
    }

    # Return full name of pH scale
    pHsc <- rep(NA,nK)
    pHsc[is_total] <- "total scale"
    pHsc[is_free]  <- "free scale"
    pHsc[!is_total & !is_free] <- "seawater scale"

    
    ##------------Warnings
    is_w <- warn == "y"
    
    if (any (is_w & (T>45 | S>45 | T<0 | S<5)) ) {warning("S and/or T is outside the range of validity of the formulation available for Kb in seacarb.")}


    attr(Kb,"unit")     = "mol/kg-soln"
    attr(Kb,"pH scale") = pHsc
    return(Kb)
}

# Copyright (C) 2008 Jean-Pierre Gattuso and Jean-Marie Epitalon and Heloise Lavigne and Aurelien Proye
    #
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
"K0" <-
function(S=35,T=25,P=0,Patm=1,warn="y"){

    nK <- max(length(S), length(T), length(P), length(Patm))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(Patm)!=nK){Patm <- rep(Patm[1], nK)}

    #-------Constantes----------------

    #---- issues de equic----
    tk = 273.15;           # [K] (for conversion [deg C] <-> [K])
    TK = T + tk;           # T[K] (absolute temperature)

    #---------------------------------------------------------------------
    #---------------------- K0 (K Henry) ---------------------------------
    #
    #               CO2(g) <-> CO2(aq.)
    #               K0      = [CO2]/ p CO2
    #
    #   Weiss (1974)   [mol/kg/atm]
    #
    #                             

    tmp = 9345.17 / TK - 60.2409 + 23.3585 * log(TK/100);
    nK0we74 = tmp + S*(0.023517-0.00023656*TK+0.0047036e-4*TK*TK);

    ##------------Warnings

    is_w <- warn == "y"

    if (any (is_w & (T>45 | S>45 | T<(-1))) ) {warning("S and/or T is outside the range of validity of the formulation available for K0 in seacarb.")}

    K0= exp(nK0we74);

    #---------------------------------------------------------------------
    #---------------------- Pressure correction ----------------------------
    Phydro_atm = P / 1.01325  # convert hydrostatic pressure from bar to atm (1.01325 bar / atm)
    Ptot = Patm + Phydro_atm  # total pressure (in atm) = atmospheric pressure + hydrostatic pressure
    R = 82.05736              # (cm3 * atm) / (mol * K)  CODATA (2006)
    vbarCO2 = 32.3            # partial molal volume (cm3 / mol) from Weiss (1974, Appendix, paragraph 3)
    K0 = K0 * exp( ((1-Ptot)*vbarCO2)/(R*TK) )   # Weiss (1974, equation 5), TK is absolute temperature (K)

    attr(K0,"unit") = "mol/kg"	
    return(K0)
}

# Copyright (C) 2009 Jean-Pierre Gattuso and H?loise Lavigne
# with a most valuable contribution of Bernard Gentili <gentili@obs-vlfr.fr>
# and valuable suggestions from Jean-Marie Epitalon <epitalon@lsce.saclay.cea.fr>
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#
oa <-
function(flag, var1, var2, pCO2f, pCO2s=1e6, S=35, T=25, P=0, Pt=0, Sit=0, k1k2='x', kf='x', ks="d", pHscale="T", plot=FALSE,  b="u74",
        eos="eos80", long=1.e20, lat=1.e20){

n <- max(length(var1), length(var2), length(pCO2f), length(pCO2s), length(S), length(T), length(P), length(Pt), length(Sit), length(k1k2), length(kf), length(pHscale), length(ks))
if(length(flag)!=n){ flag <- rep(flag[1],n)}
if(length(var1)!=n){ var1 <- rep(var1[1],n)}
if(length(var2)!=n){ var2 <- rep(var2[1],n)}
if(length(pCO2f) !=n){ pCO2f <- rep(pCO2f[1],n)}
if(length(pCO2s) !=n){ pCO2s <- rep(pCO2s[1],n)}
if(length(S)!=n){ S <- rep(S[1],n)}
if(length(T)!=n){ T <- rep(T[1],n)}
if(length(P)!=n){ P <- rep(P[1],n)}
if(length(Pt)!=n){ Pt <- rep(Pt[1],n)}
if(length(Sit)!=n){ Sit <- rep(Sit[1],n)}
if(length(k1k2)!=n){ k1k2 <- rep(k1k2[1],n)}
if(length(kf)!=n){ kf <- rep(kf[1],n)}
if(length(ks)!=n){ ks <- rep(ks[1],n)}
if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
if(length(b)!=n){ b <- rep(b[1],n)}


# Only two options for eos
if (eos != "teos10" && eos != "eos80")
    stop ("invalid parameter eos: ", eos)

# if the concentrations of total silicate and total phosphate are NA
# they are set at 0
Sit[is.na(Sit)] <- 0
Pt[is.na(Pt)] <- 0

# initial system
# --------------
SYSi <- carb(flag=flag, var1=var1, var2=var2, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

# CO2 bubbling
# -----------------------------
SYSfCO2bubbling <- carb(flag=24, var1=pCO2f, var2=SYSi$ALK, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)


# Seawater mixing
# ---------------------------------
SYSs <- carb(flag=24, var1=pCO2s, var2=SYSi$ALK, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

SYSf <- carb(flag=24, var1=pCO2f, var2=SYSi$ALK, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

DICi <- SYSi$DIC
DICs <- SYSs$DIC
DICf <- SYSf$DIC

ws <- (DICf-DICi)/(DICs-DICi)

ws <- round(ws, 8)
wi <- round(1-ws, 8)

SYSfSWmixing <- SYSf

# Addition of strong acid
# -----------------------------------

SYSfacid <- carb(flag=25, var1=pCO2f, var2=SYSi$DIC, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)
ALKf <- SYSfacid$ALK
ALKi <- SYSi$ALK

hplus <- ALKi-ALKf

hplus <- round(hplus, 8)

## on va devoir ajouter hplus mole par kg de solution de HCl

# Addition of strong acid and carbonate
# ------------------------------------------------------
SYSfCarboAcid <- carb(flag=24, var1=pCO2f, var2=SYSi$ALK, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf,  ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

deltaDIC <- SYSfCarboAcid$DIC - SYSi$DIC

## with CO3 addition
TAint <- SYSi$ALK + 2*deltaDIC
TAintCO3 <- TAint  ## we keep the value for the plot

DICint <- SYSfCarboAcid$DIC

SYSint <- carb(flag=15, var1=TAint, var2=DICint, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

deltaTA <- SYSf$ALK-SYSint$ALK

CO3 <- round(deltaDIC, 8)
ACIDco3 <- round(-deltaTA, 8)

## with HCO3 addition
TAint <- SYSi$ALK + deltaDIC
TAintHCO3 <- TAint  ## we keep the value for the plot

DICint <- SYSfCarboAcid$DIC

SYSint <- carb(flag=15, var1=TAint, var2=DICint, S=S, T=T, P=P, Sit=Sit, Pt=Pt, k1k2=k1k2, kf=kf, ks=ks, pHscale=pHscale, b=b, eos=eos, long=long, lat=lat)

deltaTA <- SYSf$ALK-SYSint$ALK

HCO3 <- round(deltaDIC, 8)
ACIDhco3 <- round(-deltaTA, 8)


###############################################################################
## OUTPUTS
###############################################################################

OUT <- new.env()

## Summary table for sw state after each experiments
## -------------------------------------------------
co <- as.data.frame(c(rep("initial",n), rep("final-CO2 bubbling", n), rep("final-SW mixing", n), rep("final-acid", n), rep("final-HCO3-and-acid", n), rep("final-CO3-and-acid", n)))

State <- rbind(SYSi, SYSfCO2bubbling, SYSfSWmixing, SYSfacid ,SYSfCarboAcid, SYSfCarboAcid)

State <- cbind(co, State)

names(State)[1] <- "comment"

OUT$summary <- State


## Summary table for manipulations of each experiments
## ---------------------------------------------------

col1 <- c(rep("CO2 bubbling", n), rep("Mixing", n), rep("Addition of acid", n), rep("Addition of HCO3 and acid", n), rep("Addition of CO3 and acid", n))
col2 <- c(rep("Air pCO2 (uatm)", n), rep("Weight fraction high-CO2", n), rep("H+ (mol/kg)", n), rep("HCO3 (mol/kg)", n), rep("CO3 (mol/kg)", n))
col3 <- c(pCO2f,ws,hplus,HCO3,CO3)
col4 <- c(rep(NA,n*3),rep("H+ (mol/kg)", n),rep("H+ (mol/kg)", n)) 
col5 <- c(rep(NA,n*3), ACIDhco3, ACIDco3)

perturbation <- cbind(col1, col2, col3, col4, col5)
perturbation <- data.frame(perturbation)
names(perturbation) <- c("Method", "1st param.", "Value 1st param.", "2nd param.", "Value 2nd param.")
OUT$perturbation <- perturbation

## Description of manipulation
## ---------------------------------------------------

col1 <- c("CO2 bubbling", "Mixing", "Addition of acid", "Addition of HCO3 and acid", "Addition of CO3 and acid")
col2 <- c( paste("Bubble seawater with a gas with pCO2=", round(pCO2f, 3)[1], "uatm", sep=""),
paste("Mix", wi[1], "kg of the normal seawater with", ws[1] ,"kg of high-pCO2 seawater" ),
paste("Add", hplus[1], "mol of H+ to 1 kg of the initial seawater"),
paste("Add", HCO3[1], "mol of HCO3 and then", ACIDhco3[1], "mol of H+ to 1 kg of initial seawater" ),
paste("Add", CO3[1], "mol of CO3 and then", ACIDco3[1], "mol of H+ to 1 kg of initial seawater" ))
description <- cbind(col1, col2)
description <- data.frame(description)
names(description) <- c("Approach", "Description of manipulations")
OUT$description <- description

OUT <- as.list(OUT)


########### GRAPHIQUE

if(plot){
par(mai=c(1,1,1.6,1))
x1 <- SYSi$DIC[1] - 0.1*(SYSf$DIC[1]-SYSi$DIC[1])  # xlim inf
x2 <- SYSf$DIC[1] + 0.1*(SYSf$DIC[1]-SYSi$DIC[1])  # xlim sup
y1 <- SYSfacid$ALK[1] - 0.1*(TAintCO3[1]-SYSfacid$ALK[1])  # ylim inf
y2 <- TAintCO3[1] + 0.1*(TAintCO3[1]-SYSfacid$ALK[1])       # ylim sup
xa <- SYSi$DIC[1]   # starting point
ya <- SYSi$ALK[1]   # starting point
plot(xa ,ya , xlim=c(x1,x2), ylim=c(y1, y2),
	xlab=expression(paste("Dissolved inorganic carbon (mol ", kg^"-1", ")", sep="" )), 
	ylab=expression(paste("Total alkalinity (mol ", kg^"-1", ")", sep="" )),
	pch=19)

## Bubbling and sea watermixing
xb <- SYSf$DIC[1]
yb <- SYSf$ALK[1] 
arrows(xa, ya, xb, yb, col=4, lwd=2)

## strong acid
xb <- SYSfacid$DIC[1]
yb <- SYSfacid$ALK[1]
arrows(xa, ya, xb, yb, col=3, lwd=2)

## strong acid  + base
# CO3
xb <- SYSfCarboAcid$DIC[1]
yb <- TAintCO3[1]
arrows(xa, ya, xb, yb, col=2, lwd=2)
ybb <- SYSfCarboAcid$ALK[1]
arrows(xb, yb, xb, ybb, col=2, lwd=2)
# HCO3
xb <- SYSfCarboAcid$DIC[1]
yb <- TAintHCO3[1]
arrows(xa, ya, xb, yb, col="orange", lwd=2)
ybb <- SYSfCarboAcid$ALK[1]
arrows(xb, yb, xb, ybb, col="orange", lty=2, lwd=2)

alk <- seq(y1, y2,, 15)
dic <- seq(x1, x2,, 15)
z <- expand.grid(ALK=alk, DIC=dic)
cp <- carb(flag=15, var1=z$ALK, var2=z$DIC, T=T[1], P=P[1], S=S[1], Pt=Pt[1], Sit=Sit[1], pHscale=pHscale[1], k1k2=k1k2[1], kf=kf[1], eos=eos, long=long, lat=lat)
pCO2 <- cp$pCO2
M <- tapply(pCO2, z, unique)
contour(dic, alk, t(M),
	levels=c(seq(0, 10000, 100)),
	labcex=0.5,
	method="flattest",
	col=1,
	lwd=1,
	lty="dashed",
	add=TRUE,
	axes = FALSE, frame = FALSE
	)
	mtext(side=3, text=expression(paste(CO[2], " bubbling and seawater mixing")), col=4, line=5, adj=0)
	mtext(side=3, text="Addition of strong acid", col=3, line=3.5, adj=0)
	mtext(side=3, text=expression(paste("Addition of ", HCO[3]^"-", " and strong acid")), col="orange" , line=2, adj=0)
	mtext(side=3, text=expression(paste("Addition of ", CO[3]^"2-", " and strong acid")), col=2 , line=0.5, adj=0)
}
 return(OUT)
}

# Copyright (C) 2017 Samir Alliouane, Lydia Kapsenberg, Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"sf_calib" <-
function(calEint=0.0865, calEext=-0.93, calpH=8.132, calT=16.2, calSal=35.6){

nK <- max(length(calEint), length(calEext), length(calpH), length(calT), length(calSal))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)
    
    if(length(calEint)!=nK){calEint <- rep(calEint[1], nK)}
    if(length(calEext)!=nK){calEext <- rep(calEext[1], nK)}
    if(length(calpH)!=nK){calpH <- rep(calpH[1], nK)}
    if(length(calT)!=nK){calT <- rep(calT[1], nK)}
    if(length(calSal)!=nK){calSal <- rep(calSal[1], nK)}
    
    # --------------------- sf_calib ----------------------------
    
    # Gas constant, Faraday constant, 
    R <- 8.3145
    FF <- 96487 
    # Temperature dependence oF standard potentials, Martz et al. 2010
    dE0int <- -0.001101
    dE0ext <- -0.001048
    # See Martz et al. 2010 For greater detail
    tempK <- calT + 273.15 # Convert temp From C to K
    S_T <- (R*tempK)/FF*log(10) # Nernst temp dependence
    
    E0int <- calEint-S_T*calpH # Calc E0int From Nernst & pH @ calibration point
    E0int25 <- E0int+dE0int*(25-calT)
    
    Z <- 19.924*calSal/(1000-1.005*calSal) # Ionic strength, Dickson et al. 2007
    SO4_tot <- (0.14/96.062)*(calSal/1.80655)  # Total conservative sulFate
    cCl <- 0.99889/35.453*calSal/1.80655 # Conservative chloride
    mCl <- cCl*1000/(1000-calSal*35.165/35) # mol/kg-H2O
    K_HSO4 <- exp(-4276.1/tempK+141.328-23.093*log(tempK) +
                    (-13856/tempK+324.57-47.986*log(tempK))*Z^0.5 + 
                    (35474/tempK-771.54+114.723*log(tempK))*Z-2698/tempK*Z^1.5 +
                    1776/tempK*Z^2+log(1-0.001005*calSal)) # BisulFate equilibrium const., Dickson et al. 2007
    pHint_free <- calpH+log10(1+SO4_tot/K_HSO4)
    cHfree <- 10^(-pHint_free) # mol/kg-sw
    pHint_free <- pHint_free+log10((1000-calSal*35.165/35)/1000) # mol/kg-H2O
    mHfree <- 10^(-pHint_free) # mol/kg-H2O
    DHconst <- 0.00000343*calT^2+0.00067524*calT+0.49172143 # Debye-Huckel, Khoo et al. 1977
    log10gamma_HCl <- 2*(-DHconst*sqrt(Z)/(1+1.394*sqrt(Z))+(0.08885-0.000111*calT)*Z)
    aHfree_aCl <- mHfree*mCl*10^(log10gamma_HCl) 
    E0ext <- calEext+S_T*log10(aHfree_aCl)
    E0ext25 <- E0ext+dE0ext*(25-calT)
    
    E0 <- data.frame(E0int25, E0ext25)
    return(E0)

    attr(sf_calib,"unit") = "Volts"
    return(sf_calib)
}

# Copyright (C) 2020 Kimberlee Baldry
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
#

sir <- function(flag, var1, var2, S = 35, T = 25, Patm = 1, P = 0, 
               Pt = 0, Sit = 0, k1k2 = "x", kf = "x", ks = "d", pHscale = "T", 
               b = "u74", gas = "potential", warn = "y", eos = "eos80", 
               long = 1e+20, lat = 1e+20){
  Carb_out = carb(flag, var1, var2, S , T , Patm , P , 
                   Pt, Sit , k1k2 , kf , ks , pHscale , 
                   b , gas , warn , eos , 
                   long , lat )
  ### pH conversions
  # carb allows multiple pH scale inputs
  # set up flag vectors
  free_flag <- NA
  total_flag <- NA
  sws_flag <- NA
  for(ph in 1:length(pHscale)){
    if(pHscale[ph] == "F"){
      free_flag[ph] <- NA
      total_flag[ph] <- 2
      sws_flag[ph] <- 6
          }
    if(pHscale[ph] == "T"){
      free_flag[ph] <- 4
      total_flag[ph] <- NA
      sws_flag[ph] <- 3
          }
    if(pHscale[ph] == "SWS"){
      free_flag[ph] <- 5
      total_flag[ph] <- 1
      sws_flag[ph] <- NA
          }}
  pH_f <- Carb_out$pH
  pH_t <- Carb_out$pH
  pH_sws <- Carb_out$pH
  #identify those that need conversion
  idx_f <- is.finite(free_flag)
  idx_t <- is.finite(total_flag)
  idx_sws <- is.finite(sws_flag)
  #convert only those not in scale
  if(length(which(idx_f)) != 0){pH_f[idx_f] = pHconv(flag = free_flag[idx_f], pH=Carb_out$pH[idx_f], S=Carb_out$S[idx_f], T=Carb_out$T[idx_f], P=Carb_out$P[idx_f], ks = ks)}
  if(length(which(idx_t)) != 0){pH_t[idx_t] = pHconv(flag = total_flag[idx_t], pH=Carb_out$pH[idx_t], S=Carb_out$S[idx_t], T=Carb_out$T[idx_t], P=Carb_out$P[idx_t], ks = ks)}
  if(length(which(idx_sws)) != 0){pH_sws[idx_sws] = pHconv(flag = sws_flag[idx_sws], pH=Carb_out$pH[idx_sws], S=Carb_out$S[idx_sws], T=Carb_out$T[idx_sws], P=Carb_out$P[idx_sws], ks = ks)}
  
  ### calculate H ions on different scales
  H_f <- (10^(-pH_f)*1000000) * 1000 / rho(S=Carb_out$S, T=Carb_out$T, P=Carb_out$P) #mol/L to mol/kg (devide by kg/L)
  H_t <- (10^(-pH_t)*1000000) * 1000 / rho(S=Carb_out$S, T=Carb_out$T, P=Carb_out$P) #mol/L to mol/kg (devide by kg/L)
  H_sws <- (10^(-pH_sws)*1000000) * 1000 / rho(S=Carb_out$S, T=Carb_out$T, P=Carb_out$P) #mol/L to mol/kg (devide by kg/L)
  
  ### SIR calculation
  sir <- (Carb_out$HCO3) / H_f
  
  
  ### output
  RES <- data.frame(flag, Carb_out$S, Carb_out$T, Carb_out$Patm, Carb_out$P, Carb_out$pH, Carb_out$CO2, Carb_out$fCO2, Carb_out$pCO2, 
                    Carb_out$fCO2pot, Carb_out$pCO2pot, Carb_out$fCO2insitu, Carb_out$pCO2insitu, Carb_out$HCO3, Carb_out$CO3, 
                    Carb_out$DIC, Carb_out$ALK, Carb_out$OmegaAragonite, Carb_out$OmegaCalcite, sir, H_f,H_t,H_sws,pH_f,pH_sws,pH_t)
  names(RES) <- c("flag", "S", "T", "Patm", "P", "pH", "CO2", 
                  "fCO2", "pCO2", "fCO2pot", "pCO2pot", "fCO2insitu", "pCO2insitu", 
                  "HCO3", "CO3", "DIC", "ALK", "OmegaAragonite", "OmegaCalcite","SIR","H_free","H_sws","H_t","pH_free","pH_sws","pH_t")
  return(RES)
}

# Copyright (C) 2009 Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"psi" <-
function(flag, var1, var2, S=35, T=20, Patm=1, P=0, Pt=0, Sit=0, pHscale="T", kf="x", k1k2="x", ks="d", eos="eos80", long=1.e20, lat=1.e20){
  # if the concentrations of total silicate and total phosphate are NA
  # they are set at 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  
  buf <- buffer(flag=flag, var1=var1, var2=var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, pHscale=pHscale, kf=kf, k1k2=k1k2, ks=ks, eos=eos, long=long, lat=lat)
	psi <- -buf$PiC/buf$PiD
out <- psi
attr(out,"unit") <- "mol CO2/mol CaCO3"
return(out)
}

# Copyright (C) 2010 Jean-Pierre Gattuso and Heloise Lavigne
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"Pcorrect" <-
function(Kvalue, Ktype, T=25, S=35, P=0, pHscale="T", kconv2ScaleP0="x", kconv2Scale="x", warn="y")
{
  #Pcoeffs <- get("Pcoeffs", envir  = environment()) # added to silence the CRAN note "Pcorrect: no visible binding for global variable ‘Pcoeffs’"
  nK <- max(length(Kvalue), length(Ktype), length(P), length(T), length(pHscale), length(S), length(kconv2ScaleP0), length(kconv2Scale))

    ##-------- Creation de vecteur pour toutes les entrees (si vectorielles)

    if(length(Kvalue)!=nK){Kvalue <- rep(Kvalue[1], nK)}
    if(length(Ktype)!=nK){Ktype <- rep(Ktype[1], nK)}
    if(length(T)!=nK){T <- rep(T[1], nK)}
    if(length(S)!=nK){S <- rep(S[1], nK)}
    if(length(P)!=nK){P <- rep(P[1], nK)}
    if(length(pHscale)!=nK){pHscale <- rep(pHscale[1], nK)}

    # If conversion factor at zero pressure is given 
    if (! missing(kconv2ScaleP0))
        # Check length of given conversion factor vector
        if (length(kconv2ScaleP0) != nK)
            kconv2ScaleP0 <- rep(kconv2ScaleP0[1], nK)
            
    # If conversion factor at given pressure is given 
    if (! missing(kconv2Scale))
        if (length(kconv2Scale) != nK)
            kconv2Scale <- rep(kconv2Scale[1], nK)
            

    # Constants 
    TK <- T + 273.15   # Temperature in Kelvin
    R = 83.14472;             # mol bar deg-1
    	
    ## loading table with coefficients
   # if(!exists("Pcoeffs", where = .GlobalEnv)) data(Pcoeffs)

    # ------------------- Pressure effect --------------------------------
    
	## 
	##  here, pHscale given in argument is the pH scale of Kvalue also given in argument
	##  if pHscale is different from seawater scale, it is necessary to convert K in seawater scale
	##  before applying the pressure correction, except with Kf, Kspc or Kspa.
	##  whith Kf, Kspc and Kspa pressure correction applied on free scale


    ##
    ## 1)
    ## Process  Ktype %in% c("K1", "K2", "K1p", "K2p", "K3p", "Kb", "Khs", "Kn", "Ksi", "K2si", "Kw")
    ## 

	# Indices of Kvalue elements where pressure correction to apply on seawater scale
    i_SWscale <- which (P > 0.0 & Ktype %in% c("K1", "K2", "K1p", "K2p", "K3p", "Kb", "Khs", "Kn", "Ksi", "K2si", "Kw"))     
    # if there is any such element
    if (length(i_SWscale) > 0)
    {
        # If conversion factor at zero pressure is not given 
        if (missing(kconv2ScaleP0) || kconv2ScaleP0[1] == "x")
        {
            ## ---------------- Compute this factor where pressure correction to apply on seawater scale
            
            # Initialise vector of conversion factor
            n_SWscale <- length(i_SWscale)
            conv <- rep(1., n_SWscale)
            
            # Indices in subvector Kvalue[i_SWscale] relevant to pHscale "T" 
            i_from_T <- which (pHscale[i_SWscale] == "T")
            if (length(i_from_T) > 0)
            {
                # Indices in initial vector Kvalue candidate to conversion from T to SWS
                i_from_T_SWS <- i_SWscale[i_from_T]             
                # compute conversion factor from Total pH scale to SW pH scale at zero pressure
                conv[i_from_T] <- kconv(S=S[i_from_T_SWS], T=T[i_from_T_SWS], P=0, warn=warn)$ktotal2SWS        
            }

            # Indices in subvector Kvalue[i_SWscale] relevant to pHscale "F" 
            i_from_F <- which (pHscale[i_SWscale] == "F")
            if (length(i_from_F) > 0)
            {
                # Indices in initial vector Kvalue candidate to conversion from F to SWS
                i_from_F_SWS <- i_SWscale[i_from_F] 
                # compute conversion factor from Free pH scale to SW pH scale at zero pressure
                conv[i_from_F] <- kconv(S=S[i_from_F_SWS], T=T[i_from_F_SWS], P=0, warn=warn)$free2SWS
            }        
        }
        else
        {
            # extract subset of relevant conversion factors
            conv <- kconv2ScaleP0[i_SWscale] 
        }
        
        # Apply conversion to SWS at zero pressure 
        Kvalue[i_SWscale] <- Kvalue[i_SWscale] * conv
    
        ## ------------------ Pressure correction
            
        l <- match(Ktype[i_SWscale], Pcoeffs$K)
        deltav  <-  Pcoeffs$a0[l] + Pcoeffs$a1[l] *T[i_SWscale] + Pcoeffs$a2[l] *T[i_SWscale]*T[i_SWscale]
        deltak  <-  Pcoeffs$b0[l]  + Pcoeffs$b1[l] *T[i_SWscale] + Pcoeffs$b2[l] *T[i_SWscale]*T[i_SWscale]
        lnkpok0 <-  -(deltav /(R*TK[i_SWscale]))*P[i_SWscale] + (0.5*deltak /(R*TK[i_SWscale]))*P[i_SWscale]*P[i_SWscale];
        Kvalue[i_SWscale] = Kvalue[i_SWscale]*exp(lnkpok0);
    
    
        # If conversion factor at given pressure is not given 
        if (missing(kconv2Scale) || kconv2Scale[1] == "x")
        {
            ## ---------------  Compute this factor where pressure correction is applied on seawater scale
            ##
    
            # Initialise vector of conversion factor
            n_SWscale <- length(i_SWscale)
            conv <- rep(1., n_SWscale)
            
            # Indices in subvector Kvalue[i_SWscale] relevant to pHscale "T" 
            i_from_T <- which (pHscale[i_SWscale] == "T")
            if (length(i_from_T) > 0)
            {
                # Indices in initial vector Kvalue candidate to conversion from T to SWS
                i_from_T_SWS <- i_SWscale[i_from_T]
                # compute conversion factor from Total pH scale to SW pH scale at given pressure
                conv[i_from_T] <- kconv(S=S[i_from_T_SWS], T=T[i_from_T_SWS], P=P[i_from_T_SWS], warn=warn)$ktotal2SWS        
            }
    
            # Indices in subvector Kvalue[i_SWscale] relevant to pHscale "F" 
            i_from_F <- which (pHscale[i_SWscale] == "F")
            if (length(i_from_F) > 0)
            {
                # Indices in initial vector Kvalue candidate to conversion from F to SWS
                i_from_F_SWS <- i_SWscale[i_from_F] 
                # compute conversion factor from Free pH scale to SW pH scale at given pressure
                conv[i_from_F] <- kconv(S=S[i_from_F_SWS], T=T[i_from_F_SWS], P=P[i_from_F_SWS], warn=warn)$free2SWS
            }        
        }
        else
        {
            # extract subset of relevant conversion factors
            conv <- kconv2Scale[i_SWscale] 
        }
        
        # Apply conversion from SWS to given pH scale 
        Kvalue[i_SWscale] <- Kvalue[i_SWscale] / conv
    }

    ##
    ## 2) 
    ## Process  Ktype %in% c("Ks", "Kspa", "Kspc", "Kf")
    ## 

	# Indices of Kvalue elements where pressure correction to apply on free pH scale
    i_Fscale <- which (P > 0.0 & Ktype %in% c("Ks", "Kspa", "Kspc", "Kf"))     
    # if there is any such element
    if (length(i_Fscale) > 0)
    {
        # There is no pH scale choice for 
        #    Ks -> "free scale"
        #    Kspa and Kspc do not need pH scale
          
        # For Kf, there is pH scale choice
        #   must convert from given scale to free pH scale
    
        # Indices in subvector Kvalue[i_Fscale] relevant to "Kf" 
        i_Kf <- which (Ktype[i_Fscale] == "Kf")
        # Indices in initial vector Ktype candidate to conversion to free scale
        i_Kf_to_free <- i_Fscale[i_Kf] 
        
        # if there is any such candidate
        if (length(i_Kf_to_free) > 0)
        {                 
            # If conversion factor at zero pressure is not given 
            if (missing(kconv2ScaleP0) || kconv2ScaleP0[1] == "x")
            {
                ## ---------------- Compute this factor where applying to Kf where P > 0.0
                
                # Initialise vector of conversion factor
                n_Kf_to_free <- length(i_Kf_to_free)
                conv <- rep(1., n_Kf_to_free)
                
                # Indices in subvector Kvalue[i_Kf_to_free] relevant to Kf on pHscale "T" 
                i_from_T <- which (pHscale[i_Kf_to_free] == "T")
                # if there is any such element in subvector
                if (length(i_from_T) > 0)
                {
                    # Indices in full vector Kvalue
                    i_from_T_full <- i_Kf_to_free[i_from_T]                
                  
                    # compute conversion factor from Total pH scale to Free pH scale at zero pressure
                    Cl = S[i_from_T_full] / 1.80655;  # Cl = chlorinity; S = salinity (per mille)
                    ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
                    FT = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
                    conv[i_from_T] <- 1./(1+ST/Ks(S=S[i_from_T_full], T=T[i_from_T_full], P=0))
                }
              
                # Indices in subvector Kvalue[i_Kf_to_free] relevant to Kf on pHscale "SWS" 
                i_from_SWS <- which (pHscale[i_SWscale] == "SWS")
                # if there is any such element in subvector
                if (length(i_from_SWS) > 0)
                {
                    # Indices in full vector Kvalue
                    i_from_SWS_full <- i_Kf_to_free[i_from_SWS]                
                  
                    # compute conversion factor from SW pH scale to Free pH scale at zero pressure
                    Cl = S[i_from_SWS_full] / 1.80655;  # Cl = chlorinity; S = salinity (per mille)
                    ST = 0.14 * Cl/96.062         # (mol/kg) total sulfate  (Dickson et al., 2007, Table 2)
                    FT = 6.7e-5 * Cl/18.9984    # (mol/kg) total fluoride (Dickson et al., 2007, Table 2)
                    conv[i_from_SWS] <- 1. / (1 + ST/Ks(S=S[i_from_SWS_full], T=T[i_from_SWS_full], P=0)
                                      + FT/Kf(T=T[i_from_SWS_full], P=0, S=S[i_from_SWS_full], pHscale="F"))
                }
            }
            else
            {
                # extract subset of relevant conversion factors
                conv <- kconv2ScaleP0[i_Kf_to_free] 
            }
            
            # Apply conversion at zero pressure (for Kf only)
            Kvalue[i_Kf_to_free] <- Kvalue[i_Kf_to_free] * conv
        }
            
        ## ------------------ Pressure correction
        
        l <- match(Ktype[i_Fscale], Pcoeffs$K)
        deltav  <-  Pcoeffs$a0[l] + Pcoeffs$a1[l] *T[i_Fscale] + Pcoeffs$a2[l] *T[i_Fscale]*T[i_Fscale]
        deltak  <-  Pcoeffs$b0[l]  + Pcoeffs$b1[l] *T[i_Fscale] + Pcoeffs$b2[l] *T[i_Fscale]*T[i_Fscale]
        lnkpok0 <-  -(deltav /(R*TK[i_Fscale]))*P[i_Fscale] + (0.5*deltak /(R*TK[i_Fscale]))*P[i_Fscale]*P[i_Fscale];
        
        Kvalue[i_Fscale] = Kvalue[i_Fscale]*exp(lnkpok0);
    }
    
    # Update pH scale information
    pHscale [ (P > 0.0) & (Ktype %in% c("Ks", "Kf")) ] <- "F" 

    phs <- rep(NA, nK)
    phs [pHscale == "SWS"] <- "Seawater scale"
    phs [pHscale == "T"]   <- "Total scale"
    phs [pHscale == "F"]   <- "Free scale"
    phs [Ktype %in% c("Kspa", "Kspc")] <- NA

    attr(Kvalue, "pH scale") <- phs
    attr(Kvalue, "unit") <- "mol/kg-soln"
    return(Kvalue)
}

# Copyright (C) 2010 Heloise Lavigne, Andreas Anderson and Jean-Pierre Gattuso
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#


"Om" <- function(x, flag, var1, var2, k1k2='x', kf='x', ks="d", pHscale="T", b="u74"){

c <- carb(flag=flag, var1=var1, var2=var2, S=35, T=25, P=0, Pt=0, Sit=0, k1k2= k1k2, kf= kf, ks=ks,  pHscale= pHscale, b=b)

S=35
x <- x*100  #conversion to percentage

#Activity coefficients Millero and Pierrot 1998
gCO3 <- 0.037
gCa <- 0.191
gMg <- 0.2029

Mg <- 0.05282*S/35   #Mg2+ concentration for a seawater at S=35 Dickson 2007
Ca <- 0.01028*S/35    #Dickson 2007
CO3 <- c$CO3

aCa <- gCa*Ca
aMg <- gMg*Mg
aCO3 <- gCO3*CO3

## to find the good IAP (equations derived from Bishoff, 1993)
IAPbiogenic <- 10^(-8.45022726020772 + -0.0363899891218405*x^1 + 0.0107168547236804*x^2 + -0.000307011230719308*x^3)
IAPcleaned <- 10^(-8.34594750694325 + -0.0492826022095371*x^1 + 0.0125595815829352*x^2 + -0.0010724925092009*x^3 + 4.33859823603212e-05*x^4 + -6.69727692992637e-07*x^5)
x <- x/100 # reconversion to proportion between 0 and 1
omegaBiogenic <- (aMg^x)*(aCa^(1-x))*(aCO3)/IAPbiogenic
omegaCleaned <- (aMg^x)*(aCa^(1-x))*(aCO3)/IAPcleaned
RES <- list(OmegaMgCa_biogenic=omegaBiogenic, OmegaMgCa_biogenic_cleaned=omegaCleaned)
return(RES)
}

# errors()
# This subroutine propagates errors to computed carbonate system variables 
# from errors (uncertainties) in six input variables 
#  - pair of carbonate system variables 
#  - nutrients (silicate and phosphate concentrations)
#  - temperature and salinity
# and from errors in dissociation constants pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
#
# It propagates error from input to output variables using one of two methods: 
#    * Gaussian (the method of moments without covariance term)
#      It is a very general technique for estimating the second moment of a variable z
#      (variance or standard deviation) based on a first-order approximation to z.
#      This is the default method.
#
#    * Monte Carlo method
#      It relies on repeated random sampling on input errors to obtain numerical results
#      from which standard error is estimated
#
# Input parameters :
#   - evar1, evar2   :  standard error (uncertainty) in var1 and var2 of input pair of carbonate system variables
#   - eS, eT         :  standard error (uncertainty) in Salinity and Temperature (default value = 0.01)
#   - ePt, eSit      :  standard error (uncertainty) in Phosphorus and Silicon total inorganic concentrations
#   - epK            :  standard error (uncertainty) in all 7 dissociation constants (a vector)
#   - method         :  case insensitive character string : "ga", "mo", or "mc"
#                       default is "ga" (gaussian)
#   - r              :  correlation coefficient (from -1 to 1) between var1 and var2 (only for Method of moments')
#   - runs           :  number of random samples (for Monte Carlo method only); default is 10000
#   - others         :  same as input of subroutine carb(): scalar or vectors
#
# All parameters may be scalars or vectors except epK, method, runs and gas.
#   * method, runs and gas must be scalar
#   * epK must be vector of seven values : errors of pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
#     these errors are assumed to be equal for all input data.
#
# In constrast, evar1, evar2, eS, eT, ePt and eSit, 
#   - if vectors, are errors associated with each data point
#   - if scalars, are one error value associated to all data points
#
# Returns a 2-dimensional data-frame, with the folowing columns :
#   - H      total error to [H+] concentration  (mol/kg)
#   - pH     total error to pH
#   - CO2    total error to CO2 concentration (mol/kg)
#   - pCO2   total error to "standard" pCO2, CO2 partial pressure computed at in situ temperature and atmospheric pressure (µatm)
#   - fCO2   total error to "standard" fCO2, CO2 fugacity computed at in situ temperature and atmospheric pressure (µatm)
#   - HCO3   total error to HCO3 concentration (mol/kg)
#   - CO3    total error to CO3 concentration (mol/kg)
#   - DIC    total error to DIC concentration (mol/kg)
#   - ALK    total error to ALK, total alkalinity (mol/kg)
#   - OmegaAragonite  total error of Omega aragonite (aragonite saturation state)
#   - OmegaCalcite    total error of Omega calcite   (calcite saturation state)
#
# Details :
#   Computation time depends on the chosen method, with the Monte Carlo approach being more computationally intensive
#
#   The Monte Carlo method computation time is proportional to the number of runs.
#   But more runs means more accurate results :
#      runs = 10000 appears to be the minimum needed to obtain results with an accuracy of less than 1%
#      Accuracy is inversely proportional to the number of runs.
#
#   Computation time also depends on the type of input pair of variables.
#     For example, the input pair DIC and Alk (flag=15) requires much more computational time 
#     than does the input pair pH and Alkalinity (flag=8).
#
errors <- 
function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, 
         evar1=0, evar2=0, eS=0.01, eT=0.01, ePt=0, eSit=0, epK=c(0.002, 0.0075, 0.015, 0.01, 0.01, 0.02, 0.02),
         eBt=0.02,
         method="ga", r=0, runs=10000, 
         k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", warn="y", eos="eos80", long=1.e20, lat=1.e20)
{
  # if the concentrations of total silicate and total phosphate are NA
  # they are set to 0
  Sit[is.na(Sit)] <- 0
  Pt[is.na(Pt)] <- 0
  
  # Input checking
    # --------------
    
    if (! method %in% c("ga", "mo", "mc"))
        stop ("Invalid input parameter: ", method)

    # Only two options for eos
    if (eos != "teos10" && eos != "eos80")
        stop ("invalid parameter eos: ", eos)

    # Input conditioning
    # -------------------
    
    n <- max(length(flag), length(var1), length(var2), length(S), length(T), length(P), length(Pt), length(Sit), 
             length(evar1), length(evar2), length(r), length(eS), length(eT), length(ePt), length(eSit),
             length(k1k2), length(kf), length(pHscale), length(ks), length(b))
    if(length(flag)!=n){flag <- rep(flag[1],n)}
    if(length(var1)!=n){var1 <- rep(var1[1],n)}
    if(length(var2)!=n){var2 <- rep(var2[1],n)}
    if(length(S)!=n){S <- rep(S[1],n)}
    if(length(T)!=n){T <- rep(T[1],n)}
    if(length(Patm)!=n){Patm <- rep(Patm[1],n)}
    if(length(P)!=n){P <- rep(P[1],n)}
    if(length(Pt)!=n){Pt <- rep(Pt[1],n)}
    if(length(Sit)!=n){Sit <- rep(Sit[1],n)}
    if(length(evar1)!=n){evar1 <- rep(evar1[1],n)}
    if(length(evar2)!=n){evar2 <- rep(evar2[1],n)}
    if(length(r)!=n){r <- rep(r[1],n)}
    if(length(eS)!=n){eS <- rep(eS[1],n)}
    if(length(eT)!=n){eT <- rep(eT[1],n)}
    if(length(ePt)!=n){ePt <- rep(ePt[1],n)}
    if(length(eSit)!=n){eSit <- rep(eSit[1],n)}
    if(length(k1k2)!=n){k1k2 <- rep(k1k2[1],n)}
    if(length(kf)!=n){kf <- rep(kf[1],n)}
    if(length(ks)!=n){ks <- rep(ks[1],n)}
    if(length(pHscale)!=n){pHscale <- rep(pHscale[1],n)}
    if(length(b)!=n){b <- rep(b[1],n)}

    # Correct sign of all input errors
    evar1 <- abs(evar1)
    evar2 <- abs(evar2)
    eS    <- abs(eS)
    eT    <- abs(eT)
    ePt   <- abs(ePt)
    eSit  <- abs(eSit)
    eBt   <- abs(eBt)
    
    # if epK=NULL, set all pK errors to zero
    if(is.null(epK)) {epK = rep(0, 7)}
  
    # Default value for epK
    if (missing(epK))
    {
      epK <- c(0.002, 0.0075, 0.015, 0.01, 0.01, 0.02, 0.02)    }
    else
    {
        # Check validity of epK
        if (length(epK) == 1 && epK == 0)
        {
            # this means that the caller does not want to account for errors on dissoc. constants
            epK <- rep(0, 7)
        }
        else if (length(epK) != 7)
            stop ("invalid parameter epK: ", epK)
        else
        {
            # Correct sign of given epK
            epK <- abs(epK)
        }
    }
    
    if (method == "ga")
    {
        r0 <- r*0.0
        errs <- .errors_ga (flag, var1, var2, S, T, Patm, P, Pt, Sit, evar1, evar2, r0, eS, eT, ePt, eSit,
                epK, eBt, k1k2, kf, ks, pHscale, b, gas, warn, eos=eos, long=long, lat=lat)
    }
    else if (method == "mo")
    {
        errs <- .errors_ga (flag, var1, var2, S, T, Patm, P, Pt, Sit, evar1, evar2, r, eS, eT, ePt, eSit,
                epK, eBt, k1k2, kf, ks, pHscale, b, gas, warn, eos=eos, long=long, lat=lat)
    }
    else if (method == "mc")
    {
        errs <- .errors_mc (flag, var1, var2, S, T, Patm, P, Pt, Sit, evar1, evar2, eS, eT, ePt, eSit,
                epK, eBt, k1k2, kf, ks, pHscale, b, gas, runs, warn, eos=eos, long=long, lat=lat)
    }

    return (errs)
}

#==============================================================================================
#                                                                                             #
#    1st method : Gaussian (r=0) or Method of moments (r nonzero, i.e. with covariance terms) #
#                                                                                             #
#==============================================================================================

# .errors_ga()
#
# This routine estimates uncertainties in computed carbonate system variables 
# by propagating errors (uncertainties) in the six input variables, including 
#  - the pair of carbonate system variables, 
#  - the 2 nutrients (total dissolved inorganic silicon and phosphorus concentrations), and
#  - temperature and salinity, as well as
# the errors in dissociation constants pK0, pK1, pK2, pKb, pKw, pKspa and pKspc.
#
# It computes numerical derivatives then applies them to propagate errors using the Gaussian method
# The Gaussian method is the standard technique for estimating the second moment of a computed variable z
# (its variance or standard deviation) based on a first-order approximation to z.
# The Gaussian method is the same as the method of moments but without covariance terms.
#
# Input parameters :
#   - evar1, evar2   :  standard error (uncertainty) in var1 and in var2 of input pair of carbonate system variables
#   - r              :  correlation coefficient between var1 and var2 
#   - eS, eT         :  standard error (uncertainty) in salinity and in temperature
#   - ePt, eSit      :  standard error (uncertainty) in total dissolved inorganic phosphorus and in total dissolved inorganic silicon concentrations
#   - epK            :  standard error (uncertainty) in the 7 dissociation constants mentioned above (a vector)
#   - others         :  same as input of subroutine  carb() : scalar or vectors
#
# All parameters may be scalars or vectors except epK and gas.
#   * gas must be a scalar
#   * epK must be vector of 7 values : errors of pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
#     these errors are assumed to be the same for all input data.
#
# In contrast, evar1, evar2, eS, eT, ePt and eSit, 
#   - if vectors, are errors associated with each data point
#   - if scalars, are one error value associated to all data points
#
# Returns a 2-dimensional data-frame, with the folowing columns :
# - H      total error to [H+] concentration  (mol/kg)
# - pH     total error to pH
# - CO2    total error to CO2 concentration (mol/kg)
# - fCO2   total error to "standard" pCO2, CO2 partial pressure computed at in situ temperature and atmospheric pressure (µatm)
# - pCO2   total error to "standard" fCO2, CO2 fugacity computed at in situ temperature and atmospheric pressure (µatm)
# - HCO3   total error to HCO3 concentration (mol/kg)
# - CO3    total error to CO3 concentration (mol/kg)
# - DIC    total error to DIC concentration (mol/kg)
# - ALK    total error to ALK, total alkalinity (mol/kg)
# - OmegaAragonite  total error of Omega aragonite (aragonite saturation state)
# - OmegaCalcite    total error of Omega calcite   (calcite saturation state)

.errors_ga <- 
function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, evar1=0, evar2=0, r=0, eS=0.01, eT=0.01,
         ePt=0, eSit=0, epK=NULL, eBt=NULL, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential",
         warn="y",
         eos="eos80", long=1.e20, lat=1.e20)
{
    # names of dissociation constants
    Knames <- c ('K0','K1','K2','Kb','Kw','Kspa', 'Kspc')

    # Constant table :  names of input pair variables sorted by flag number
    varnames  = rbind (
        c("H", "CO2"),            # flag = 1    
        c("CO2", "HCO3"),         # flag = 2    
        c("CO2", "CO3"),          # flag = 3    
        c("CO2", "ALK"),          # flag = 4    
        c("CO2", "DIC"),          # flag = 5    
        c("H", "HCO3"),           # flag = 6    
        c("H", "CO3"),            # flag = 7    
        c("H", "ALK"),            # flag = 8    
        c("H", "DIC"),            # flag = 9    
        c("HCO3", "CO3"),         # flag = 10   
        c("HCO3", "ALK"),         # flag = 11   
        c("HCO3", "DIC"),         # flag = 12   
        c("CO3", "ALK"),          # flag = 13   
        c("CO3", "DIC"),          # flag = 14   
        c("ALK", "DIC"),          # flag = 15 
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("pCO2", "H"),           # flag = 21   
        c("pCO2", "HCO3"),        # flag = 22   
        c("pCO2", "CO3"),         # flag = 23   
        c("pCO2", "ALK"),         # flag = 24   
        c("pCO2", "DIC")          # flag = 25   
    )
    var1name <- varnames[flag,1]
    var2name <- varnames[flag,2]

    # Convert error on pH to error on [H+] concentration
    isH <- (var1name == 'H')
    r[isH]  <- -1.0*r[isH]       # Inverse sign of 'r' if var1 is pH
    pH <- var1[isH]
    epH <- evar1[isH]       # Error on pH
    H  <- 10**(-pH)         # H+ concentration

    # dpH = d(-log10[H])
    #     = d(- ln[H] / ln[10] )
    #     = -(1/ln[10]) * d (ln[H])
    #     = -(1/ln[10]) * (dH / H)
    # Thus dH = - ln[1O] * [H] dpH
    eH <-  log(10) * H * epH   # the formula is del H = -log(10) H epH, but we drop minus sign here (errors are always positive)
    evar1[isH] <- eH

    # Same conversion for second variable
    isH <- (var2name == 'H')
    r[isH]  <- -1.0*r[isH]       # Inverse sign of 'r' if var2 is pH
    pH <- var2[isH]
    epH <- evar2[isH]       # Error on pH
    H  <- 10**(-pH)         # H+ concentration
        
    eH <-  log(10) * H * epH   
    evar2[isH] <- eH

    # initialise total square error
    n <- length(flag)
    sq_err <- numeric(n)
        
    # Contribution of var1 to squared standard error
    if (any (evar1 != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv1 <- derivnum ('1', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                            pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv1 * evar1
        sq_err <- sq_err + err * err
    }

    # Contribution of var2 to squared standard error
    if (any (evar2 != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv2 <- derivnum ('2', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                            pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv2 * evar2
        sq_err <- sq_err + err * err
    }

    # Contribution of covariance of var1 and var2 to squared standard error
    if (any (r != 0.0) & any (evar1 != 0.0) & any (evar2 != 0.0))
    {
        # Compute covariance from correlation coeff. and standard deviations
        covariance <- r * evar1 * evar2
        # Contribution to squared error
        err2 <- 2 * deriv1 * deriv2 * covariance
        sq_err <- sq_err + err2
    }

    # Contribution of Silicon (total dissolved inorganic concentration) to squared standard error
    if (any (eSit != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv <- derivnum ('sil', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                           pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv * eSit
        sq_err <- sq_err + err * err
    }

    # Contribution of Phosphorus (total dissolved inorganic concentration) to squared standard error
    if (any (ePt != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv <- derivnum ('phos', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                           pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv * ePt
        sq_err <- sq_err + err * err
    }

    # Contribution of T (temperature) to squared standard error
    if (any (eT != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv <- derivnum ('T', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                           pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv * eT
        sq_err <- sq_err + err * err
    }

    # Contribution of S (salinity) to squared standard error
    if (any (eS != 0.0))
    {
        # Compute sensitivities (partial derivatives)
        deriv <- derivnum ('S', flag, var1, var2, S=S, T=T, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                           pHscale=pHscale, b=b, gas=gas, warn=warn, eos=eos, long=long, lat=lat)
        err <- deriv * eS
        sq_err <- sq_err + err * err
    }
    
    # Salinity and Temperature converted to EOS-80 (if necessary)
    SP    <- rep(NA, n)
    InsT  <- rep(NA, n)
 
    # if use of TEOS-10 standard
    if (eos == "teos10")
    {
        # Must convert temperature and salinity from TEOS-10 to EOS-80
        # convert temperature: from Conservative (CT) to in-situ temperature
        # and salinity from Absolute to Practical (SP)
        STeos <- teos2eos_geo (S, T, P, long, lat)
        InsT <- STeos$T
        SP <- STeos$SP
    }
    else
    {
        InsT <- T
        SP <- S
    }

    # Preliminary calculations for dissociation constants
    if (any (epK != 0))
    {

        # Ks (free pH scale) at zero pressure and given pressure
        Ks_P0 <- Ks(S=SP, T=InsT, P=0, ks=ks, warn=warn)
        Ks    <- Ks(S=SP, T=InsT, P=P, ks=ks, warn=warn)

        # Kf on free pH scale
        Kff_P0 <- Kf(S=SP, T=InsT, P=0, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
        Kff <- Kf(S=SP, T=InsT, P=P, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
        # Kf on given pH scale
        Kf <- Kf(S=SP, T=InsT, P=P, pHscale=pHscale, kf=kf, Ks_P0, Ks, warn=warn)

        # Conversion factor from total to SWS pH scale at zero pressure
        ktotal2SWS_P0 <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks_P0,Kff=Kff_P0,warn=warn)$ktotal2SWS

        # Conversion factor from SWS to chosen pH scale
        conv <- kconv(S=SP,T=InsT,P=P,kf=kf,Ks=Ks,Kff=Kff,warn=warn)
        kSWS2chosen <- rep(1.,n)
        kSWS2chosen [pHscale == "T"] <- conv$kSWS2total [pHscale == "T"]
        kSWS2chosen [pHscale == "F"] <- conv$kSWS2free [pHscale == "F"]  
    }
        
    # Contribution of all pKi to squared standard error
    for (i in 1:length(epK))
    {
        # if error on Ki is given
        if (epK[i] != 0.0)
        {
            # Compute Ki
            Ki <- switch (i,
                          K0(S=SP, T=InsT, Patm=Patm, P=0, warn=warn),
                          K1(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn),
                          K2(S=SP, T=InsT, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0, warn=warn),
                          Kb(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0, warn=warn),
                          Kw(S=SP, T=InsT, P=P, pHscale=pHscale, kSWS2chosen, warn=warn),
                          Kspa(S=SP, T=InsT, P=P, warn=warn),
                          Kspc(S=SP, T=InsT, P=P, warn=warn)
                          )
            # compute error on Ki from that on pKi
            eKi <- - epK[i] * Ki * log(10)

            # Compute sensitivities (partial derivatives)
            # No need to pass option "eos" since conversion to eos has been done already
            deriv <- derivnum (Knames[i], flag, var1, var2, S=SP, T=InsT, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                pHscale=pHscale, b=b, gas=gas, warn=warn)
            err <- deriv * eKi
            sq_err <- sq_err + err * err
        }
    }

    # Contribution of Total Boron (computed from Bt/S ratio) to squared standard error
    if (eBt != 0.0)
    {
        # Compute sensitivities (partial derivatives)
        deriv <- derivnum ('bor', flag, var1, var2, S=SP, T=InsT, Patm=Patm, P=P, Pt=Pt, Sit=Sit, k1k2=k1k2, kf=kf, ks=ks, 
                           pHscale=pHscale, b=b, gas=gas, warn=warn)
        # err is derivative * absolute error in boron (i.e., eBt=0.02 is a 2% error)
        err <- deriv * bor(S=S, b=b) * eBt  
        sq_err <- sq_err + err * err
    }

    
    # Compute resulting total error (or uncertainty)
    error <- sqrt (sq_err)

    return (error)
}


#===========================================================================================
#                                                                                          #
#    2nd method : Monte Carlo                                                              #
#                                                                                          #
#===========================================================================================

# Function that computes standard deviation 
# on first dimension of a given 3D matrix "x"
.Sd_3D_1stD <- function(x)
{          
    first_dim <- dim(x)[1]
    # Compute mean over first dimension
    x_mean <- colMeans(x)
    # x_mean is a 2D array
    # Extend array x_mean to 3D by replication
    new_dim <- c(dim(x_mean), first_dim)
    x_mean <- array (x_mean, new_dim)
    # Transpose extended x_mean so that it conforms to array x
    x_mean <- aperm (x_mean, c(3,1,2))

    # Compute mean-centered array
    y <- x - x_mean
    # Compute variance estimator
    var <- colMeans(y^2)*(first_dim/(first_dim-1))

    return (sqrt(var))
}


# Function that generates deviate values for dissoc. constants Kx from given error on pKx (and same for Bt from eBt)
#
# Special case for K0 :
#    This function generates a set of small deltas departing from 0 whose distribution is close to normal
#    These deltas will then be used to generate deviate samples of K0, from standard value of K0
#
# Other cases : K1, K2, Kb, Kw, Kspa and Kspc
#    This function generates a set of deviate values for dissociation constant Kx
#    Distribution of that set is close to a Normal distribution centered on standard value of Kx
#
# Input parameters :
#   - epK      :  standard error (or uncertainty) on all seven dissociation constants (a vector of length 7)
#   - S, T, P  :  Salinity, Temperature and Pressure, vectors of same length n
#   - Patm     :  Surface atmospheric pressure in atm, vector of length n
#   - pHscale  :  pH scale, character ("T" for the total scale, "F" for the free scale and "SWS" for using the seawater scale)
#                 vector of length n
#   - k1k2     :  option for dissociation constants K1 and K2, vector of length n
#   - kf, ks   :  options for dissociation constants Kf and Ks, vectors of length n
#   - runs     :  number of runs of Monte Carlo (= number of deviates per sample)
#
# Returns a 2-dimensional data-frame, with 7 columns: K0, K1, K2, Kb, Kw, Kspa and Kspc 
#   each column contains deviate delta values of one dissociation constant
#   column length is n * runs
#
.gen_delta_Kx <- function (epK, eBt, S, T, P, Patm, pHscale, k1k2, kf, ks, b, runs, warn="y")
{
    n <- length(S)

    # For convenience, add eBt (fractional absolute err in Bt) to
    # end of epK vector (errors on K values in terms of pK)
    epKplus <- c(epK, eBt)

    # names of dissociation constants
    Kplusnames <- c ('K0','K1','K2','Kb','Kw','Kspa', 'Kspc', 'bor')

    # Initalise output data frame
    nrows <- n * runs
    delta_Kx <- data.frame (matrix( numeric(0), nrow=nrows, ncol=0))

    # Devise a function that generates simulation samples
    # for one variable with central value "val" and standard error "std_err"
    gen_sim <- function (val, std_err)
    {
        # if no standard error"
        if (all(std_err == 0))
            # Replicate variable value
            sim_array <- rep(val, times=runs)
        else
        {
            # Generate deviates from value using Normal distribution
            sim_array <- rnorm(runs, val, std_err)
            # Set to zero all negative values
            sim_array[sim_array < 0] = 0.0
        }
        return(sim_array)
    }

    # Preliminary calculations for dissociation constants

    # Ks (free pH scale) at zero pressure and given pressure
    Ks_P0 <- Ks(S=S, T=T, P=0, ks=ks, warn=warn)
    Ks    <- Ks(S=S, T=T, P=P, ks=ks, warn=warn)

    # Kf on free pH scale
    Kff_P0 <- Kf(S=S, T=T, P=0, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
    Kff <- Kf(S=S, T=T, P=P, pHscale="F", kf=kf, Ks_P0, Ks, warn=warn)
    # Kf on given pH scale
    Kf <- Kf(S=S, T=T, P=P, pHscale=pHscale, kf=kf, Ks_P0, Ks, warn=warn)

    # Conversion factor from total to SWS pH scale at zero pressure
    ktotal2SWS_P0 <- kconv(S=S,T=T,P=P,kf=kf,Ks=Ks_P0,Kff=Kff_P0,warn=warn)$ktotal2SWS

    # Conversion factor from SWS to chosen pH scale
    conv <- kconv(S=S,T=T,P=P,kf=kf,Ks=Ks,Kff=Kff, warn=warn)
    kSWS2chosen <- rep(1.,n)
    kSWS2chosen [pHscale == "T"] <- conv$kSWS2total [pHscale == "T"]
    kSWS2chosen [pHscale == "F"] <- conv$kSWS2free [pHscale == "F"]  

    # Convert error on pKi to error on Ki
    for (i in 1:length(epKplus))
    {
        # Compute Ki
        Ki <- switch (i,
                      seacarb::K0(S=S, T=T, Patm=Patm, P=0),
                      seacarb::K1(S=S, T=T, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0),
                      seacarb::K2(S=S, T=T, P=P, pHscale=pHscale, k1k2=k1k2, kSWS2chosen, ktotal2SWS_P0),
                      seacarb::Kb(S=S, T=T, P=P, pHscale=pHscale, kSWS2chosen, ktotal2SWS_P0),
                      seacarb::Kw(S=S, T=T, P=P, pHscale=pHscale, kSWS2chosen),
                      seacarb::Kspa(S=S, T=T, P=P),
                      seacarb::Kspc(S=S, T=T, P=P),
                      seacarb::bor(S=S, b=b)
                      )
        if (i == 1)
            center_value = 0.0    # special case for K0
        else
            center_value = Ki

        # if error on Ki is given
        if (epKplus[i] != 0.0)
        {
            # compute error (not signed) on Ki from that on pKi
            if ( i == 8 ) 
            {
                eKi <-  epKplus[i] * Ki
            } else {
                eKi <-  epKplus[i] * Ki * log(10)
            }
            # Generate deviate values for Ki or deltas for K0
            spl_Ki <- mapply (gen_sim, center_value, eKi)
            # Reshape samples from 2D matrix to vector
            dim(spl_Ki) <- c(n*runs)
        }
        else
        {
            # Simply replicate value of Ki or  replicate 0.0 for K0
            spl_Ki <- rep (center_value, each=runs)
        }

        # Store (deviate) values for Ki in a data frame column
        Kname <- Kplusnames[i]
        delta_Kx[[Kname]] <- spl_Ki
    }
    return (delta_Kx)
}


# errors()
# This subroutine does error propagation on the computation of carbonate system variables 
# from errors (or uncertainties) on six input 
#  - pair of carbonate system variables 
#  - nutrients (total inorganic silicon and phosphorus concentrations)
#  - temperature and salinity
# plus errors on dissociation constants pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
#
# It computes standard error on carbonate system output variables using Monte Carlo method.
#
# Input parameters :
#   - evar0, evar2   :  standard error (or uncertainty) on var1 and var2 of input pair of carbonate system variables
#   - eS, eT         :  standard error (or uncertainty) on Salinity and Temperature
#   - ePt, eSit      :  standard error (or uncertainty) on Phosphorus and Silicon total inorganic concentrations
#   - epK            :  standard error (or uncertainty) on all seven dissociation constants (a vector)
#   - eBt            :  standard error (or uncertainty) on total boron (relative factional error, e.g. eBt=0.02 is a 2% error in total boron

#   - runs           :  number of runs of Monte Carlo (= number of simulated samples)
#                       default is 10000
#   - others         :  same as input of subroutine  carb() : scalar or vectors
#
# All parameters may be scalars or vectors except epK, eBt, method, runs, and gas.
#   * runs must be a scalar
#   * gas and method must each be a character string
#   * epK must be a vector of seven values : errors of pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
#   * eBt must be a scalar : error in total boron, e.g. eBt=0.02 is a 2% error 
#     These 3 types of errors are assumed to be the same for all input data.
#
# In constrast, for evar1, evar2, eS, eT, ePt and eSit:
#   - if they are vectors, they represent errors associated with each data point
#   - if they are scalars, they represent one error value each associated to all data points
#
# Returns a 2-dimensional data-frame, with the folowing columns :
# - H      total error to [H+] concentration  (mol/kg)
# - pH     total error to pH
# - CO2    total error to CO2 concentration (mol/kg)
# - fCO2   total error to "standard" pCO2, CO2 partial pressure computed at in situ temperature and atmospheric pressure (µatm)
# - pCO2   total error to "standard" fCO2, CO2 fugacity computed at in situ temperature and atmospheric pressure (µatm)
# - HCO3   total error to HCO3 concentration (mol/kg)
# - CO3    total error to CO3 concentration (mol/kg)
# - DIC    total error to DIC concentration (mol/kg)
# - ALK    total error to ALK, total alkalinity (mol/kg)
# - OmegaAragonite  total error of Omega aragonite (aragonite saturation state)
# - OmegaCalcite    total error of Omega calcite   (calcite saturation state)
#
# Remarks :
#   Accuracy is strongly depending on number of runs. Default run number is 10,000, 
#   You may want to choose a number of 100,000 runs or more, but beware that such calculations may take several seconds per data point.
#
.errors_mc <- 
function(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, evar1=0, evar2=0, eS=0.01, eT=0.01, ePt=0, eSit=0,
         epK=NULL, eBt=NULL, k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", runs=10000, warn="y",
         eos="eos80", long=1.e20, lat=1.e20)
{
    # Constant table :  names of input pair variables sorted by flag number
    varnames  = rbind (
        c("H", "CO2"),            # flag = 1    
        c("CO2", "HCO3"),         # flag = 2    
        c("CO2", "CO3"),          # flag = 3    
        c("CO2", "ALK"),          # flag = 4    
        c("CO2", "DIC"),          # flag = 5    
        c("H", "HCO3"),           # flag = 6    
        c("H", "CO3"),            # flag = 7    
        c("H", "ALK"),            # flag = 8    
        c("H", "DIC"),            # flag = 9    
        c("HCO3", "CO3"),         # flag = 10   
        c("HCO3", "ALK"),         # flag = 11   
        c("HCO3", "DIC"),         # flag = 12   
        c("CO3", "ALK"),          # flag = 13   
        c("CO3", "DIC"),          # flag = 14   
        c("ALK", "DIC"),          # flag = 15 
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("", ""),
        c("pCO2", "H"),           # flag = 21   
        c("pCO2", "HCO3"),        # flag = 22   
        c("pCO2", "CO3"),         # flag = 23   
        c("pCO2", "ALK"),         # flag = 24   
        c("pCO2", "DIC")          # flag = 25   
    )

    # if use of EOS-10 standard
    if (eos == "teos10")
    {
        # Must convert temperature and salinity from TEOS-10 to EOS-80
        # convert temperature: from Conservative (CT) to in-situ temperature
        # and salinity from Absolute to Practical (SP)
        STeos <- teos2eos_geo (S, T, P, long, lat)
        InsT <- STeos$T
        SP <- STeos$SP
    }
    else
    {
        InsT <- T
        SP <- S
    }
    
    # Devise a function that generates simulation samples 
    # for one variable with central value "val" and standard error "std_err"
    gen_sim <- function (val, std_err)
    {
        # if no standard error"
        if (all(std_err == 0))
            # Replicate variable value
            sim_array <- rep(val, times=runs)
        else
        {
            # Generate deviates from value using Normal distribution
            sim_array <- rnorm(runs, val, std_err)
            # Set to zero all negative values
            sim_array[sim_array < 0] = 0.0
        }
        return(sim_array)
    }

    # Generate deviate sample values for var1
    spl_var1 <- mapply (gen_sim, var1, evar1)
    # Same for var2, Salinity, Temperature, total inorganic Phosphorus and Silicon
    spl_var2 <- mapply (gen_sim, var2, evar2)
    spl_S    <- mapply (gen_sim, S, eS)
    spl_T    <- mapply (gen_sim, T, eT)
    spl_Pt   <- mapply (gen_sim, Pt, ePt)
    spl_Sit  <- mapply (gen_sim, Sit, eSit)

    # Reshape samples from 2D matrix to vector
    n <- length(var1)
    dim(spl_var1) <- c(n*runs)
    dim(spl_var2) <- c(n*runs)
    dim(spl_S)    <- c(n*runs)
    dim(spl_T)    <- c(n*runs)
    dim(spl_Pt)   <- c(n*runs)
    dim(spl_Sit)  <- c(n*runs)
        
    # Generate deviate delta values for K0
    # Generate deviate values for other dissoc. constants Kx
    # Note : use of Salinity and Temperature converted to EOS-80 as they are when routine carb() computes dissociation constants  
    spl_Kx <- .gen_delta_Kx (epK, eBt, S=SP, T=InsT, P, Patm, pHscale, k1k2, kf, ks, b, runs, warn=warn)

    # All other parameters and variables
    spl_flag <- rep (flag, each=runs)
    spl_Patm <- rep (Patm, each=runs)
    spl_P    <- rep (P, each=runs)
    spl_k1k2 <- rep (k1k2, each=runs)
    spl_kf   <- rep (kf, each=runs)
    spl_ks   <- rep (ks, each=runs)
    spl_pHscale <- rep (pHscale, each=runs)
    spl_b    <- rep (b, each=runs)

    # Define local functions K0, K1, K2, ...
    # --------------------------------------
    # Special case for K0
    #
    # It computes original K0 values then add precalculated deltas to generate 
    # and return deviate values
    K0 <- function(S=35,T=25,P=0,Patm=1,...)
    #K0 <- function(S=35,T=25,P=0,Patm=1,warn='n')
    {
        # Call original K0 function
       #out <- seacarb::K0(S, T, P, Patm=Patm, warn=warn)
        out <- seacarb::K0(S, T, P, Patm=Patm)
        # perturb value of K0 by adding deltas
        out = out + spl_Kx$K0
        return (out)
    }
    # General case : all other
    # 'function' returns precalculated deviate values
    K1 <- function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0, warn="y")  spl_Kx$K1
    K2 <- function(S=35,T=25,P=0,k1k2='x',pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0, warn="y")  spl_Kx$K2
    Kw <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0, warn="y")  spl_Kx$Kw
    Kb <- function(S=35,T=25,P=0,pHscale="T",kSWS2scale=0,ktotal2SWS_P0=0, warn="y")  spl_Kx$Kb
    Kspa <- function(S=35,T=25,P=0, warn="y")  spl_Kx$Kspa
    Kspc <- function(S=35,T=25,P=0, warn="y")  spl_Kx$Kspc
    bor <- function(S=35,b="u74")  spl_Kx$bor

    # Note : in the general case (K1, K2,...) the function Kx is called once by the function carb()
    #        We can calculate in anticipation deviate values and substitute real values with deviate ones
    #        at run time.
    #        In the special case of K0, function K0() is called twice by function carb() 
    #        with two different parameter sets so that we cannot precalculate deviate values.
    #        We must generate deviate values at run time, when K0() is called.
    #        This is the purpose of locally defined K0() function
    
    # Save then change execution environment of function carb()
    saved_env <- environment(carb)
    # Using environment(K0) will hide original seacarb::K0, seacarb::K1, ... functions
    environment(carb) <- environment(NULL)
    # Note :  environment(NULL) is the environment created when this function .errors_mc() is executed
    #         It then contains all locally defined functions, like K0, K1, K2, ...
    #         Then, when carb() is executed and looks for a name (variable or function) that is not a local variable,
    #         it will look into evironment(NULL), which has been attached to it, 
    #         before looking into global or seacarb module environment.

    # Compute output carbonate system variables
    seacarb = carb(spl_flag, spl_var1, spl_var2, S=spl_S, T=spl_T, Patm=spl_Patm, P=spl_P, Pt=spl_Pt, Sit=spl_Sit, 
                spl_k1k2, spl_kf, spl_ks, spl_pHscale, spl_b, warn=warn, eos=eos, long=long, lat=lat)

    # Restore environment of carb()
    environment(carb) <- saved_env

    # Add one column for [H+]
    H <- 10^(-seacarb$pH)
    seacarb <- cbind(H,seacarb)

    # Drop unnecessary columns
    drops=c('flag', 'S', 'T', 'P', 'Patm', 'fCO2pot', 'pCO2pot', 'fCO2insitu', 'pCO2insitu')
    seacarb <- seacarb[,!(names(seacarb) %in% drops)]
        
    # if all input pairs are of same type
    if (all(flag == flag[1]))
    {
        # Drop input pair
        var1name <- varnames[flag[1],1]
        var2name <- varnames[flag[1],2]
        drops=c(var1name, var2name)
        # if to drop H, drop also pH
        if (var1name == 'H' || var2name == 'H')   drops = c(drops, 'pH')
        seacarb <- seacarb[,!(names(seacarb) %in% drops)]
    }
    
    # if more than one data point
    if (n > 1)
    {
        # Reshape results to a matrix
        seacarb_3D <- data.matrix(seacarb)
        # Reshape to a 3D array
        nvars <- dim(seacarb)[2]
        dim(seacarb_3D) <- c(runs, n, nvars)

        # Compute standard deviation on all simulated samples
        std_dev <- .Sd_3D_1stD (seacarb_3D)
        # Convert to data frame
        std_dev <- data.frame (std_dev)
        colnames(std_dev) <- colnames(seacarb)
    }
    else
    {
        # devise a function that computes standard deviation on columns of a given 2D matrix "x"
        colSd_2D <- function(x)sqrt(rowMeans((t(x)-colMeans(x))^2)*((dim(x)[1])/(dim(x)[1]-1)))

        # Compute standard deviation on all simulated samples
        std_dev <- data.frame(t(colSd_2D(seacarb)))
    }
        
    return (std_dev)
}

# Copyright (C) 2010 Jean-Pierre Gattuso and Héloïse Lavigne
#
# This file is part of seacarb.
#
# Seacarb is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or any later version.
#
# Seacarb is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with seacarb; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#


#Reference:  
#    Saunders, Peter M. (1981) Practical conversion of pressure
#     to depth, J. Phys. Oceanogr., 11, 573-574, (1981)
#
#     From FORTRAN code of

#     R. Millard
#     March 9, 1983
#     check value: p80=7500.004 dbars;for lat=30 deg., depth=7321.45 meters

#     Modified (slight format changes + added ref. details):
#     J. Orr, MEL-IAEA, 16 April 2009


"p2d" <- function (pressure, lat=40)

{
## Checking dimension of input data
npres <- length(pressure)
nlat <- length(lat)
nmax <- max(npres, nlat)
if(npres < nmax){pressure <- rep(pressure[1], nmax)}
if(nlat < nmax){lat <- rep(lat[1], nmax)}

pi <- 3.141592654
depth <- rep(NA, npres)

for(i in (1:nmax)){

## Checking for lat positive
lat[i] <- abs(lat[i])

plat <- abs(lat[i]*pi)/180 
d <- sin(plat)
c1 <- 5.92e-3 + d^2 * 5.25e-3

depth[i] <- ((1-c1)^2 - ((1-c1)-pressure[i]*4.42e-6)^2)/(8.84e-6)

}

return(depth)

}

\encoding{latin1}
\name{fH}
\alias{fH}
\title{Total activity coefficient for H+}
\description{Compute total hydrogen ion activity coefficient.
The activity coefficient (fH) is used to convert from H+ concentration on SWS scale 
to H+ activity (ah), as used for NBS scale; likewise fH is used to make the conversion 
in the opposite direction, from the NBS scale to the SWS scale. Here, fH is taken from 
Takahashi et al (1982, GEOSECS Pacific Expedition, Chap 3, p. 80) who say:
fH is the total activity coeff., which includes contributions from HSO4- 
and HF [as well as H+].

Takahashi et al. (1982) computed a relationship for fH based on the experimental data
from Culberson & Pytkowicz (1973), who determined it experimentally as a function
of temperature and salinity.  The approach is old and full of uncertainty.
Newer approaches are more complicated (Pitzer equations) 
and big uncertainties remain (Marion et al., 2011; Pilson, 2013).
}
\usage{
fH(S=35, T=25)
}
\arguments{
  \item{S}{Salinity on the practical salinity scale, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  }

\value{
  \item{fH}{Total activity coefficient for H+}
}

\details{
This total activity coefficient appears in
the following basic chemistry equation: \eqn{ ah = fH * hsws}, 
where  \eqn{ah}  is the activity of hydrogen ion,
\eqn{fH}  is the total activity coefficient, and
\eqn{hsws = [H+] + [HSO4-] + [HF]}. In other words, 
\eqn{hsws} is the total hydrogen ion conccentration on the seawater scale.

The two pH scales of concern are defined as \eqn{pHNBS = -log10(ah)} and 
\eqn{pHSWS = -log10(hsws)}.  
}

\references{
Culberson, C.H., & Pytkowicz, R.M. (1973). Ionization of water in seawater. 
\emph{Marine Chemistry}, \bold{1(4)}, 309-316.

Marion G.M., Millero F.J., Camoes M.F., Spitzer P., Feistel R., Chen C.T.A. 2011. pH of seawater. 
\emph{Marine Chemistry} \bold{126} 89-96.

Pilson M.E.Q. (2013) An introduction to the chemistry of the sea, 
2 edn. Cambridge, UK: Cambridge University Press.

Takahashi T., Williams R.T., and Ros D.L. (1982) Carbonate chemistry. 
GEOSECS Pacific Expedition, Volume 3, Hydrographic Data 1973-1974, 77-83.
}
             
\author{
James Orr \email{james.orr@lsce.ipsl.fr}
}

\seealso{
	\code{\link{pHnbs2sws}} and \code{\link{pHsws2nbs}}
}

\examples{
## Compute fH
   f = fH(T=25, S=35)
   print(f)
## Check value: The result is 0.7134043
}

\keyword{utilities}

\encoding{latin1}
\name{rho}
\alias{rho}

\title{Density of seawater  (kg/m3)}
\description{Calculates the density of seawater (\eqn{kg\ m^{-3}})}
\usage{
rho(S = 35, T = 25, P = 0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Practical Salinity (PSS-78), default is 35}
  \item{T}{Temperature in degrees Celsius (ITS-90), default is 25oC}
  \item{P}{Hydrostatic pressure in bar (surface = 0), default is 0}
}


\value{
  \item{rho}{Density of seawater (kg/m3)}

}
\references{Millero F. J. and Poisson A., 1981 International one-atmosphere equation of state of seawater. \emph{Deep-Sea Research} \bold{28A}, 625-629.}

\author{
Aurelien Proye and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}


\examples{
	rho(35,25,0)
}
\keyword{utilities}

\name{seaFET}

\docType{data}

\alias{seaFET}

\title{Test seaFET data file}

\description{
Short test file for using with functions sf_calc and sf_calib. It is an excerpt of a file produced by a SeaFET pH sensor.
}
\usage{seaFET}

\format{
A data frame with 8 variables (datetime, Eint, Eext, Salinity, Temperature, pHspectro, E0int25, E0ext25) and 10 rows
}

\keyword{datasets}

\encoding{latin1}
\name{bor}
\alias{bor}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Total boron concentration (mol/kg)}
\description{total boron concentration (\eqn{mol\ kg^{-1}})}
\usage{
bor(S, b)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity, default is 35}
  \item{b}{"l10" for using the formulation of Lee et al. (2010), "u74" for using the Uppstrom (1974), or "k18" for using the Kulinski et al. (2018), default is "u74" }
  }

\details{Note that the formulation of Kulinski et al. (2018) is specifically designed for the Baltic Sea. Three formulations are described in their paper:
\itemize{
  \item based on their measurements: TB = [umol/kg] = 10.838 * S + 13.821
  \item based on Kremling (1970 and 1972): TB [umol/kg] = 11.44 * S + 12.6; R2 = 0.95
  \item consensus regression (Kremling + their data): TB [umol/kg] = 11.405 * S + 11.869; R2 = 0.97s
}
The latter formulation is used here.
}

\value{
  \item{bor}{total boron concentration (\eqn{mol\ kg^{-1}}))}

}
\references{
DOE 1994 \emph{Handbook of methods for the analysis of the various parameters of the carbon dioxide system in sea water}. ORNL/CDIAC-74. Oak Ridge,Tenn.: Carbon Dioxide Information Analysis Center, Oak Ridge National Laboratory.

Kulinski K., Szymczycha B., Koziorowska K., Hammer K. & Schneider B., 2018. Anomaly of total boron concentration in the brackish waters of the Baltic Sea and its consequence for the CO2 system calculations. \emph{Marine Chemistry}. doi:s10.1016/j.marchem.2018.05.007.

Lee K., Tae-Wook K., Byrne R.H., Millero F.J., Feely R.A. and Liu Y-M, 2010 The universal ratio of the boron to chlorinity for the North Pacific and North Atlantic oceans. \emph{Geochimica et Cosmochimica Acta} \bold{74} 1801-1811.

Uppstrom L.R., 1974 The boron/chlorinity ratio of the deep-sea water from the Pacific Ocean. \emph{Deep-Sea Research I} \bold{21} 161-162.
}

\author{
Heloise Lavigne, Aurelien Proye and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}


\examples{
bor(35, "l10")
}

\keyword{utilities}

\encoding{latin1}
\name{pHsws2nbs}
\alias{pHsws2nbs}
\title{Converts pH from SWS to NBS scale}
\description{Converts pH on SWS scale to pH on NBS scale pCO2. 
The SWS-to-NBS conversion is done with the total activity coefficient fH (combined activity coeff for H+, HSO4-, and HF)
from Takahashi et al. (1982) based on data from Culberson and Pytkowicz (1973). The approach is old and full of uncertainty.
Newer approaches are more complicated (Pitzer equations) and big uncertainties remain (Marion et al., 2011; Pilson, 2013).
}
\usage{
pHsws2nbs(pHSWS, S=35, T=25)
}
\arguments{
  \item{pHSWS}{pH on SWS scale}
  \item{S}{Salinity on the practical salinity scale, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  }

\value{
  \item{pHNBS}{pH on NBS scale}
}

\details{
The pHNBS is computed from pHSWS using the total activity coefficient and 
relying on the following basic chemistry equation: \eqn{ ah = fH * hsws}, 
where  \eqn{ah}  is the activity of hydrogen ion,
\eqn{fH}  is the total activity coefficient, and
\eqn{hsws = [H+] + [HSO4-] + [HF]}. In other words, 
\eqn{hsws} is the hydrogen ion concentration on the seawater scale.

The two pH scales are defined as \eqn{pHNBS = -log10(ah)} and 
\eqn{pHSWS = -log10(hsws)}.  
}

\references{
Culberson, C.H., & Pytkowicz, R.M. (1973). Ionization of water in seawater. 
\emph{Marine Chemistry}, \bold{1(4)}, 309-316.

Marion G.M., Millero F.J., Camoes M.F., Spitzer P., Feistel R., Chen C.T.A. 2011. pH of seawater. 
\emph{Marine Chemistry} \bold{126} 89-96.

Pilson M.E.Q. (2013) An introduction to the chemistry of the sea, 
2 edn. Cambridge, UK: Cambridge University Press.

Takahashi T., Williams R.T., and Ros D.L. (1982) Carbonate chemistry. 
GEOSECS Pacific Expedition, Volume 3, Hydrographic Data 1973-1974, 77-83.
}
             
\author{
James Orr \email{james.orr@lsce.ipsl.fr}
}

\seealso{
	\code{\link{fH}} and \code{\link{pHnbs2sws}}
}

\examples{
## Convert pHSWS to pHNBS
   pHSWS = 7.853336                     # pH on the SWS scale
   pHNBS = pHsws2nbs(pHSWS, T=25, S=35) # pH on the NBS scale
   print(pHNBS)
## Check value: the result should be 8.0
}

\keyword{utilities}

\encoding{latin1}
\name{Pcorrect}
\alias{Pcorrect}
\title{Pressure correction of equilibrium constants}
\description{Computes the pressure correction of the equilibrium constants}
\usage{
Pcorrect(Kvalue, Ktype, T=25, S=35, P=0, pHscale="T", 
  kconv2ScaleP0="x", kconv2Scale="x", warn="y")
}

\arguments{
  \item{Kvalue}{Value of the constant at P=0 (hydrostatic pressure in bar, surface = 0)}
  \item{Ktype}{Name of the constant, 
	\itemize{
	\item K1	First dissociation constant of carbonic acid (mol/kg)
	\item K2	Second dissociation constant of carbonic acid (mol/kg)
	\item Kb	Dissociation constant of boric acid (mol/kg)
	\item Kw	Ion product of water (mol2/kg2)
	\item Ks	Stability constant of hydrogen sulfate (mol/kg)
	\item Kf	Stability constant of hydrogen fluoride (mol/kg)
	\item Kspc	Solubility product of calcite (mol/kg)
	\item Kspa	Solubility product of aragonite (mol/kg)
	\item K1p	First dissociation constant of phosphoric acid (mol/kg)
	\item K2p	Second dissociation constant of phosphoric acid (mol/kg)
	\item K3p	Third dissociation constant of phosphoric acid (mol/kg)
	\item Khs	Dissociation constant of hydrogen sufide(mol/kg)
	\item Kn	Dissociation constant of ammonium (mol/kg)
	\item Ksi	Dissociation constant of Si(OH)4 (mol/kg)
	\item K2si	Second dissociation constant of Si(OH)4 (mol/kg)
}}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{S}{Salinity, default is 35}
  \item{P}{Hydrostatic pressure in bar (surface = 0), default is 0}
  \item{pHscale}{pHscale of the constant given in Kvalue}
  \item{kconv2ScaleP0}{Conversion factor from the pH scale selected to the SWS (or free for Kf) scale at pressure zero. It is advised to use default value "x", in which case it is computed when required. (may slow down the computation)}
  \item{kconv2Scale}{Conversion factor from the pH scale selected to the SWS (or free for Kf) scale at the hydrostatic pressure value indicated. It is advised to use default value "x", in which case it is computed when required. (may slow down the computation)}
  \item{warn}{"y" to show warnings when T or S go beyond the valid range for constants; "n" to supress warnings. The default is "y".}
}    

\details{
  \itemize{
  \item The pressure correction is applied on the seawater scale for K1, K2, K1p, K2p, K3p, Kb, Khs, Kn, Ksi, K2si and Kw. Hence the K value is first converted on the seawater scale if needed. After pressure correction, the constant is converted back to the initial pH scale.
  \item The pressure correction is applied on the free scale for Kf.
  \item There is no issue of pH scale for Ks, Kspa and Kspc.
}
}

\value{
  The equilibrium constant given in argument but after pressure correction
}

\references{
Millero F. J., 1995 Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica et Cosmochimica Acta} \bold{59} 661-677.
}

\author{
  Heloise Lavigne and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
Pcoeffs
}

\examples{
k10 <- K1(T=25, P=0, S=35)
Pcorrect(Kvalue=k10, Ktype="K1", P=300, T=25, S=35, pHscale="T")
}

\keyword{utilities}

\encoding{latin1}
\name{buffergen}
\alias{buffergen}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{Buffer factors of the seawater carbonate system as defined by Hagens and 
  Middelburg (2016)}
\description{Returns the suite of buffer factors presented in Table 3 of Hagens and Middelburg 
  (2016), as well as the proton concentration buffer factor (beta.H of Hofmann et al, 2010) and 
  the classic Revelle factor. For practical purposes, this function excludes the nitrate and 
  nitrite acid-base systems presented in this paper, as well as the fully protonoted form of 
  sulfate (H2SO4) and fully deprotonated form of sulfide (S2-), as their contributions to total 
  alkalinity under natural seawater conditions are negligible. Its input arguments are 
  identical to those in the carbfull function of seacarb.}
\usage{buffergen(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, k1k2="x", kf="x",
                 ks="d", pHscale="T", b="u74", gas="potential", NH4t=0, HSt=0)}
%- maybe also 'usage' for other objects documented here.
\arguments{
	\item{flag}{select the couple of variables available. The flags which can be used are: 

flag = 1      pH and CO2 given

flag = 2      CO2 and HCO3 given

flag = 3      CO2 and CO3 given

flag = 4      CO2 and ALK given

flag = 5      CO2 and DIC given

flag = 6      pH and HCO3 given

flag = 7      pH and CO3 given

flag = 8      pH and ALK given

flag = 9      pH and DIC given

flag = 10     HCO3 and CO3 given

flag = 11     HCO3 and ALK given

flag = 12     HCO3 and DIC given

flag = 13     CO3 and ALK given

flag = 14     CO3 and DIC given

flag = 15     ALK and DIC given

flag = 21     pCO2 and pH given

flag = 22     pCO2 and HCO3 given

flag = 23     pCO2 and CO3 given

flag = 24     pCO2 and ALK given

flag = 25     pCO2 and DIC given
}
	\item{var1}{Value of the first  variable in mol/kg, except for pH and for pCO2 in \eqn{\mu}atm}
	\item{var2}{Value of the second  variable in mol/kg, except for pH}
	\item{S}{Salinity}
	\item{T}{Temperature in degrees Celsius}
  \item{Patm}{Surface atmospheric pressure in atm, default is 1 atm}
  \item{P}{Hydrostatic pressure in bar (surface = 0)}
  \item{Pt}{Concentration of total phosphate in mol/kg; set to 0 if NA}
  \item{Sit}{Concentration of total silicate in mol/kg; set to 0 if NA}
  \item{k1k2}{"cw" for using K1 and K2 from Cai & Wang (1998), "l" from Lueker et al. (2000), "m02" from Millero et al. (2002), "m06" from Millero et al. (2006), "m10" from Millero (2010), "mp2" from Mojica Prieto et al. (2002), "p18" from Papadimitriou et al. (2018), "r" from Roy et al. (1993), "sb21" from Shockman & Byrne (2021), "s20" from Sulpis et al. (2020), and "w14" from Waters et al. (2014). "x" is the default flag; the default value is then "l", except if T is outside the range 2 to 35oC and/or S is outside the range 19 to 43. In these cases, the default value is "w14".}
	\item{kf}{"pf" for using Kf from Perez and Fraga (1987) and "dg" for using Kf from Dickson and Riley (1979 in Dickson and Goyet, 1994). "x" is the default flag; the default value is then "pf", except if T is outside the range 9 to 33oC and/or S is outside the range 10 to 40. In these cases, the default is "dg".}
	\item{ks}{"d" for using Ks from Dickson (1990) and "k" for using Ks from Khoo et al. (1977), default is "d"} 
	\item{pHscale}{"T" for the total scale, "F" for the free scale and "SWS" for using the seawater scale, default is "T" (total scale)}
	\item{b}{Concentration of total boron. "l10" for the Lee et al. (2010) formulation or "u74" for the Uppstrom (1974) formulation, default is "u74" }
  \item{gas}{used to indicate the convention for INPUT pCO2, i.e., when it is an input variable (flags 21 to 25): "insitu" indicates it is referenced to in situ pressure and in situ temperature; "potential" indicates it is referenced to 1 atm pressure and potential temperature; and "standard" indicates it is referenced to 1 atm pressure and in situ temperature. All three options should give identical results at surface pressure. This option is not used when pCO2 is not an input variable (flags 1 to 15). The default is "potential" and should be a unique value..}
  \item{NH4t}{Concentration of total ammonium in mol/kg; set to 0 if NA}
  \item{HSt}{Concentration of total hydrogen sulfide in mol/kg; set to 0 if NA}
}

\details{The Lueker et al. (2000) constants for K1 and K2, the Perez and Fraga (1987) constant for Kf and the Dickson (1990) constant for Ks are recommended by Dickson et al. (2007). It is, however, critical to consider that each formulation is only valid for specific ranges of temperature and salinity:

\emph{For K1 and K2:}
\itemize{
\item Cai and Wang (1998): S ranging between 0 and 40 and T ranging between 0.2 and 30oC.

\item Lueker et al. (2000): S ranging between 19 and 43 and T ranging between 2 and 35oC.

\item Millero et al. (2002): S ranging from 34 to 37 and T ranging between -1.6 and 35oC.

\item Millero et al. (2006): S ranging between 0.1 and 50 and T ranging between 1 and 50oC.

\item Millero (2010): S ranging between 1 and 50 and T ranging between 0 and 50oC. Millero (2010) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale". 

\item Mojica Prieto et al. (2002): S ranging from 5 to 42 and T ranging between 0 and 45oC.

\item Papadimitriou et al. (2018): S ranging from 33 to 100 and T ranging between -6 to 25oC.

\item Roy et al. (1993): S ranging between 5 and 45 and T ranging between 0 and 45oC.

\item Shockman & Byrne (2021): for K2, S ranging from 19.6 to 41 and T ranging between 15 to 35oC. For K1, formulation is that of Waters et al.

\item Sulpis et al. (2020): S ranging from 30.7 to 37.6 and T ranging between -1.7 to 31.8oC.

\item Waters et al.(2014): S ranging between 1 and 50 and T ranging between 0 and 50oC. Waters (2014) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale".
}

\emph{For Kf:}
\itemize{
\item Perez and Fraga (1987): S ranging between 10 and 40 and T ranging between 9 and 33oC.

\item Dickson and Riley (1979 in Dickson and Goyet, 1994): S ranging between 0 and 45 and T ranging between 0 and 45oC.
}

\emph{For Ks:}
\itemize{
\item Dickson (1990): S ranging between 5 and 45 and T ranging between 0 and 45oC.

\item Khoo et al. (1977): S ranging between 20 and 45 and T ranging between 5 and 40oC.
}

The arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It is recommended to use either vectors with the same dimension or one vector for one argument and numbers for the other arguments.

\emph{Pressure corrections and pH scale:}

\itemize{
\item For K0, the pressure correction term of Weiss (1974) is used.

\item For K1, K2, pK1, pK2, pK3, Kw, Kb, Khs, Ksi and K2si, the pressure correction was applied on the seawater scale. Hence, if needed, values were first transformed from the total scale to the seawater scale, the pressure correction applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).

\item For Kf, the pressure correction was applied on the free scale. The formulation of Dickson and Riley (1979 in Dickson and Goyet, 1994) provides Kf on the free scale but that of Perez and Fraga (1987) provides it on the total scale. Hence, in that case, Kf was first transformed from the total scale to the free scale. With both formulations, the pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS). 

\item For Ks, the pressure correction was applied on the free scale. The pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS). 

\item For Kn, the pressure correction was applied on the seawater scale. The pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).
}
}

\value{The function returns a list containing the following matrices:
  \item{Carbfull}{Output of the carbfull function that is used within buffergen}
  \item{dALK.dH}{Sensitivity of ALK to a change in proton concentration (dimensionless). Species-specific.}
  \item{dtotX.dH}{Sensitivity of an acid-base species to a change in proton concentration (dimensionless). Species-specific.}
  \item{dALK.dX}{Sensitivity of ALK to a change in an acid-base species (dimensionless). Species-specific.}
  \item{dtotX.dX}{Sensitivity of an acid-base species to a change in its total concentration (dimensionless). Species-specific.}
  \item{dALK.dpH}{Sensitivity of ALK to a change in pH (mol/kg-soln). Species-specific.}
  \item{dtotX.dpH}{Sensitivity of an acid-species to a change in pH (mol/kg-soln). Species-specific.}
  \item{dH.dALK}{Sensitivity of proton concentration to a change in ALK (dimensionless). Values are the same for all species and all acid-base systems, except for the fluoride and sulfate acid-base systems, which slightly deviate due to pH scale conversion effects.}
  \item{dH.dtotX}{Sensitivity of an acid-species to a change in its total concentration (dimensionless). Values are the same for all species of a specific acid-base system.}
  \item{dX.dALK}{Sensitivity of an acid-species to a change in its total concentration (dimensionless). Species-specific.}
  \item{dX.dtotX}{Sensitivity of an acid-species to a change in its total concentration (dimensionless). Species-specific.}
  \item{dpH.dALK}{Sensitivity of pH due to a change in ALK ((mol/kg-soln)-1). Values are the same for all species and all acid-base systems, except for the fluoride and sulfate acid-base systems, which slightly deviate due to pH scale conversion effects.}
  \item{dpH.dtotX}{Sensitivity of pH due to a change in the total concentration of an acid-base system ((mol/kg-soln)-1). Values are the same for all species of a specific acid-base system.}
  \item{beta.H}{proton concentration buffer factor (Eq.4 of Hagens and Middelburg (2016), dimensionless)}   
  \item{RF}{Revelle factor (dimensionless)}
  If the total concentration of an acid-base system is 0, the values of the buffer factors corresponding to all species of that acid-base system return NA.
}

\references{
Hagens M. and Middelburg J. J., 2016 Generalised expressions for the response of pH to changes in ocean chemistry. \emph{Geochimica et Cosmochimica Acta} \bold{187} 334-349.
}

\author{
Mathilde Hagens \email{m.hagens@uu.nl}
}

\examples{

## With a couple of variables
buffergen(flag=8, var1=8.2, var2=0.00234, S=35, T=25, P=0, Patm=1.0, Pt=0, Sit=0,
	pHscale="T", kf="pf", k1k2="l", ks="d", b="u74", gas="potential", NH4t=0, HSt=0)

## With a couple of variables and non-zero nutrient concentrations
buffergen(flag=8, var1=8.2, var2=0.00234, S=35, T=25, P=0, Patm=1.0, Pt=5e-6, Sit=2e-6,
	pHscale="T", kf="pf", k1k2="l", ks="d", b="u74", gas="potential", NH4t=10e-6, HSt=0.1e-6)

## Using vectors as arguments
flag <- c(8, 2, 8)
var1 <- c(8.2, 7.477544e-06, 8.2)
var2 <- c(0.002343955, 0.001649802, 2400e-6)
S <- c(35, 35, 30)
T <- c(25, 25, 30)
P <- c(0, 0, 0)
Pt <- c(0, 0, 0)
Sit <- c(0, 0, 0)
kf <- c("pf", "pf", "pf")
k1k2 <- c("l", "l", "l")
pHscale <- c("T", "T", "T")
b <- c("l10", "l10", "l10")
gas <- c("potential")
NH4t <- c(0, 0, 0)
HSt <- c(0, 0, 0)
buffergen(flag=flag, var1=var1, var2=var2, S=S, T=T, P=P, Pt=Pt, Sit=Sit, 
  kf=kf, k1k2=k1k2, pHscale=pHscale, b=b, gas=gas, NH4t=NH4t, HSt=HSt)

## Test with all flags 
flag <- c((1:15), (21:25))
var1 <- c(8.200000, 7.308171e-06, 7.308171e-06, 7.308171e-06, 7.308171e-06, 
	8.2, 8.2, 8.2, 8.2, 0.001646857, 0.001646857, 0.001646857, 0.0002822957, 
	0.0002822957, 0.00234, 258.2164, 258.2164, 258.2164, 258.2164, 258.2164 )
var2 <- c(7.308171e-06, 0.001646857, 0.0002822957, 0.00234, 0.001936461, 
	0.001646857, 0.0002822957, 0.00234, 0.001936461, 0.0002822957, 
	0.00234, 0.001936461,  0.00234, 0.001936461, 0.001936461, 8.2, 
	0.001646857, 0.0002822957, 0.00234, 0.001936461)
buffergen(flag=flag, var1=var1, var2=var2)
}
\keyword{utilities}

\encoding{latin1}
\name{sf_calib}
\alias{sf_calib}

\title{Calibration coefficients for seaFET sensor}
\description{Calibration coefficients E0INT,25, E0EXT,25}
\usage{
sf_calib(calEint=0.0865, calEext=-0.93, calpH=8.132, calT=16.2, calSal=35.6)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{calEint}{EINT (V), default is 0.0865}
  \item{calEext}{EEXT (V), default is -0.93}
  \item{calpH}{spectrophotometric pH in Total scale, default is 8.132}
  \item{calT}{Temperature in degrees Celsius, default is 16.2}
  \item{calSal}{Salinity, default is 35.6}
}

\details{
Outputs E0INT25 and E0EXT25 must be calculated for each reference sample collected during a SeaFET deployment. Multiple E0INT25 and E0EXT25 may be calculated if there is more than one reference sample for a given deployment. As such, arguments can be given as unique numbers or as vectors (vectors should be of the same length).

It is critical that Eint (calEint) and Eext (calEext) recorded by the SeaFET match reference sample measurements of temperature (calT), salinity (calSal), and spectrophotometric pH (calpH, total hydrogen ion scale) taken at the same time. Note that SeaFET temperature measurements may require calibration via an applied offset. When possible, calibrated CTD temperature and salinity measurements may be used, while spectrophotometric pH measurements always require discrete 'reference' water samples (unless in situ, certified, seawater-based, Tris pH buffer is used). The accepted time offset between collection of reference samples and SeaFET measurements depends on the hydrology and pH variability of the location. For detailed SeaFET calibration instructions and recommendations see Bresnahan et al. (2014) and Rivest et al. (2016).
} 

\value{
This function returns a dataframe comprising 2 variables: 
  \item{E0int25}{Calibration coefficients of the internal sensor at 25oC.}
  \item{E0ext25}{Calibration coefficients of the external sensor at 25oC.}

}
\references{
Bresnahan, P. J. J., Martz, T. R., Takeshita, Y., Johnson, K. S., and LaShomb, M., 2014. Best practices for autonomous measurement of seawater pH with the Honeywell Durafet. \emph{Methods Oceanogr.} \bold{9}, 44-60.

Martz, T.R., Connery, J.G., and Johnson, K.S., 2010. Testing the Honeywell Durafet for seawater pH applications. \emph{Limnol. Oceanogr. Meth.} \bold{8}, 172-184.

Rivest, E.B., O'Brien, M., Kapsenberg, L., Gotschalk, C.C., Blanchette, C.A., Hoshijima, U., and Hofmann, G.E., 2016. Beyond the benchtop and the benthos: dataset management planning and design for time series of ocean carbonate chemistry associated with Durafet(c)-based pH sensors. \emph{Ecological Informatics} \bold{36}, 209-220.
}

\author{
Samir Alliouane, Lydia Kapsenberg, Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{sf_calc}}.
}

\examples{
sf_calib(calEint=0.0865, calEext=-0.93, calpH=8.132, calT=16.2, calSal=35.6)

## Using the test file seaFET
sf_calib(calEint=seaFET$Eint, calEext=seaFET$Eext, 
  calpH=seaFET$pHspectro, calT=seaFET$Temperature, 
  calSal=seaFET$Salinity)
}
\keyword{utilities}

\encoding{latin1}
\name{K2si}
\alias{K2si}
\title{Second dissociation constant of Si(OH)4}
\description{Second dissociation constant of Si(OH)4 (mol/kg)}
\usage{
K2si(S=35, T=25, P=0, pHscale="T", kSWS2scale="x", ktotal2SWS_P0="x")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{P}{Hydrostatic pressure in bar (surface = 0), default is 0}
  \item{pHscale}{choice of pH scale: "T" for using the total scale, "F" for using the free scale and "SWS" for using the seawater scale, default is total scale}
  \item{kSWS2scale}{Conversion factor from the seawater scale (SWS) to the pH scale selected at the hydrostatic pressure value indicated. It is advised to use default value "x", in which case it is computed when required.}
  \item{ktotal2SWS_P0}{Conversion factor from the total scale to the SWS at an hydrostatic pressure of 0. It is advised to use default value "x", in which case it is computed when required.}
}

\details{This equation is modified from Wischmeyer et al. (2003), who fitted the temperature-dependent K2si from Nordstrom et al. (1990) for freshwater to a value of 12.56 for T=25 and an ionic strength of 0.5 mol/kg. The temperature and salinity ranges in which it is valid are not well constrained.

The pressure correction is applied on the seawater scale. Hence, values are first transformed from the total scale to the seawater scale, the pressure correction applied as described by Millero (1995), and the value is transformed back to the required scale (T, F or SWS).

Note that the arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length.
} 

\value{
	\item{K2si}{Second dissociation constant of Si(OH)4 (mol/kg)}
}

\references{
Millero F. J., 1995 Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica Cosmochimica Acta} \bold{59}: 661-677.

Nordstrom D. K., L. N. Plummer, D. Langmuir, E. Busenberg, H. M. May, B. F. Jones, D. L. Parkhurst, 1990 Revised chemical equilibrium data from major mineral reactions and their limitations. In: Melchior, D.C., R. L. Bassett (Eds.) Chemical Modeling of Aqueous Systems. \emph{IIACS Series} \bold{416}. American Chemical Society, Washington, DC.

Wischmeyer A. G., Y. Del Amo, M. Brzezinski, D. A. Wolf-Gladrow, 1995 Theoretical constraints on the uptake of silicic acid species by marine diatoms. \emph{Marine Chemistry} \bold{82}: 13-29.
}
             
\author{
Mathilde Hagens (\email{M.Hagens@uu.nl})
}

%\seealso{}

\examples{
  K2si(S=35, T=25, P=0, pHscale="T")
}

\keyword{utilities}

\encoding{latin1}
\name{speciation}
\alias{speciation}
\title{ionic forms as a function of pH}
\description{Estimates the concentration of the various ionic forms of a molecule as a function of pH}
\usage{
speciation(K1=K1(), K2=NULL, K3=NULL, pH, conc=1)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{K1}{First dissociation constant}
  \item{K2}{Second dissociation constant, default is NULL}
  \item{K3}{Third dissociation constant, default is NULL}
  \item{pH}{pH value, default is 8}
  \item{conc}{concentration of molecule in mol/kg, default is 1 mol/kg}
}

\value{The function returns a data frame containing the following concentrations (in mol/kg if conc is given in mol/kg):
  \item{C1}{ionic form 1, univalent, bivalent and trivalent molecules}
  \item{C2}{ionic form 2, univalent, bivalent and trivalent molecules}
  \item{C3}{ionic form 3, bivalent and trivalent molecules}
  \item{C4}{ionic form 4, trivalent molecules}    
}


\references{
Zeebe R. E. and Wolf-Gladrow D. A., 2001 \emph{CO2 in seawater: equilibrium, kinetics, isotopes}. Amsterdam: Elsevier, 346 pp.
}
             
\author{
Karline Soetaert \email{K.Soetaert@nioo.knaw.nl}
}

\seealso{
	\code{\link[seacarb]{bjerrum}}.
}

\examples{
## Speciation of divalent species; example to estimate the various ionic forms
## of dissolved inorganic carbon  (DIC = 0.0021 mol/kg) at a salinity of 35,
## a temperature of 25oC and an hydrostatic pressure of 0:
spec <- speciation (K1(35, 25, 0), K2(35, 25, 0), pH=8, conc=0.0021)
## where (spec\$C1=[CO2], spec\$C2=[HCO3-], spec\$C3=[CO3--])

## Speciation of trivalent species (e.g.,  H3PO4, H2PO4-, HPO4--, PO4---)
speciation(K1p(), K2p(), K3p(), conc=0.001)

## Effect of temperature on pCO2 - Figure 1.4.18 of Zeebe and Wolf-Gladrow (2001)
Tseq <- seq(0, 30, by=0.5)
pHseq <- carb(flag=15, var1=2300e-6, var2=1900e-6, S=35, T=Tseq, P=0)$pH
CO2  <- speciation(K1(T=Tseq), K2(T=Tseq), conc=1900, pH=pHseq)$C1
pCO2 <- CO2/K0(T=Tseq)
plot(Tseq, pCO2, xlab="Temperature (oC)", ylab="pCO2 (uatm)", type="l", 
	main="effect of temperature on pCO2")
legend("topleft", c(expression(sum(CO[2])==1900~umol~kg^"-1"), 
	expression(TA==2300~umol~kg^"-1")))
}

%\details{}

\keyword{utilities}

\encoding{latin1}
\name{sp2sa_geo}
\alias{sp2sa_geo}

\title{From practical to absolute salinity}
\description{Converts from practical to absolute salinity based on depth and geographic location.}

\usage{
sp2sa_geo(SP, P=0, long=1.e20, lat=1.e20)
}
%- maybe also 'usage' for other objects documented here.

\arguments{
  \item{SP}{Practical salinity on the practical salinity scale}
  \item{P}{Sea water pressure in dbar}
  \item{long}{Longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]}
  \item{lat}{Latitude in decimal degrees [-90 ... 90]}
  }

\value{
  \item{SA}{Absolute salinity (g/kg)}
}

\details{
This function is almost an alias of function gsw_SA_from_SP of the gsw package on which it relies. The only difference is in that depth and location are optional.  If location is not given, or incomplete (either longitude or latitude missing), an arbitrary location is chosen: the mid equatorial atlantic ocean. Note that this implies an error on computed SA ranging from 0 up to 0.02 g/kg.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}

\seealso{
sa2sp_geo does the reverse, sp2sa_chem
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate the absolute salinity of a sample whose practical Salinity is 35,
   # depth is 10 dbar and location is 188 degrees East and 4 degrees North.
   SA <- sp2sa_geo(35, 10, 188, 4)     # 34.711778344814114
}

\keyword{utilities}

\encoding{latin1}
\name{eos2teos_chem}
\alias{eos2teos_chem}

\title{Convert temperature and salinity from EOS-80 to TEOS-10}
\description{Converts in situ temperature to conservative temperature and practical to absolute salinity (SA). Salinity conversion depends on total alkalinity as well as the concentrations of dissolved inorganic carbon, nitrate and silicate.}
\usage{
eos2teos_chem(SP, T, P=0, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{SP}{Practical salinity on the practical salinity scale}
  \item{T}{In situ temperature in deg. C}
  \item{P}{Sea water pressure in dbar}
  \item{TA}{Total alkalinity, in mol/kg, default is 2300 µmol/kg}
  \item{DIC}{Dissolved inorganic carbon concentration in mol/kg, default is 2000 µmol/kg}
  \item{NO3}{Total nitrate concentration in mol/kg, default is 0}
  \item{SIOH4}{Total silicate concentration in mol/kg, default is 0}
  }

\value{The function returns a data frame containing the following columns:
  \item{CT}{Conservative temperature (deg C)}
  \item{SA}{Absolute salinity (g/kg)}
}

\details{
Conversion from practical to absolute salinity depends on carbonate system parameters and ion concentration which mostly affect water density anomalies.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., Wright D. G. and Millero F. J., 2011. The effects of biogeochemical processes on oceanic conductivity/salinity/density relationships and the characterization of real seawater. \emph{Ocean Science} \bold{7}, 363-387.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
teos2eos_chem does the reverse, eos2teos_geo, sp2sa_chem

package gsw
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate Conservative Temperature and Absolute Salinity of a sample with 
   # Practical salinity of 35 psu, in-situ temperature of 18 deg C,
   # at 0 dbar and total alkalinity of 0.00234 mol/kg and DIC of 0.00202 mol/kg
   f <- eos2teos_chem(SP=35, T=18, P=0, TA=0.00234, DIC=0.00202)
   CT <- f$CT     # Conservative Temperature
   SA <- f$SA     # Absolute Salinity
}

\keyword{utilities}

\encoding{latin1}
\name{K1p}
\alias{K1p}
\title{First dissociation constant of phosphoric acid (mol/kg)}

\description{First dissociation constant of phosphoric acid (mol/kg)}

\usage{
K1p(S=35, T=25, P=0, pHscale="T", kSWS2scale="x", warn="y")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{P}{Hydrostatic pressure in bar (surface = 0), default is 0}
  \item{pHscale}{Choice of pH scale: "T" for using the total scale, "F" for using the free scale and "SWS" for using the seawater scale, default is total scale}
  \item{kSWS2scale}{Conversion factor from the seawater scale (SWS) to the pH scale selected at the hydrostatic pressure value indicated. It is advised to use default value "x", in which case it is computed when required.}
  \item{warn}{"y" to show warnings when T or S go beyond the valid range for K1p; "n" to supress warnings. The default is "y".}
}

\details{
This formulation is only valid for specific ranges of temperature and salinity: 

\itemize{
\item S ranging between 0 and 45 and T ranging between 0 and 45oC.
}

The pressure correction was applied on the seawater scale. Hence, values were first transformed from the total scale to the seawater scale, the pressure correction applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).

Note that the arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length.
} 

\value{
  \item{K1p}{First dissociation constant of phosphoric acid (mol/kg)}
 
}
\references{
Millero F. J., 1995 Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica Cosmochimica Acta} \bold{59}: 661-677.

Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.}

\author{
Jean-Marie Epitalon, Heloise Lavigne, Aurelien Proye and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{K2p}}, \code{\link{K3p}}.
}

\examples{
  K1p(35,25,0)
}
\keyword{utilities}

\encoding{latin1}
\name{K1}
\alias{K1}
\title{First dissociation constant of carbonic acid (mol/kg)}
\description{First dissociation constant of carbonic acid (mol/kg)}
\usage{
K1(S=35, T=25, P=0, k1k2="x", pHscale="T", kSWS2scale="x", ktotal2SWS_P0="x", warn="y")
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{P}{Hydrostatic pressure in bar (surface = 0), default is 0}
  \item{k1k2}{"cw" for using K1 and K2 from Cai & Wang (1998), "l" from Lueker et al. (2000), "m02" from Millero et al. (2002), "m06" from Millero et al. (2006), "m10" from Millero (2010), "mp2" from Mojica Prieto et al. (2002), "p18" from Papadimitriou et al. (2018), "r" from Roy et al. (1993), "sb21" from Shockman & Byrne (2021), "s20" from Sulpis et al. (2020), and "w14" from Waters et al. (2014). "x" is the default flag; the default value is then "l", except if T is outside the range 2 to 35oC and/or S is outside the range 19 to 43. In these cases, the default value is "w14".}
  \item{pHscale}{choice of pH scale: "T" for using the total scale, "F" for using the free scale and "SWS" for using the seawater scale, default is total scale}
  \item{kSWS2scale}{Conversion factor from the seawater scale (SWS) to the pH scale selected at the hydrostatic pressure value indicated. It is not required for all formulations of K1 and K2, nor when the hydrostatic pressure is 0. It is advised to use default value "x", in  which case it is computed when required.}
  \item{ktotal2SWS_P0}{Conversion factor from the total scale to the SWS at an hydrostatic pressure of 0. It is not required for all formulations of K1 and K2. It is advised to use default value "x", in  which case it is computed when required.}
  \item{warn}{"y" to show warnings when T or S go beyond the valid range for K1; "n" to supress warnings. The default is "y".}
  }

\details{The Lueker et al. (2000) constant is recommended by Guide to Best Practices for Ocean CO2 Measurements (2007). It is, however, critical to consider that each formulation is only valid for specific ranges of temperature and salinity:

\itemize{
\item Cai and Wang (1998): S ranging between 0 and 40 and T ranging between 0.2 and 30oC.

\item Lueker et al. (2000): S ranging between 19 and 43 and T ranging between 2 and 35oC.

\item Millero et al. (2002): S ranging from 34 to 37 and T ranging between -1.6 and 35oC.

\item Millero et al. (2006): S ranging between 0.1 and 50 and T ranging between 1 and 50oC.

\item Millero (2010): S ranging between 1 and 50 and T ranging between 0 and 50oC. Millero (2010) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale". 

\item Mojica Prieto et al. (2002): S ranging from 5 to 42 and T ranging between 0 and 45oC.

\item Papadimitriou et al. (2018): S ranging from 33 to 100 and T ranging between -6 to 25oC.

\item Roy et al. (1993): S ranging between 5 and 45 and T ranging between 0 and 45oC.

\item Shockman & Byrne (2021): formulation is that of Waters et al (2014). See below.

\item Sulpis et al. (2020): S ranging from 30.7 to 37.6 and T ranging between -1.7 to 31.8oC.

\item Waters et al.(2014): S ranging between 1 and 50 and T ranging between 0 and 50oC. Waters (2014) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale". 
}

The arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length. 

The pressure correction was applied on the seawater scale. Hence, if needed, values were first transformed from the total scale to the seawater scale, the pressure correction applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).}

\value{
  \item{K1}{First dissociation constant of carbonic acid (mol/kg)}
}

\references{
Cai W. J., and Wang Y., 1998. The chemistry, fluxes, and sources of carbon dioxide in the estuarine waters of the Satilla and Altamaha Rivers, Georgia. \emph{Limnology and Oceanography} \bold{43}, 657-668.

Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.

DOE 1994 \emph{Handbook of methods for the analysis of the various parameters of the carbon dioxide system in sea water}. ORNL/CDIAC-74. Oak Ridge,Tenn.: Carbon Dioxide Information Analysis Center, Oak Ridge National Laboratory.

Lueker T. J., Dickson A. G., and Keeling C. D., 2000 Ocean pCO2 calculated from dissolved inorganic carbon, alkalinity, and equations for K1 and K2: validation based on laboratory measurements of CO2 in gas and seawater at equilibrium. \emph{Marine Chemistry} \bold{70} 105-119.

Millero F. J., 1995 Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica Cosmochimica Acta} \bold{59}: 661-677.

Millero F. J., Pierrot D., Lee K., Wanninkhof R., Feely R., Sabine C. L., Key R. M. and Takahashi T., 2002. Dissociation constants for carbonic acid determined from field measurements. Deep Sea Research Part I: Oceanographic Research Papers \bold{49}:1705-1723.

Millero F. J., Graham T. B., Huang F., Bustos-Serrano H., and Pierrot D., 2006 Dissociation constants of carbonic acid in seawater as a function of salinity and temperature.  \emph{Marine Chemistry} \bold{100}, 80-84.

Millero F. J., 2010 Carbonate constant for estuarine waters. \emph{Marine and Freshwater Research} \bold{61}: 139-142.

Mojica Prieto F. J. and Millero F. J., 2002. The values of pK1 + pK2 for the dissociation of carbonic acid in seawater. \emph{Geochimica et Cosmochimica Acta} \bold{66}, 2529-2540.

Papadimitriou S., Loucaides S., Rérolle V. M. C., Kennedy P., Achterberg E. P., Dickson A. G., Mowlem M. and Kennedy H., 2018. The stoichiometric dissociation constants of carbonic acid in seawater brines from 298 to 267 K. \emph{Geochimica et Cosmochimica Acta} \bold{220}, 55-70.

Roy R. N., Roy L. N., Vogel K. M., Porter-Moore C., Pearson T., Good C. E., Millero F. J. and Campbell D. M., 1993. The dissociation constants of carbonic acid in seawater at salinities 5 to 45 and temperatures 0 to 45oC. \emph{Marine Chemistry} \bold{44}, 249-267.

Schockman, K.M., Byrne, R.H., 2021. Spectrophotometric determination of the bicarbonate dissociation constant in seawater, \emph{Geochimica et Cosmochimica Acta}..

Sulpis O., Lauvset S. K. and Hagens M., 2020. Current estimates of K1* and K2* appear inconsistent with measured CO2 system parameters in cold oceanic regions. \emph{Ocean Science} \bold{16}, 847-862.

Waters, J., Millero, F. J., and Woosley, R. J., 2014. Corrigendum to ``The free proton concentration scale for seawater pH'', [MARCHE: 149 (2013) 8-22], \emph{Marine Chemistry} \bold{165}, 66-67.
}
             
\author{
Jean-Marie Epitalon and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{K2}}.
}

\examples{
  K1(S=35,T=25,P=0,k1k2="l",pHscale="T")
}

\keyword{utilities}

\encoding{latin1}
\name{fCO2insi}
\alias{fCO2insi}
\title{fCO2 at in situ temperature}
\description{Correction to compensate for the difference in temperature between the temperature of measurement and in situ temperature.}
\usage{
fCO2insi(fCO2lab = 400, Tlab = 20, SST = 19)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{fCO2lab}{Fugacity of CO2 measured in the lab in \eqn{\mu}atm}
  \item{Tlab}{Temperaure of measurment in the lab in in degrees Celsius, default is 20oC}
  \item{SST}{Temperature in degrees Celsius, default is 19oC}
  }

\value{
  \item{fCO2insi}{Fugacity of CO2 at in situ temperature in \eqn{\mu}atm, the same units as that for the fCO2 input.}
}

\note{The empirical correction applied is from Takahashi (1993) as recommended by Pierrot et al. (2009)}

\references{
Pierrot D., Neill C., Sullivan K., Castle R., Wanninkhof R., Lüger H., Johannessen T., Olsen A., Feely R. A. and Cosca C. E., 2009. Recommendations for autonomous underway pCO2 measuring systems and data-reduction routines. \emph{Deep-Sea Res. II} \bold{56}, 512-522.

Takahashi T., Olafsson J., Goddard J. G., Chipman D. W. and Sutherland S. C., 1993. Seasonal variation of CO2 and nutrients in the high-latitude surface oceans: a comparative study. \emph{Glob. Biogeochem. Cycles} \bold{7}, 843-878.
}
             
\author{
Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{f2pCO2}}, \code{\link{pCO2insi}}.
	}

\examples{
fCO2insi(fCO2lab = 400, SST = 15, Tlab = 16)
}

\keyword{utilities}

\encoding{latin1}
\name{sa2sp_geo}
\alias{sa2sp_geo}

\title{From absolute to practical salinity}
\description{Converts from absolute to practical salinity based on depth and geographic location.}
\usage{
sa2sp_geo(SA, P=0, long=1.e20, lat=1.e20)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{SA}{Absolute salinity in g/kg}
  \item{P}{Sea water pressure in dbar}
  \item{long}{Longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]}
  \item{lat}{Latitude in decimal degrees [-90 ... 90]}
  }

\value{
  \item{SP}{Practical salinity (psu)}
}

\details{
This function is almost an alias of subroutine gsw_SP_from_SA from gsw package on which it relies. The only difference is in that depth and location are optional. If location is not given, or incomplete (either longitude or latitude missing), an arbitrary location is chosen: the mid equatorial atlantic ocean. Note that this implies an error on computed SA ranging from 0 up to 0.02 g/kg.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
sp2sa_geo does the reverse, sa2sp_chem
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate the practical salinity of a sample whose absolute Salinity is 35,
   # depth is 10 dbar and location is 188 degrees East and 4 degrees North.
   SP <- sa2sp_geo(35, 10, 188, 4)
}

\keyword{utilities}

\encoding{latin1}
\name{teos2eos_geo}
\alias{teos2eos_geo}

\title{Convert temperature and salinity from TEOS-10 to EOS-80}
\description{Converts conservative temperature to in situ temperature and absolute salinity to practical salinity (SP). Salinity conversion depends on depth and geographic location.
}
\usage{
teos2eos_geo(SA, CT, P=0, long=1.e20, lat=1.e20)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{SA}{Absolute salinity in g/kg}
  \item{CT}{Conservative temperature in degrees C}
  \item{P}{Sea water pressure in dbar}
  \item{long}{Longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]}
  \item{lat}{Latitude in decimal degrees [-90 ... 90]}
  }

\value{The function returns a data frame containing the following columns:
  \item{T}{In situ temperature (deg C)}
  \item{SP}{Practical salinity (psu)}
}

\details{
Conversion from absolute to practical salinity depends on water density anomaly which is correlated with silicate concentration. This function relies on silicate concentration taken from WOA (World Ocean Atlas) to evaluate the density anomaly.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., Wright D. G. and Millero F. J., 2011. The effects of biogeochemical processes on oceanic conductivity/salinity/density relationships and the characterization of real seawater. \emph{Ocean Science} \bold{7}, 363-387.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}

\seealso{
eos2teos_geo does the reverse, teos2eos_chem, sa2sp_geo, package gsw
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate in situ temperature and practical salinity of a sample with 
   # Absolute salinity of 35 g/kg, conservative temperature of 18 deg C,
   # depth is 10 dbar and location is 188 degrees East and 4 degrees North.
   f <- teos2eos_geo(SA=35, CT=18, P=10, long=188, lat=4)
   T <- f$T     # in situ temperature
   SP <- f$SP     # Practical salinity
}

\keyword{utilities}

\name{seacarb_test_P300}
\docType{data}
\alias{seacarb_test_P300}
\title{Test data file (at P=300) to test the use of the carb function}
\description{
The variables are:

\itemize{
	\item Flag indicating which couple of variables is used
	\item Value of the first variable in mol/kg, except for pH and for pCO2 in \eqn{\mu}atm
	\item Value of the second variable in mol/kg, except for pH
	\item Salinity
	\item Temperature in degrees Celsius
	\item Hydrostatic pressure in bar (P=300)
	\item Value of the concentration of total phosphate in mol/kg
	\item Value of the total silicate in mol/kg
}
}
\usage{seacarb_test_P300}
\format{A data frame with 20 rows and 8 variables}
\source{None, these data were invented for this purpose. The input variables were chosen in order to check that the carbonate chemistry is identical for all flags.}
\keyword{datasets}

\encoding{latin1}
\name{f2pCO2}
\alias{f2pCO2}
\title{Converts the CO2 fugacity to CO2 partial pressure}
\description{Converts fCO2 (fugacity of CO2) into pCO2 (partial pressure in CO2)}
\usage{
f2pCO2(T = 25, Patm=1, P=0, fCO2)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{Patm}{Surface atmospheric pressure in atm, default 1 atm}
  \item{P}{Hydrostatic pressure in bar, default is 0 bar (surface)}
  \item{fCO2}{Fugacity of CO2 in \eqn{\mu}atm, the same units as that for the pCO2 output}
  }

\value{
  \item{pCO2}{Partial pressure of CO2 in \eqn{\mu}atm, the same units as that for the fCO2 input}
}

\note{\bold{Warning:} pCO2 estimates below 100 m are subject to considerable uncertainty. See Weiss (1974) and Orr et al. (2015)}

\references{
Dickson A. G., Sabine C. L. and Christian J. R., 2007. Guide to best
practices for ocean CO2 measurements. \emph{PICES Special Publication}
\bold{3}, 1-191.

Orr J. C., Epitalon J.-M. and Gattuso J.-P., 2015. Comparison of seven packages that compute ocean carbonate chemistry. \emph{Biogeosciences} \bold{12}, 1483-1510.

Weiss, R. F. and Price, B. A., 1980. Nitrous oxide solubility in water and seawater, \emph{Marine Chemistry}, \bold{8}, 347-359, 1980.

Weiss, R. F., 1974. Carbon dioxide in water and seawater: the solubility of a non-ideal gas, \emph{Mar.  Chem.}, \bold{2}, 203-215.

Weiss, R. F. and Price, B. A., 1980. Nitrous oxide solubility in water and seawater, \emph{Marine Chemistry}, \bold{8}, 347-359.
}
             
\author{
Heloise Lavigne, Jean-Pierre Gattuso, and James Orr \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{p2fCO2}}.
}

\examples{
f2pCO2(T=25, Patm=1.0, P=0, fCO2=380)  
}

\keyword{utilities}

\encoding{latin1}
\name{sa2sp_chem}
\alias{sa2sp_chem}

\title{From absolute to practical salinity}
\description{Converts from absolute to practical salinity (SP). Salinity conversion depends on total alkalinity as well as the concentrations of dissolved inorganic carbon, nitrate and silicate.}
\usage{
sa2sp_chem(SA, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{SA}{Absolute salinity in g/kg}
  \item{TA}{Total alkalinity, in mol/kg, default is 2300 µmol/kg}
  \item{DIC}{Dissolved inorganic carbon concentration in mol/kg, default is 2000 µmol/kg}
  \item{NO3}{Total nitrate concentration in mol/kg, default is 0}
  \item{SIOH4}{Total silicate concentration in mol/kg, default is 0}
  }

\value{
  \item{SP}{Practical salinity (in psu)}
}

\details{
Convert from absolute to practical salinity from carbon system parameters and ion concentration which most affect water density anomalies.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., Wright D. G. and Millero F. J., 2011. The effects of biogeochemical processes on oceanic conductivity/salinity/density relationships and the characterization of real seawater. \emph{Ocean Science} \bold{7}, 363-387.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
sp2sa_chem does the reverse, sa2sp_geo
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate the practical salinity of a sample with Absolute Salinity of 35 g/kg,
   # Total Alkalinity of 0.00234 mol/kg and DIC of 0.00202 mol/kg
   SP <- sa2sp_chem(SA=35, TA=0.00234, DIC=0.00202)
}

\keyword{utilities}

\name{alkalinity}
\docType{data}
\alias{alkalinity}
\title{Example data file for function at}
\description{
The variables are:

\itemize{
	\item volume: Volume of acid added to the sample in ml	
	\item E: Potential measured during the titration in mV	
	\item temperature: Temperature in degrees Celsius	
	\item weight: Weight of the sample in g
	\item S: Salinity
	\item normality : Normality of the acid
	\item ETris: Potential used for the calibration of the electrode in mV
	\item pHTris: pH used for the calibration of the electrode with the TRIS buffer
	}
}
\usage{alkalinity}

\format{A data frame with 29 rows and 8 variables}

\source{
Data come from a potentiometric titration performed by Steeve Comeau.
}

\keyword{datasets}

\encoding{latin1}
\name{pHnbs2sws}
\alias{pHnbs2sws}
\title{Converts pH from NBS to SWS scale}
\description{Converts pH on NBS scale to pH on SWS scale pCO2. 
The NBS-to-SWS conversion is done with the total activity coefficient fH (combined activity coeff for H+, HSO4-, and HF)
from Takahashi et al. (1982) based on data from Culberson and Pytkowicz (1973). The approach is old and full of uncertainty.
Newer approaches are more complicated (Pitzer equations) and big uncertainties remain (Marion et al., 2011; Pilson, 2013).
}
\usage{
pHnbs2sws(pHNBS, S=35, T=25)
}
\arguments{
  \item{pHNBS}{pH on NBS scale}
  \item{S}{Salinity on the practical salinity scale, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  }

\value{
  \item{pHSWS}{pH on SWS scale}
}

\details{
The pHSWS is computed from pHNBS using the total activity coefficient and 
relying on the following basic chemistry equation: \eqn{ ah = fH * hsws}, 
where  \eqn{ah}  is the activity of hydrogen ion,
\eqn{fH}  is the total activity coefficient, and
\eqn{hsws = [H+] + [HSO4-] + [HF]}. In other words, 
\eqn{hsws} is the hydrogen ion concentration on the seawater scale.

The two pH scales are defined as \eqn{pHNBS = -log10(ah)} and 
\eqn{pHSWS = -log10(hsws)}.  
}

\references{
Culberson, C.H., & Pytkowicz, R.M. (1973). Ionization of water in seawater. 
\emph{Marine Chemistry}, \bold{1(4)}, 309-316.

Marion G.M., Millero F.J., Camoes M.F., Spitzer P., Feistel R., Chen C.T.A. 2011. pH of seawater. 
\emph{Marine Chemistry} \bold{126} 89-96.

Pilson M.E.Q. (2013) An introduction to the chemistry of the sea, 
2 edn. Cambridge, UK: Cambridge University Press.

Takahashi T., Williams R.T., and Ros D.L. (1982) Carbonate chemistry. 
GEOSECS Pacific Expedition, Volume 3, Hydrographic Data 1973-1974, 77-83.

}
             
\author{
James Orr \email{james.orr@lsce.ipsl.fr}
}

\seealso{
	\code{\link{fH}} and \code{\link{pHsws2nbs}}
}

\examples{
## Convert pHNBS to pHSWS
   pHNBS = 8.0                          # pH on the NBS scale
   pHSWS = pHnbs2sws(pHNBS, T=25, S=35) # pH on the SWS scale
   print(pHSWS)
## Check value: the result should be 7.853336
}

\keyword{utilities}

\encoding{latin1}
\name{eos2teos_geo}
\alias{eos2teos_geo}

\title{Convert temperature and salinity from EOS-80 to TEOS-10}
\description{Converts in situ to conservative temperature and practical to absolute salinity (SA). Salinity conversion depends on depth and geographic location. 
}
\usage{
eos2teos_geo(SP, T, P=0, long=1.e20, lat=1.e20)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{SP}{Practical salinity on the practical salinity scale}
  \item{T}{In situ temperature in deg. C}
  \item{P}{Sea water pressure in dbar}
  \item{long}{Longitude in decimal degrees [ 0 ... +360 ] or  [ -180 ... +180 ]}
  \item{lat}{Latitude in decimal degrees [-90 ... 90]}
  }

\value{The function returns a data frame containing the following columns:
  \item{CT}{Conservative temperature (deg C)}
  \item{SA}{Absolute salinity (g/kg)}
}

\details{
Conversion from practical to absolute salinity depends on water density anomaly which is correlated with silicate concentration. This function relies on silicate concentration taken from WOA (World Ocean Atlas) to evaluate density anomaly.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
teos2eos_geo does the reverse, eos2teos_chem, sp2sa_geo, package gsw
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate conservative temperature and absolute salinity of a sample with 
   # Practical salinity of 35 psu, in situ temperature of 18 deg C,
   # depth is 10 dbar and location is 188 degrees East and 4 degrees North.
   f <- eos2teos_geo(SP=35, T=18, P=10, long=188, lat=4)
   CT <- f$CT     # Conservative temperature
   SA <- f$SA     # Absolute salinity
}

\keyword{utilities}

\encoding{latin1}
\name{p2xCO2}
\alias{p2xCO2}
\title{Converts partial pressure of CO2 to mole fraction of CO2}
\description{Converts pCO2 (partial pressure of CO2) into xCO2 (mole fraction of CO2)}
\usage{
p2xCO2(S=35, T=25, Patm=1, pCO2)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity on the practical salinity scale, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{Patm}{Atmospheric pressure in atmospheres, default is 1.0 (this is not the hydrostatic pressure)}
  \item{pCO2}{Partial pressure of CO2 in \eqn{\mu}atm}
  }

\value{
  \item{xCO2}{Mole fraction of CO2 in ppm.}
}

\details{
The xCO2 (ppm) is computed from pCO2 (\eqn{\mu}atm) using the
following equation: \eqn{ xCO2 = pCO2 / (Patm - pH2O)}, where pH20 is the vapor pressure of seawater computed following best practices (Dickson et al., 2007). That computed pH20 is identical, when rounded to the 4th decimal place, with that computed by the equation from Weiss and Price (1980).}

\note{\bold{Warning:} pCO2 estimates below 100 m are subject to considerable uncertainty. See Weiss (1974) and Orr et al. (2015)}

\references{
Dickson A. G., Sabine C. L. and Christian J. R. (2007) Guide to best
practices for ocean CO2 measurements. \emph{PICES Special Publication}
\bold{3}, 1-191.

Orr J. C., Epitalon J.-M. and Gattuso J.-P., 2015. Comparison of seven packages that compute ocean carbonate chemistry. \emph{Biogeosciences} \bold{12}, 1483-1510.

Weiss, R. F. (1974) Carbon dioxide in water and seawater: the
solubility of a non-ideal gas, Marine Chemistry, \bold{2}, 203-215.

Weiss, R. F. and Price, B. A. (1980) Nitrous oxide solubility in water
and seawater, Marine Chemistry, \bold{8}, 347-359.
}
             
\author{
James Orr \email{james.orr@lsce.ipsl.fr}
}

\seealso{
	\code{\link{x2pCO2}} and \code{\link{vapress}}
}

\examples{
## Convert atmospheric pressure from mbar to atm
   Patm_mbar = 1052                 # in millibar
   Patm      = Patm_mbar / 1013.25  # in atm
## Compute xCO2 from pCO2
   pCO2 = 380
   xCO2 = p2xCO2(T=25, S=35, Patm=Patm, pCO2=pCO2)
   print(xCO2)
## The result is 377.1546 ppm 
}

\keyword{utilities}

\encoding{latin1}
\name{d2p}
\alias{d2p}
\title{Converts depth in meters to pressure in dbar}
\description{Converts depth in meters to pressure in dbar}
\usage{
d2p(depth, lat=40)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{depth}{Depth in meters}
  \item{lat}{Latitude in degrees, N and S is irrelevant, default is 40o}
  }

\value{
  \item{pressure}{Pressure corresponding to the depth given, in dbar}
}

\references{
Saunders P. M., 1981. Practical conversion of pressure to depth. \emph{J. Phys. Oceanogr.} \bold{11}: 573-574.
}
             
\author{
Heloise Lavigne and Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{p2d}}
}

\examples{
d2p(depth=7500, lat=30)  
}

\keyword{utilities}

\encoding{latin1}
\name{pHslope}
\alias{pHslope}

\title{Slope of the calibration curve of a pH electrode}
\description{Slope of the calibration curve of a pH electrode (percent of theoretical slope)}
\usage{
pHslope(Etris=-72.4,Eamp=4.9,S=35,T=25)
%- maybe also 'usage' for other objects documented here.
}

\arguments{
  \item{Etris}{e.m.f. of the TRIS buffer in mV, default is -72.4}
  \item{Eamp}{e.m.f. of the AMP buffer in mV, default is 4.9}
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
}

\details{Note that the arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length.} 

\value{
  \item{pHslope}{Slope of the calibration curve (in percent of theoretical slope)}

}
\references{
Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.
}

\author{
Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{tris}}, \code{\link{amp}}, \code{\link{pH}}.
}

\examples{
	##Example from Dickson et al. (2007)
	pHslope(Etris=-72.4,Eamp=4.9,S=35,T=25)
}
\keyword{utilities}

\encoding{latin1}
\name{bjerrum}
\alias{bjerrum}
\title{Bjerrum plot}
\description{Plot the concentration of the various ionic forms of a molecule as a function of pH}
\usage{
bjerrum(K1=K1(), K2=NULL, K3=NULL, phmin=2, phmax=12, by=0.1, conc=1, 
     type="l", col="black",  ylab="Relative concentration (\%)", add=FALSE, ...)
     }
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{K1}{First dissociation constant}
  \item{K2}{Second dissociation constant, default is NULL}
  \item{K3}{Third dissociation constant, default is NULL}
  \item{phmin}{Minimum pH value, default is 2}
  \item{phmax}{Maximum pH value, default is 12}
  \item{by}{Increment on the pH axis, default is 0.1}
  \item{conc}{concentration of molecule, default is 1}
  \item{type}{Type of plot, default is line}
  \item{col}{Color of plot, default is black}
  \item{ylab}{Label of Y axis, default is (mol/kg)}
  \item{add}{false:start new, true: add to current, default is false}
  \item{\dots}{Graphical parameters (see \code{\link{par}}) and any further arguments of plot, typically \code{\link{plot.default}}, may also be supplied as arguments to this function. Hence, the high-level graphics control arguments described under \code{\link{par}} and the arguments to \code{\link{title}} may be supplied to this function.}
}

%\value{XXX}

\references{
Zeebe, R. E. and Wolf-Gladrow D. A., 2001 \emph{CO2 in seawater: equilibrium, kinetics, isotopes}. Amsterdam: Elsevier, 346 pp.
}
             
\author{
Karline Soetaert \email{K.Soetaert@nioo.knaw.nl}
}

\seealso{
	\code{\link{matplot}}, \code{\link{par}}, \code{\link[seacarb]{speciation}}.
}

\details{
Note that the concentration is plotted in mol/kg only if conc is given is mol/kg
}

\examples{
## Plot the bjerrum plot for the carbonate system using the default values
bjerrum(K1(),K2(),main="DIC speciation",lwd=2) 
abline(v=-log10(K1()),col="grey")
mtext(side=3,at=-log10(K1()),"pK1")
abline(v=-log10(K2()),col="grey")
mtext(side=3,at=-log10(K2()),"pK2")
legend("left",lty=1:3,lwd=2,legend=c(expression(CO[2]),expression(HCO[3]^"-"),
	expression(CO[3]^"2-")))

## Plot the bjerrum plot for phosphate using the default values
bjerrum(K1p(),K2p(),K3p(),main="phosphate speciation",lwd=2)
legend("left",lty=1:4,lwd=2,legend=c(expression(H[3]~PO[4]),
	expression(H[2]~PO[4]^"-"),
expression(HPO[4]^"2-"),expression(PO[4]^"3-")))

## Plot the bjerrum plot for the carbonate system using the values other 
##	than the default ones, showing the effect of temperature
bjerrum(K1(T=25,S=35),K2(T=25,S=35),conc=1.3,main="effect of temperature" )
bjerrum(K1(T=0,S=35),K2(T=0,S=35),conc=1.3,add=TRUE,col="red")
legend("left",lty=1,col=c("black","red"),legend=c("T=25 oC","T=0 oC"))
legend("right",lty=1:3,legend=c(expression(CO[2]),expression(HCO[3]^"-"),
	expression(CO[3]^"2-")))

## Plot the bjerrum plot for the carbonate system using the values other 
##	than the default ones, showing the effect of salinity
bjerrum(K1(T=25,S=35),K2(T=25,S=35),conc=1.3,main="effect of salinity" )
bjerrum(K1(T=25,S=5),K2(T=25,S=5),conc=1.3,add=TRUE,col="blue")
legend("left",lty=1,col=c("black","blue"),legend=c("S=35","S=5"))
legend("right",lty=1:3,legend=c(expression(CO[2]),expression(HCO[3]^"-"),
	expression(CO[3]^"2-")))

## Plot the bjerrum plot for the carbonate system using the values other 
##	than the default ones, showing the effect of pressure
bjerrum(K1(P=0),K2(P=0),conc=1.3,main="effect of pressure" )
bjerrum(K1(P=300),K2(P=300),conc=1.3,add=TRUE,col="green")
legend("left",lty=1,col=c("black","green"),legend=c("P=0","P=300"),title="atm")
legend("right",lty=1:3,legend=c(expression(CO[2]),expression(HCO[3]^"-"),
	expression(CO[3]^"2-")))
}

\keyword{utilities}

\encoding{latin1}
\name{ppH}
\alias{ppH}
%- Also NEED an '\alias' for EACH other topic documented here.
\title{ppH}
\description{Calculates the carbonate chemistry after pH manipulations through addition of acid or base}
\usage{
ppH(flag, sys, var1, var2, pCO2a, vol, N, S=35, T=20, P=0, Pt=0, 
	Sit=0, pHscale="T", k1k2="x", kf="x", ks="d", eos = "eos80", 
	long = 1e+20, lat = 1e+20)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
	\item{flag}{Select the couple of variables available. The flags which can be used are: 

flag = 1      pH and CO2 given

flag = 2      CO2 and HCO3 given

flag = 3      CO2 and CO3 given

flag = 4      CO2 and ALK given

flag = 5      CO2 and DIC given

flag = 6      pH and HCO3 given

flag = 7      pH and CO3 given

flag = 8      pH and ALK given

flag = 9      pH and DIC given

flag = 10     HCO3 and CO3 given

flag = 11     HCO3 and ALK given

flag = 12     HCO3 and DIC given

flag = 13     CO3 and ALK given

flag = 14     CO3 and DIC given

flag = 15     ALK and DIC given

flag = 21     pCO2 and pH given

flag = 22     pCO2 and HCO3 given

flag = 23     pCO2 and CO3 given

flag = 24     pCO2 and ALK given

flag = 25     pCO2 and DIC given
}
	\item{sys}{0 if the manipulation is carried out in a system closed to the atmosphere or 1 if its is carried out in a system open to the atmosphere}
	\item{var1}{Value of the first variable in mol/kg, except for pH and for pCO2 in \eqn{\mu}atm}
	\item{var2}{Value of the second variable in mol/kg, except for pH}
	\item{pCO2a}{CO2 partial pressure in the atmosphere pCO2 in \eqn{\mu}atm. It is only used in systems open to the atmosphere (i.e. when sys=1)}
	\item{vol}{Volume of acid or base added in liter. By convention, it is given a negative sign for acid additions and a positive sign for base additions. The acid must be fortified with NaCl in order to have the same salinity than seawater.}
	\item{N}{Normality of the acid or base in mol/kg}
	\item{S}{Salinity}
	\item{T}{Temperature in degrees Celsius}
	\item{P}{Hydrostatic pressure in bar (surface = 0)}
	\item{Pt}{Concentration of total phosphate in mol/kg}
	\item{Sit}{Concentration of total silicate in mol/kg}
        \item{k1k2}{"cw" for using K1 and K2 from Cai & Wang (1998), "l" from Lueker et al. (2000), "m02" from Millero et al. (2002), "m06" from Millero et al. (2006), "m10" from Millero (2010), "mp2" from Mojica Prieto et al. (2002), "p18" from Papadimitriou et al. (2018), "r" from Roy et al. (1993), "sb21" from Shockman & Byrne (2021), "s20" from Sulpis et al. (2020), and "w14" from Waters et al. (2014). "x" is the default flag; the default value is then "l", except if T is outside the range 2 to 35oC and/or S is outside the range 19 to 43. In these cases, the default value is "w14".}
        \item{kf}{"pf" for using Kf from Perez and Fraga (1987) and "dg" for using Kf from Dickson and Riley (1979 in Dickson and Goyet, 1994). "x" is the default flag; the default value is then "pf", except if T is outside the range 9 to 33oC and/or S is outside the range 10 to 40. In these cases, the default is "dg".}
	\item{ks}{"d" for using Ks from Dickon (1990), "k" for using Ks from Khoo et al. (1977), default is "d"} 
	\item{pHscale}{choice of pH scale: "T" for using the total scale, "F" for using the free scale and "SWS" for using the seawater scale, default is total scale}
	 \item{eos}{"teos10" to specify T and S according to Thermodynamic Equation Of Seawater - 2010 (TEOS-10); "eos80" to specify T and S according to EOS-80.}
  \item{long}{longitude of data point, used when eos parameter is "teos10" as a conversion parameter from absolute to practical salinity.}
  \item{lat}{latitude of data point, used when eos parameter is "teos10".}
}

\details{The Lueker et al. (2000) constants for K1 and K2, the Perez and Fraga (1987) constant for Kf and the Dickson (1990) constant for Ks are recommended by Dickson et al. (2007). It is, however, critical to consider that each formulation is only valid for specific ranges of temperature and salinity:

\emph{For K1 and K2:}
\itemize{
\item Cai and Wang (1998): S ranging between 0 and 40 and T ranging between 0.2 and 30oC.

\item Lueker et al. (2000): S ranging between 19 and 43 and T ranging between 2 and 35oC.

\item Millero et al. (2002): S ranging from 34 to 37 and T ranging between -1.6 and 35oC.

\item Millero et al. (2006): S ranging between 0.1 and 50 and T ranging between 1 and 50oC.

\item Millero (2010): S ranging between 1 and 50 and T ranging between 0 and 50oC. Millero (2010) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale". 

\item Mojica Prieto et al. (2002): S ranging from 5 to 42 and T ranging between 0 and 45oC.

\item Papadimitriou et al. (2018): S ranging from 33 to 100 and T ranging between -6 to 25oC.

\item Roy et al. (1993): S ranging between 5 and 45 and T ranging between 0 and 45oC.

\item Shockman & Byrne (2021): for K2, S ranging from 19.6 to 41 and T ranging between 15 to 35oC. For K1, formulation is that of Waters et al.

\item Sulpis et al. (2020): S ranging from 30.7 to 37.6 and T ranging between -1.7 to 31.8oC.

\item Waters et al.(2014): S ranging between 1 and 50 and T ranging between 0 and 50oC. Waters (2014) provides a K1 and K2 formulation for the seawater, total and free pH scales. Therefore, when this method is used and if P=0, K1 and K2 are computed with the formulation corresponding to the pH scale given in the flag "pHscale".
}

\emph{For Kf:}
\itemize{
\item Perez and Fraga (1987): S ranging between 10 and 40 and T ranging between 9 and 33oC.

\item Dickson and Riley (1979 in Dickson and Goyet, 1994): S ranging between 0 and 45 and T ranging between 0 and 45oC.
}

\emph{For Ks:}
\itemize{
\item Dickson (1990): S ranging between 5 and 45 and T ranging between 0 and 45oC.

\item Khoo et al. (1977): S ranging between 20 and 45 and T ranging between 5 and 40oC.
}

The arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It is recommended to use either vectors with the same dimension or one vector for one argument and numbers for the other arguments.

\emph{Pressure corrections and pH scale:}

\itemize{
\item For K0, the pressure correction term of Weiss (1974) is used.

\item For K1, K2, pK1, pK2, pK3, Kw, Kb, Khs and Ksi, the pressure correction was applied on the seawater scale. Hence, if needed, values were first transformed from the total scale to the seawater scale, the pressure correction applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).

\item For Kf, the pressure correction was applied on the free scale. The formulation of Dickson and Riley (1979 in Dickson and Goyet, 1994) provides Kf on the free scale but that of Perez and Fraga (1987) provides it on the total scale. Hence, in that case, Kf was first transformed from the total scale to the free scale. With both formulations, the pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS). 

\item For Ks, the pressure correction was applied on the free scale. The pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS). 

\item For Kn, The pressure correction was applied on the seawater scale. The pressure correction was applied as described by Millero (1995), and the value was transformed back to the required scale (T, F or SWS).
}
long and lat are used as conversion parameters from absolute to practical salinity: when seawater is not of standard composition, practical salinity alone is not sufficient to compute absolute salinity and vice-versa. One needs to know the density. When long and lat are given, density is inferred from WOA silicate concentration at given location. When they are not, an arbitrary geographic point is chosen: mid equatorial Atlantic. Note that this implies an error on computed salinity up to 0.02 g/kg.
}

\value{The function returns a data frame containing the following columns:

  \item{comment}{The initial or final state water}
  \item{S}{Salinity}
  \item{T}{Temperature in degrees Celsius}
  \item{P}{Pressure in bar}
  \item{pH}{pH}
  \item{CO2}{CO2 concentration (mol/kg)}
  \item{pCO2}{pCO2, CO2 partial pressure (\eqn{\mu}atm)}
  \item{fCO2}{fCO2, CO2 fugacity (\eqn{\mu}atm)}
  \item{HCO3}{HCO3 concentration (mol/kg)}
  \item{CO3}{CO3 concentration (mol/kg)}
  \item{DIC}{DIC concentration (mol/kg)}
  \item{ALK}{ALK, total alkalinity (mol/kg)}
  \item{OmegaAragonite}{Omega aragonite, aragonite saturation state}
  \item{OmegaCalcite}{Omega calcite, calcite saturation state}
}

\note{\bold{Warning:} pCO2 estimates below 100 m are subject to considerable uncertainty. See Weiss (1974) and Orr et al. (2015)}

\references{
Cai W. J., and Wang Y., 1998. The chemistry, fluxes, and sources of carbon dioxide in the estuarine waters of the Satilla and Altamaha Rivers, Georgia. \emph{Limnology and Oceanography} \bold{43}, 657-668.

Dickson A. G., 1990 Standard potential of the reaction: AgCI(s) + 1/2H2(g) = Ag(s) + HCI(aq), and the standard acidity constant of the ion HSO4 in synthetic sea water from 273.15 to 318.15 K. \emph{Journal of Chemical Thermodynamics} \bold{22}, 113-127.

Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.

Khoo H. K., Ramette R. W., Culberson C. H. and Bates R. G., 1977 Determination of Hydrogen Ion Concentration in Seawater from 5 to 40oC: Standard Potentials at Salinities from 20 to 45. \emph{Analytical Chemistry} \bold{49}, 29-34. 

Gattuso J.-P. and Lavigne H., 2009 Perturbation experiments to investigate the impact of ocean acidification: approaches and software tools. \emph{Biogeosciences} \bold{6}, 4413-4439.

Millero F. J., 2010 Carbonate constant for estuarine waters. \emph{Marine and Freshwater Research} \bold{61}: 139-142.

Millero F. J., Graham T. B., Huang F., Bustos-Serrano H. and Pierrot D., 2006 Dissociation constants of carbonic acid in seawater as a function of salinity and temperature.  \emph{Marine Chemistry} \bold{100}, 80-84.

Orr J. C., Epitalon J.-M. and Gattuso J.-P., 2015. Comparison of seven packages that compute ocean carbonate chemistry. \emph{Biogeosciences} \bold{12}, 1483-1510.

Waters, J., Millero, F. J., and Woosley, R. J., 2014. Corrigendum to ``The free proton concentration scale for seawater pH'', [MARCHE: 149 (2013) 8-22], Marine Chemistry 165, 66-67.
}

\author{
Jean-Pierre Gattuso and Heloise Lavigne\email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link[seacarb]{buffer}}.
}

\examples{
ppH(flag=24, sys=0, var1=384, var2=2302e-6, pCO2a=384, vol=-12e-3, 
	N=0.01, S=34.3, T=16, P=0, pHscale="T", kf="pf", k1k2="l", ks="d")

ppH(flag=24, sys=1, var1=384, var2=2302e-6, pCO2a=384, vol=-12e-3, 
	N=0.01, S=34.3, T=16, P=0, pHscale="T", kf="pf", k1k2="l", ks="d")
}

\keyword{utilities}

\encoding{latin1}
\name{pHspec}
\alias{pHspec}

\title{Calculates pHT from results of spectrophotometric measurements}
\description{Calculates pH of a water sample from absorbance ratios R, obtained from spectrophotometric measurements with pH indicator dyes (on the total scale in mol/kg-soln)}
\usage{
pHspec(S=35, T=25, R=1, d="mCP", k="m18", warn="y")
%- maybe also 'usage' for other objects documented here.
}

\arguments{
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{R}{Absorbance ratio, default is 1}
  \item{d}{Dye used for spectrophotometric measurement, default is "mCP"}
  \item{k}{"m18" for using mCP characterization by Mueller and Rehder (2018)} 
  \item{warn}{"y" to show warnings when S and/or T go beyond the valid range for the chosen d and k; "n" to supress warnings. The default is "y".}
}

\details{The model used to calculate the return value of this function is based on experimental data. It is critical to consider that the formulation refers to the conditions studied during the characterization experiment and is only valid for the studied ranges of temperature and salinity:
\itemize{
\item Mueller and Rehder (2018): S ranging between 0 and 40, T ranging between 5 and 35oC, and the dye used being m-Cresol purple (mCP) with R referring to the ratio of absorbances at wavelengths 578 and 434 nm.

}
Note that the arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length.
} 

\value{
  \item{pHspec}{The function returns the pH value of a water sample from absorbance ratios R, obtained from spectrophotometric measurements with pH indicator dyes (on the total scale in mol/kg-soln)}

}
\references{
Mueller, J. D. and Rehder, G., 2018 Metrology for pH Measurements in Brackish Waters - Part 2: Experimental Characterization of Purified meta-Cresol Purple for Spectrophotometric pHT Measurements. \emph{Frontiers in Marine Science} \bold{5:177}, 1-9. https://doi.org/10.3389/fmars.2018.00177


}

\author{
Jens Daniel Mueller \email{jens.mueller@io-warnemuende.de}
Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{amp}}, \code{\link{pHslope}}, \code{\link{pH}}. \code{\link{tris}},
}

\examples{
	##Example should give test value pHT = 7.6713
	 pHspec(S=35, T=25, R=1, d="mCP", k="m18", warn="y")
}
\keyword{utilities}

\encoding{latin1}
\name{errors}
\alias{errors}
\title{Uncertainty propagation for computed marine carbonate system variables}
\description{Estimates combined standard uncertainties in computed carbonate system variables by propagating inout uncertainties (standard uncertainties) in six input variables, including (Orr et al., Mar. Chem., in press):
\itemize{
\item the input pair of carbonate system variables, 
\item the 2 input nutrients (silicate and phosphate concentrations),
\item temperature and salinity. It also accounts for
\item the errors in the key dissociation constants pK0, pK1, pK2, pKb, pKw, pKspa and pKspc
\item the error in total boron
}
}
\usage{errors(flag, var1, var2, S=35, T=25, Patm=1, P=0, Pt=0, Sit=0, 
              evar1=0, evar2=0, eS=0.01, eT=0.01, ePt=0, eSit=0, 
              epK=c(0.002, 0.0075, 0.015, 0.01, 0.01, 0.02, 0.02),
              eBt=0.02, method = "ga", r=0.0, runs=10000,
              k1k2='x', kf='x', ks="d", pHscale="T", b="u74", gas="potential", 
              warn="y", eos = "eos80", long = 1e+20, lat = 1e+20)}
%- maybe also 'usage' for other objects documented here.
\arguments{
\item{flag}{select the pair of carbonate system input variables. The flags to be used are as follows:

flag = 1      pH and CO2 given

flag = 2      CO2 and HCO3 given

flag = 3      CO2 and CO3 given

flag = 4      CO2 and ALK given

flag = 5      CO2 and DIC given

flag = 6      pH and HCO3 given

flag = 7      pH and CO3 given

flag = 8      pH and ALK given

flag = 9      pH and DIC given

flag = 10     HCO3 and CO3 given

flag = 11     HCO3 and ALK given

flag = 12     HCO3 and DIC given

flag = 13     CO3 and ALK given

flag = 14     CO3 and DIC given

flag = 15     ALK and DIC given

flag = 21     pCO2 and pH given

flag = 22     pCO2 and HCO3 given

flag = 23     pCO2 and CO3 given

flag = 24     pCO2 and ALK given

flag = 25     pCO2 and DIC given
}
\item{var1}{Value of the first variable (in mol/kg, except for pH and for pCO2 in \eqn{\mu}atm)}
\item{var2}{Value of the second variable (in mol/kg, except for pH)}
\item{S}{Salinity (practical salinity scale)}
\item{T}{Temperature in degrees Celsius}
\item{Patm}{Surface atmospheric pressure in atm, default is 1 atm}
\item{P}{Hydrostatic pressure in bar (surface = 0)}
\item{Pt}{Concentration of total dissolved inorganic phosphorus (mol/kg); set to 0 if NA}
\item{Sit}{Concentration of total dissolved inorganic silicon (mol/kg); set to 0 if NA}
\item{evar1}{Standard uncertainty in var1 of input pair of carbonate system variables}
\item{evar2}{Standard uncertainty in var2 of input pair of carbonate system variables}
\item{eS}{Standard uncertainty in salinity; default is 0.01}
\item{eT}{Standard uncertainty in temperature (degree C); default is 0.01}
\item{ePt}{Standard uncertainty in total dissolved inorganic phosphorus concentration (mol/kg)}
\item{eSit}{Standard uncertainty in total dissolved inorganic silicon concentration (mol/kg)}
\item{epK}{Standard uncertainty) in 7 key dissociation constants: pK0, pK1, pK2, pKb, pKw, pKspa and pKspc. This is a vector. The default is c(0.002, 0.0075, 0.015, 0.01, 0.01, 0.02, 0.02).}
\item{eBt}{Standard uncertainty in total boron, given as a
  relative fractional error. The default is 0.02, which equates to a 2\% error}
\item{method}{Case insensitive character string to choose the error-propagation method: 1) Gaussian, 
2) Method of Moments, or 3) Monte Carlo).\cr
These methods are specified using the 2-letter codes "ga", "mo", or "mc", respectively. The default is "ga" (Gaussian).}
\item{r}{Correlation coefficient between standard uncertainties of var1 and var2 (only useful with method="mo",
i.e., ignored for the 2 other methods, the default is r=0.0}
\item{runs}{Number of random samples (ignored unless method="mc"; the default is runs=10000}
\item{k1k2}{"cw" for using K1 and K2 from Cai & Wang (1998), "l" from Lueker et al. (2000), "m02" from Millero et al. (2002), "m06" from Millero et al. (2006), "m10" from Millero (2010), "mp2" from Mojica Prieto et al. (2002), "p18" from Papadimitriou et al. (2018), "r" from Roy et al. (1993), "sb21" from Shockman & Byrne (2021), "s20" from Sulpis et al. (2020), and "w14" from Waters et al. (2014). "x" is the default flag; the default value is then "l", except if T is outside the range 2 to 35oC and/or S is outside the range 19 to 43. In these cases, the default value is "w14".}
\item{kf}{"pf" for using Kf from Perez and Fraga (1987) and "dg" for using Kf from Dickson and Riley (1979 in Dickson and Goyet, 1994). "x" is the default flag; the default value is then "pf", except if T is outside the range 9 to 33oC and/or S is outside the range 10 to 40. In these cases, the default is "dg".}
\item{ks}{"d" for using Ks from Dickson (1990) and "k" for using Ks from Khoo et al. (1977), default is "d"}
\item{pHscale}{"T" for the total scale, "F" for the free scale and "SWS" for using the seawater scale, default is "T" (total scale)}
\item{b}{Concentration of total boron. "l10" for the Lee et al. (2010) formulation or "u74" for the Uppstrom (1974) formulation, default is "u74" }
\item{gas}{used to indicate the convention for INPUT pCO2, i.e., when it is an input variable (flags 21 to 25): "insitu" indicates it is referenced to in situ pressure and in situ temperature; "potential" indicates it is referenced to 1 atm pressure and potential temperature; and "standard" indicates it is referenced to 1 atm pressure and in situ temperature. All three options should give identical results at surface pressure. This option is not used when pCO2 is not an input variable (flags 1 to 15). The default is "potential".}
\item{warn}{"y" to show warnings when T or S go beyond the valid range for constants; "n" to supress warnings. The default is "y".}
\item{eos}{"teos10" to specify T and S according to Thermodynamic Equation Of Seawater - 2010 (TEOS-10); "eos80" to specify T and S according to EOS-80.}
\item{long}{longitude of data point, used when eos parameter is "teos10" as a conversion parameter from absolute to practical salinity.}
\item{lat}{latitude of data point, used when eos parameter is "teos10".}
}

\details{
Complete information on routine uncertainty propagation for the marine carbon dioxide system can be found in Orr et al. (in press). This function requires users to specify each input standard uncertainty as either the standard deviation or the standard error of the mean. The latter implies much smaller propagated uncertainties, but is appropriate only when interested in the error in the mean, not the error of a given measurement. Beware that it is easy to fool oneself when using the standard error of the mean rather than the standard deviation.

This function requires different types of standard uncertainties:
\itemize{
\item Standard uncertainties for evar1, evar2, eS, eT, ePt, eSit (same units as the input data, e.g., mol/kg);
\item Standard uncertainties in pK units for epK; and
\item Standard uncertainties in relative fractional units (between 0.0 and 1.0) for eBt.
}

This function propagates standard uncertainty from input to output variables using one of three methods: 
\itemize{
\item Gaussian:
     The Gaussian method is the standard technique for estimating a
     computed variable's (z) second moment (its variance or standard
     deviation) based on a first-order approximation to z.  More
     precisely, we use here the basic 1st order, 2nd moment
     uncertainty analysis (a type of Taylor expansion), assuming no
     covariance between input variables.  This is the approach used by
     Dickson and Riley (1978). It is the default method.

\item Method of moments: The method of moments is a more general form
     of the Gaussian method. But in addition, it also accounts for
     covariance between input variables.  In this case, the 'errors'
     routine allows the user to specify a value of the correlation
     coefficient 'r', having a value between -1.0 and 1.0, to indicate
     the correlation between standard uncertainties of the input pair of
     carbonate system variables. That correlation is used to compute
     the covariance. But by default, it is assumed that there is no
     covariance (r=0.0).

\item Monte Carlo:
     The Monte Carlo method is a brute-force approach relying on
     repeated random sampling of input errors, adding those to each
     input variables, calculating the corresponding output variables
     for each sample, and finally assessing the standard deviation in
     each output variables.
}

This function has many input parameters that are identical to those in the carb function. For their details, refer to the 'carb' documentation.

All parameters may be scalars or vectors except epK, eBt, method, runs, and gas.

\itemize{
  \item runs and eBt must be scalars
  \item method and gas must each consist of a character string
  \item epK may be a vector of 7 values. In that case, it must list errors for 
    pK0, pK1, pK2, pKb, pKw, pKspa and pKspc, respectively.
    That set of errors is identical for all input data.
    Alternatively, users may specify 'epK=NULL' or 'epK=0' to set all 7 values to zero
    and thus neglect errors in the equilibrium constants. 
}
In constrast, for evar1, evar2, r, eS, eT, ePt and eSit:

\itemize{
  \item  if they are vectors, they represent standard uncertainties associated with each data point
  \item  if they are scalars (single real numbers), they represent one standard uncertainty value each associated to all data points
}

The same remark applies to parameter r (correlation coefficient).

long and lat are used as conversion parameters from absolute to practical salinity: when seawater is not of standard composition, practical salinity alone is not sufficient to compute absolute salinity and vice-versa. One needs to know the density. When long and lat are given, density is inferred from WOA silicate concentration at given location. When they are not, an arbitrary geographic point is chosen: mid equatorial Atlantic. Note that this implies an error on computed salinity up to 0.02 g/kg.
}

\section{Correlation coefficient}{
By default, 'r' is zero. However, for some pairs the user may want to specify a different value. For example, measurements of pCO2 and pH are often anti-correlated. The same goes for two other pairs: 'CO2 and CO3' and 'pCO2 and CO3'.  But even for these cases, care is needed before using non-zero values of 'r'.

When the user wishes to propagate standard uncertainties for an individual measurement, 'r' should ALWAYS be zero if each member of the input pair is measured independently. In this case, we are interested in the correlation between the uncertainties in those measurements, not in the correlation between the measurements themselves. Uncertainties from those measurements are probably not correlated if they come from different instruments. Conversely, if users are interested in the error in the mean of a distribution of measurements (i.e., if they are propagating standard errors instead of standard deviations), one should then also account for the correlation between the measurements of the two variables of the input pair.

For input pairs where one member is pH (flags 1, 6, 7, 8, 9, and 21), this 'errors' function automatically inverses the sign of 'r'. The reason for that is that the associated derivatives are computed in terms of the hydrogen ion concentration (H+), not pH. Therefore for each of these 6 flags, if the user wants to compute their own 'r' that should be done by (1) using the H+ concentration instead of pH, and (2) inversing the sign of that computed 'r' before passing it as an argument to this routine. Usually though (when not calculating r for pH), the user may just use the 'r' in the expected way.  For example, to include the covariance term when there is a perfect anticorrelation of pH with pCO2, one would use 'r=-1.0'.

}

\section{Computation time}{
Computation time depends on the method chosen; the Monte Carlo method takes much longer to execute. The computational time required for the Monte Carlo method is proportional to the number of runs. More runs, implies improved accuracy: runs = 10000 appears a minimum to obtain an accuracy of less than 1\%. Accuracy is inversely proportional to the number of runs.

Computation time also depends on the chosen pair of input variables. For example, with the input pair DIC and Total alkalinity (flag=15), it is much longer than for input pair pH and Total alkalinity (flag=8)
}

\value{The function returns a 2-dimensional dataframe, with the following columns:
 \item{H}{combined standard uncertainty in [H+] concentration (mol/kg)}
 \item{pH}{combined standard uncertainty in pH}
 \item{CO2}{combined standard uncertainty in CO2 concentration (mol/kg)}
 \item{pCO2}{combined standard uncertainty in "standard" pCO2, CO2 partial pressure computed at in situ temperature and atmospheric pressure (uatm)}
 \item{fCO2}{combined standard uncertainty in "standard" fCO2, CO2 fugacity computed at in situ temperature and atmospheric pressure (uatm)}
 \item{HCO3}{combined standard uncertainty in HCO3 concentration (mol/kg)}
 \item{CO3}{combined standard uncertainty in CO3 concentration (mol/kg)}
 \item{DIC}{combined standard uncertainty in DIC concentration (mol/kg)}
 \item{ALK}{combined standard uncertainty in ALK, total alkalinity (mol/kg)}
 \item{OmegaAragonite}{combined standard uncertainty in Omega aragonite (aragonite saturation state)}
 \item{OmegaCalcite}{combined standard uncertainty in Omega calcite   (calcite saturation state)}
 
If all input data have the same 'flag' value, the returned data frame does not show combined standard uncertainties on input pair of carbonate system variables. For example, if all input flags are 15, the input pair is DIC and ALK; hence, errors on DIC and ALK are not returned.
}

\references{
Cai W. J., and Wang Y., 1998. The chemistry, fluxes, and sources of carbon dioxide in the estuarine waters of the Satilla and Altamaha Rivers, Georgia. \emph{Limnology and Oceanography} \bold{43}, 657-668.

Dickson, A. G. and Riley, J. P., 1978 The effect of analytical error on the evaluation of the components of the aquatic carbon-dioxide system, \emph{Marine  Chemistry}, \bold{6}, 77-85.

Dickson A. G. and Riley J. P., 1979 The estimation of acid dissociation constants in seawater media from potentiometric titrations with strong base. I. The ionic product of water. \emph{Marine Chemistry} \bold{7}, 89-99.

Dickson A. G., 1990 Standard potential of the reaction: AgCI(s) + 1/2H2(g) = Ag(s) + HCI(aq), and the standard acidity constant of the ion HSO4 in synthetic sea water from 273.15 to 318.15 K. \emph{Journal of Chemical Thermodynamics} \bold{22}, 113-127.

Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.

Khoo H. K., Ramette R. W., Culberson C. H. and Bates R. G., 1977 Determination of hydrogen ion concentration in seawater from 5 to 40oC: standard potentials at salinities from 20 to 45. \emph{Analytical Chemistry} \bold{22}, vol49 29-34.

Lee K., Tae-Wook K., Byrne R.H., Millero F.J., Feely R.A. and Liu Y-M, 2010 The universal ratio of the boron to chlorinity for the North Pacific and North Atlantoc oceans. \emph{Geochimica et Cosmochimica Acta} \bold{74} 1801-1811.

Lueker T. J., Dickson A. G. and Keeling C. D., 2000 Ocean pCO2 calculated from dissolved inorganic carbon, alkalinity, and equations for K1 and K2: validation based on laboratory measurements of CO2 in gas and seawater at equilibrium. \emph{Marine Chemistry} \bold{70} 105-119.

Millero F. J., 1995. Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica Cosmochimica Acta} \bold{59}: 661-677.

Millero F. J., 2010. Carbonate constant for estuarine waters. \emph{Marine and Freshwater Research} \bold{61}: 139-142.

Millero F. J., Graham T. B., Huang F., Bustos-Serrano H. and Pierrot D., 2006. Dissociation constants of carbonic acid in seawater as a function of salinity and temperature.  \emph{Marine Chemistry} \bold{100}, 80-84.

Orr J. C., Epitalon J.-M., Dickson A. and Gattuso J.-P., in press. Routine uncertainty propagation for the marine carbon dioxide system. \emph{Marine Chemistry}. 

Orr J. C., Epitalon J.-M. and Gattuso J.-P., 2015. Comparison of seven packages that compute ocean carbonate chemistry. \emph{Biogeosciences} \bold{12}, 1483-1510.

Perez F. F. and Fraga F., 1987 Association constant of fluoride and hydrogen ions in seawater. \emph{Marine Chemistry} \bold{21}, 161-168.

Roy R. N., Roy L. N., Vogel K. M., Porter-Moore C., Pearson T., Good C. E., Millero F. J. and Campbell D. M., 1993. The dissociation constants of carbonic acid in seawater at salinities 5 to 45 and temperatures 0 to 45oC. \emph{Marine Chemistry} \bold{44}, 249-267.

Schockman, K.M., Byrne, R.H., 2021. Spectrophotometric determination of the bicarbonate dissociation constant in seawater, \emph{Geochimica et Cosmochimica Acta}.

Uppstrom L.R., 1974 The boron/chlorinity ratio of the deep-sea water from the Pacific Ocean. \emph{Deep-Sea Research I} \bold{21} 161-162.

Waters, J., Millero, F. J., and Woosley, R. J., 2014. Corrigendum to ``The free proton concentration scale for seawater pH'', [MARCHE: 149 (2013) 8-22], \emph{Marine  Chemistry}, \bold{165}, 66-67.

Weiss, R. F., 1974. Carbon dioxide in water and seawater: the solubility of a non-ideal gas, \emph{Marine  Chemistry}, \bold{2}, 203-215.

Weiss, R. F. and Price, B. A., 1980. Nitrous oxide solubility in water and seawater, \emph{Marine  Chemistry}, \bold{8}, 347-359.

Zeebe R. E. and Wolf-Gladrow D. A., 2001 \emph{CO2 in seawater: equilibrium, kinetics, isotopes}. Amsterdam: Elsevier, 346 pp.
}

\author{
Jean-Marie Epitalon, James Orr, and Jean-Pierre Gattuso\email{jean-pierre.gattuso@imev-mer.fr}
}

\examples{

## 1) For the input pair ALK and DIC (var1 and var2 when flag=15),
## compute resulting uncertainty from given uncertainty on ALK and DIC (5 umol/kg)
## and default uncertainties in dissociation constants and total boron
## using the default method (Gaussian)
errors(flag=15, var1=2300e-6, var2=2000e-6, S=35, T=25, P=0, Patm=1.0, Pt=0, Sit=0, 
       evar1=5e-6, evar2=5e-6, eS=0, eT=0, ePt=0, eSit=0, 
       pHscale="T", kf="pf", k1k2="l", ks="d", b="u74")
## Typical output:
## H             pH          CO2           fCO2      pCO2      HCO3          ...
## 3.721614e-10  0.01796767  5.441869e-07  19.25338  19.31504  9.170116e-06  ...

## 2) Do the same as in one, but assign a 4% uncertainty to total boron
##    This uncertainty is the amount by which estimates from Lee et al (2010) and 
##    Uppstrom (1974) differ. The default for the latter is eBt=0.02.
errors(flag=15, var1=2300e-6, var2=2000e-6, S=35, T=25, P=0, Patm=1.0, Pt=0, Sit=0, 
       evar1=5e-6, evar2=5e-6, eS=0, eT=0, ePt=0, eSit=0, eBt=0.04,
       pHscale="T", kf="pf", k1k2="l", ks="d", b="u74")

## 3) For the input pair pH and ALK (var1 and var2 when flag=8)
## compute standard errors in output variables from errors in input variables, i.e., 
## for pH (0.005 pH units) and in ALK (5 umol/kg), along with
## errors in total dissolved inorganic phosphorus (0.1 umol/kg) and
## total dissolved inorganic silicon (2 umol/kg) concentrations, while
## assuming no uncertainty in dissociation constants & boron, using the Gaussian method:
errors(flag=8, var1=8.25, var2=2300e-6,  S=35, T=25, P=0, Patm=1.0, Pt=0, Sit=0, 
       evar1=0.005, evar2=5e-6, eS=0, eT=0, ePt=0.1, eSit=2, epK=0, eBt=0,
       method="ga", pHscale="T", kf="pf", k1k2="l", ks="d", b="u74")

## 4) For the input pair pCO2 and pH (var1 and var2 when flag=21)
## compute standard errors in output variables from errors in input variables, i.e., 
## for pCO2 (2 uatm) and pH (0.005 pH units), with no uncertainties in Pt and Sit
## nor in the dissociation constants BUT a perfect anticorrelation between pCO2 and pH,
## (the input pair) using the Method of moments:
errors(flag=21, var1=400, var2=8.1,  S=35, T=25, P=0, Patm=1.0, Pt=0, Sit=0, 
       evar1=2, evar2=0.005, eS=0, eT=0, ePt=0.0, eSit=0, epK=0, eBt=0, 
       method="mo", r=-1.0, pHscale="T", kf="pf", k1k2="l", ks="d", b="u74")

## 5) Use vectors as arguments and compute errors on all output variables
## using Monte Carlo method taking into account input errors on pH, ALK, DIC
## and dissociation constants (pKx)
flag <- c(8, 15)
var1 <- c(8.2, 0.002394, 8.25)
var2 <- c(0.002343955, 0.002017)
S <- c(35, 35)
T <- c(25, 25)
P <- 0
Pt <- 0
Sit <- 0
evar1 <- c(0.005, 2e-6)
evar2 <- c(2e-6, 2e-6)
epKx <- c(0.002, 0.01, 0.02, 0.01, 0.01, 0.01, 0.01)
eBtx = 0.01
method <- "mc"
kf <- "pf"
k1k2 <- "l"
pHscale <- "T"
b <- "u74"
## NOTE that the following is executable but enclosed in "donttest" 
## because it takes too long to run when submiting to CRAN
## and is therefore rejected
\donttest{errors(flag=flag, var1=var1, var2=var2, S=S, T=T, P=P, Pt=Pt, Sit=Sit, 
       evar1=evar1, evar2=evar2, eS=0, eT=0, ePt=0, eSit=0, epK=epKx, eBt=eBtx,
       method=method, runs=10000, kf=kf, k1k2=k1k2, pHscale=pHscale, b=b)}
}

\keyword{utilities}

\encoding{latin1}
\name{sp2sa_chem}
\alias{sp2sa_chem}

\title{From Practical to absolute salinity}
\description{Converts from practical to absolute salinity based on total alkalinity as well as on the concentrations of dissolved inorganic carbon, nitrate and silicate.}
\usage{
sp2sa_chem(SP, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
}
%- maybe also 'usage' for other objects documented here.

\arguments{
  \item{SP}{Practical salinity on the practical salinity scale}
  \item{TA}{Total alkalinity, in mol/kg, default is 2300 µmol/kg}
  \item{DIC}{Dissolved inorganic carbon concentration in mol/kg, default is 2000 µmol/kg}
  \item{NO3}{Total nitrate concentration in mol/kg, default is 0}
  \item{SIOH4}{Total silicate concentration in mol/kg, default is 0}
  }

\value{
  \item{SA}{Absolute salinity (g/kg)}
}

\details{
Converts from practical to absolute salinity from carbonate system parameters and ion concentration which mostly affect water density anomalies.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., Wright D. G. and Millero F. J., 2011. The effects of biogeochemical processes on oceanic conductivity/salinity/density relationships and the characterization of real seawater. \emph{Ocean Science} \bold{7}, 363-387.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
sa2sp_chem does the reverse, sp2sa_geo
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate the absolute salinity of a sample with practical Salinity of 35,
   # Total Alkalinity of 0.00234 mol/kg and DIC of 0.00202 mol/kg
   SA <- sp2sa_chem(SP=35, TA=0.00234, DIC=0.00202)
}

\keyword{utilities}

\encoding{latin1}
\name{theta}
\alias{theta}

\title{Potential temperature of seawater}
\description{Computes theta, the potential temperature of seawater given
original temperature, salinity, pressure, and reference pressure}
\usage{
theta(S=35, T=25, P=0, Pref=0)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{S}{Salinity on the practical salinity scale, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{P}{Hydrostatic pressure in bar (surface = 0; 1000 db = 100 bar), default is 0}
  \item{Pref}{Reference hydrostatic pressure in bar, default is 0}
  }

\value{
  \item{theta}{potential temperature of seawater (C)}
}

\details{
Computes the potential temperature of seawater relative to a
chosen reference pressure following Fofonoff and Millard (1983).
The potential temperature \eqn{\theta}{theta} is the temperature that a
water parcel would have if were moved adiabatically to another pressure
level \code{Pref}.  Typically, the potential temperature is referenced
to the surface (\eqn{Pref=0}).  The potential teperature depends on
the original salinity \code{S}, \emph{in-situ} temperature \code{T}
and pressure \code{P}.

This routine is essentially a wrapper for the \code{\link[oce]{swTheta}}
routine of the 'oce' package. Unlike the latter, pressure units here
are given in bars and method="unesco" is prescribed.
}

\references{
Fofonoff, P. and R. C. Millard Jr, 1983. Algorithms for computation of fundamental properties of seawater. \emph{Unesco Technical Papers in Marine Science}, \bold{44}, 53 pp.}
             
\author{
James Orr \email{james.orr@lsce.ipsl.fr}
}

\seealso{
	\code{\link[oce]{swTheta}}.
}

\examples{
   #Calculate the potential temperature for a sample at 1000 db referenced to the surface
   theta <- theta(S=35, T=25, P=100, Pref=0)
}

\keyword{utilities}

\encoding{latin1}
\name{tris}
\alias{tris}

\title{pH of TRIS buffer}
\description{Calculates the pH value of TRIS buffered artificial seawater solutions (on the total scale in mol/kg-soln)}
\usage{
tris(S=35,T=25,b=0.04,k="d98",warn="y")
%- maybe also 'usage' for other objects documented here.
}

\arguments{
  \item{S}{Salinity, default is 35}
  \item{T}{Temperature in degrees Celsius, default is 25oC}
  \item{b}{Molality if TRIS/TRISH+ in moles per kg of water, default is 0.04 mol/kg-H20}
  \item{k}{"d98" for DelValls and Dickson 1998, "m18" for using tris characterization by Mueller et al (2018), default is "d98"} 
  \item{warn}{"y" to show warnings when S,T and/or b go beyond the valid range for the chosen k; "n" to supress warnings. The default is "y".}
}

\details{The models used to calculate the return value of this function are based on experimental data. It is critical to consider that each formulation refers to the artificial seawater solution applied during the characterization experiment and is only valid for the studied ranges of temperature and salinity:
\itemize{
  \item DelValls and Dickson (1998): S ranging between 20 and 40, T ranging between 0 and 45oC, and b being 0.04 mol/kg-H20.
  \item Mueller et al. (2018): S ranging between 5 and 40, T ranging between 5 and 45oC, and b ranging between 0.01 and 0.04 mol/kg-H20.
}
Note that the arguments can be given as a unique number or as vectors. If the lengths of the vectors are different, the longer vector is retained and only the first value of the other vectors is used. It can therefore be critical to use vectors of the same length.
} 

\value{
  \item{tris}{The function returns the pH value of TRIS buffered artificial seawater solutions (on the total scale in mol/kg-soln)}

}
\references{
DelValls, T. A., and Dickson, A. G., 1998 The pH of buffers based on 2-amino-2-hydroxymethyl-1,3-propanediol ('tris') in synthetic sea water. \emph{Deep Sea Research Part I: Oceanographic Research Papers} \bold{45(9)}, 1541-1554. https://doi.org/10.1016/S0967-0637(98)00019-3

Dickson A. G., Sabine C. L. and Christian J. R., 2007 Guide to best practices for ocean CO2 measurements. \emph{PICES Special Publication} \bold{3}, 1-191.

Mueller, J. D., Bastkowski, F., Sander, B., Seitz, S., Turner, D. R., Dickson, A. G., and Rehder, G., 2018 Metrology for pH measurements in brackish waters- Part 1: Extending electrochemical pHT measurements of TRIS buffers to salinities 5-20. \emph{Frontiers in Marine Science} \bold{5:176}, 1-12. https://doi.org/10.3389/fmars.2018.00176
}

\author{
Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr} and Jens Daniel Mueller \email{jens.mueller@io-warnemuende.de} 
}

\seealso{
	\code{\link{amp}}, \code{\link{pHslope}}, \code{\link{pH}}.
}

\examples{
	##Example from Mueller et al. (2018), should give test value pHT = 8.0703
	 tris(S=20,T=25,b=0.04,k="m18")
}
\keyword{utilities}

\name{Pcoeffs}

\docType{data}

\alias{Pcoeffs}

\title{Coefficients used for pressure-correcting the equilibrium constants}

\description{
Pressure corrections are based on the following equations: 

\deqn{
  \ln{\frac{K_{i}^{P}}{K_{i}^{0}}} = -\frac{\Delta V_{i}}{RT}.P + 0.5\frac{\Delta K_{i}}{RT}.P^{2} 
 }

with 
\deqn{
   \Delta V_{i} = a_{0} + a_{1}T + a_{2}T^{2}
 }
and 
\deqn{
   \Delta K_{i} = b_{0} + b_{1}T + b_{2}T^{2}
}


The variables are:

\itemize{
	\item K indicating the type of equilibrium constant
	\item coefficient \eqn{a_0}
	\item coefficient \eqn{a_1}
	\item coefficient \eqn{a_2}
	\item coefficient \eqn{b_0}
	\item coefficient \eqn{b_1}
	\item coefficient \eqn{b_2}

	}
}
\usage{Pcoeffs}

\format{A data frame with 15 rows and 7 variables}

\source{
Millero F. J., 1979 The thermodynamics of the carbonate system in seawater. \emph{Geochemica et Cosmochemica Acta} \bold{43}: 1651-1661.

Millero F. J., 1983 Influence of pressure on chemical processes in the sea. pp. 1-88. In J. P. Riley and R. Chester (eds.), Chemical Oceanography. Academic Press, New York.

Millero F. J., 1995 Thermodynamics of the carbon dioxide system in the oceans. \emph{Geochimica Cosmochimica Acta} \bold{59}: 661-677.
}

\seealso{
Pcorrect
}

\details{
For Kb, to be consistent with Millero (1979) a2 was changed to -2.608e-3 instead of 2.608e-3 (value given in Millero, 1995)
For Kw, coefficients are from Millero (1983).
}

\keyword{datasets}

\encoding{latin1}
\name{sf_calc}
\alias{sf_calc}

\title{Calculation of calibrated pH for seaFET sensor}

\description{
The function sf_calc() calculates pH time series (pHint_tot and pHext_tot) for SeaFET pH sensors, using calibration coefficients E0int25 and E0ext25 from the function sf_calib(). Both functions are R-adaptations from MATLAB scripts published by Bresnahan et al. (2014).
}

\usage{
sf_calc(calEint=0.0865, calEext= -0.93, E0int25 =-0.39, 
  E0ext25=-1.46, calT=16.2, calSal=35.6)
}
%- maybe also 'usage' for other objects documented here.
\arguments{
  \item{calEint}{EINT (V), default is 0.0865}
  \item{calEext}{EEXT (V), default is -0.93}
  \item{E0int25}{Coeficient of calibration related to the internal sensor and obtained via sf_calib function, default is -0.39. If time-serie, we use the mean per periode of deployement}
  \item{E0ext25}{Coeficient of calibration related to the external sensor and obtained via sf_calib function, default is -1.46. If time-serie, we use the mean per periode of deployement}
  \item{calT}{Temperature in degrees Celsius, default is 16.2}
  \item{calSal}{Salinity, default is 35.6}
}

\details{
Input values should be vectors of equal length. E0int25 and E0ext25 should be constant throughout the time series. When multiple reference samples are available for one SeaFET deployment, mean E0int25 and mean E0ext25 should be calculated and used in sf_calc(). Each unique SeaFET deployment requires a new calculation of mean E0int25 and mean E0ext25 based on reference pH samples (total hydrogen ion scale). For detailed SeaFET calibration instructions and recommendations see Bresnahan et al. (2014) and Rivest et al. (2016).
} 

\value{
This function returns a dataframe comprising 2 variables: 
  \item{pHint_tot}{Calibrated pH of the internal sensor at in situ temperature.}
  \item{pHext_tot}{Calibrated pH of the external sensor at in situ temperature.}

}
\references{
Bresnahan, P. J. J., Martz, T. R., Takeshita, Y., Johnson, K. S., and LaShomb, M., 2014. Best practices for autonomous measurement of seawater pH with the Honeywell Durafet. \emph{Methods Oceanogr.} \bold{9}, 44-60.

Martz, T.R., Connery, J.G., and Johnson, K.S., 2010. Testing the Honeywell Durafet for seawater pH applications. \emph{Limnol. Oceanogr. Meth.} \bold{8}, 172-184.

Rivest, E.B., O'Brien, M., Kapsenberg, L., Gotschalk, C.C., Blanchette, C.A., Hoshijima, U., and Hofmann, G.E., 2016. Beyond the benchtop and the benthos: dataset management planning and design for time series of ocean carbonate chemistry associated with Durafet(c)-based pH sensors. \emph{Ecological Informatics} \bold{36}, 209-220.
}

\author{
Samir Alliouane, Lydia Kapsenberg, Jean-Pierre Gattuso \email{jean-pierre.gattuso@imev-mer.fr}
}

\seealso{
	\code{\link{sf_calib}}.
}

\examples{
sf_calc(calEint=0.0865, calEext= -0.93, E0int25 =-0.39, 
  E0ext25=-1.46, calT=16.2, calSal=35.6)

## Using the test file seaFET
sf_calc(calEint=seaFET$Eint, calEext=seaFET$Eext, 
  E0int25=seaFET$E0int25, E0ext25=seaFET$E0ext25, 
  calT=seaFET$Temperature, calSal=seaFET$Salinity)
}
\keyword{utilities}

\encoding{latin1}
\name{teos2eos_chem}
\alias{teos2eos_chem}

\title{Convert temperature and salinity from TEOS-10 to EOS-80}
\description{Converts conservative temperature to in situ temperature and absolute salinity to practical salinity (SP). Salinity conversion depends on total alkalinity as well as on the concentrations of dissolved inorganic carbon, nitrate and silicate.}

\usage{
teos2eos_chem(SA, CT, P=0, TA=2300e-6, DIC=2000e-6, NO3=0, SIOH4=0)
}
%- maybe also 'usage' for other objects documented here.

\arguments{
  \item{SA}{Absolute salinity in g/kg}
  \item{CT}{Conservative temperature in degrees C}
  \item{P}{Sea water pressure in dbar}
  \item{TA}{Total alkalinity, in mol/kg, default is 2300 µmol/kg}
  \item{DIC}{Dissolved inorganic carbon concentration in mol/kg, default is 2000 µmol/kg}
  \item{NO3}{Total nitrate concentration in mol/kg, default is 0}
  \item{SIOH4}{Total silicate concentration in mol/kg, default is 0}
  }

\value{The function returns a data frame containing the following columns:
  \item{T}{In situ temperature (deg C)}
  \item{SP}{Practical salinity (psu)}
}

\details{
Conversion from absolute to practical salinity depends on carbonate system parameters and ion concentration which mostly affect water density anomalies.
}

\references{
McDougall T. J., Jackett D. R., Millero F. J., Pawlowicz R. and Barker P. M., 2012. A global algorithm for estimating Absolute Salinity. \emph{Ocean Science} \bold{8}, 1123-1134.

Pawlowicz R., Wright D. G. and Millero F. J., 2011. The effects of biogeochemical processes on oceanic conductivity/salinity/density relationships and the characterization of real seawater. \emph{Ocean Science} \bold{7}, 363-387.

Pawlowicz R., 2013. What every oceanographer needs to know about TEOS-10 (The TEOS-10 Primer). http://www.teos-10.org/
}
             
\seealso{
eos2teos_chem, teos2eos_geo, sa2sp_cham, package gsw
}

\author{
Jean-Marie Epitalon
}

\examples{
   # Calculate in situ temperature and practical salinity of a sample with 
   # Absolute salinity of 35 g/kg, Conservative temperature of 18 deg C,
   # at 0 dbar and Total alkalinity of 0.00234 mol/kg and DIC of 0.00202 mol/kg
   f <- teos2eos_chem(SA=35, CT=18, P=0, TA=0.00234, DIC=0.00202)
   T <- f$T     # insitu temperature
   SP <- f$SP     # Practical salinity
}

\keyword{utilities}
